/*
 *  Lucidor - e-book reader
 *  Copyright © 2013  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("chrome://lucidor/content/utils.jsm");
Components.utils.import("resource://gre/modules/NetUtil.jsm");

const ADE_NS = "http://ns.adobe.com/digitaleditions/annotations";
const DC_NS = "http://purl.org/dc/elements/1.1/";
const XHTML_NS = "http://www.w3.org/1999/xhtml";
const XUL_NS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";

var g_book = null;
var g_note_manager = null;

function init() {
    let ioService = Components.classes["@mozilla.org/network/io-service;1"].
	getService(Components.interfaces.nsIIOService);
    let docuri = ioService.newURI(document.documentURI, "UTF-8", null);
    docuri.QueryInterface(Components.interfaces.nsIURL);

    let bookid = null;
    let params = docuri.query.split("&");
    for (let par in loop(params)) {
	let pair = par.split("=");
	if (pair[0] == "bookid")
	    bookid = pair[1];
    }

    debug("bookid: " + bookid);

    let bookman = Components.classes["@lucidor.org/bookmanagerService;1"].
	getService().wrappedJSObject;
    g_book = bookman.getBook(bookid);
    debug("book: " + g_book.sha1);

    g_note_manager = Components.classes["@lucidor.org/notemanagerService;1"].getService().wrappedJSObject;
    g_note_manager.open();

    list_notes();
}

function unload() {
}

/*
 * source: local, loaded1, loaded2 ..., server
 * book: book1, book2 ...
 * notes
 */

function list_pubs() {
    if (!g_nm)
	g_nm = g_note_manager;

    let pubs = g_nm.find_all_publications();
    for (let pid in loop(pubs)) {
	let pub = g_nm.get_publication(pid);
    }
}

function list_notes() {
    if (!g_nm)
	g_nm = g_note_manager;

    let iframe = document.getElementById("notes");
    while (iframe.childNodes.length > 0)
	iframe.removeChild(iframe.firstChild);

    let pubs = g_nm.find_all_publications();
    for (let pid in loop(pubs)) {
	let nids = g_nm.find_notes_id(pid);

	for (let nid in loop(nids)) {
	    list_note(nid, iframe);
	}
    }

    /*
    if (notes.length > 0)
	document.getElementById("export_notes_button").hidden = false;
    else
	document.getElementById("export_notes_button").hidden = true;
    */
}

function sort_notes(notes) {
    let sord = g_book.get_spine_order();
    notes.sort(function(na, nb) {
	if (na.targets.length == 0 && nb.targets.length == 0)
	    return 0;
	if (na.targets.length == 0)
	    return 1;
	if (nb.targets.length == 0)
	    return -1;

	let a = na.targets[0];
	let b = nb.targets[0];

	if (a.part != b.part) {
	    if (sord[a.part] == sord[b.part])	// One is fallback item
		return 0;
	    if (sord[a.part] == -1 || sord[a.part] == undefined)
		return 1;
	    if (sord[b.part] == -1 || sord[b.part] == undefined)
		return -1;
	    if (sord[a.part] < sord[b.part])
		return -1;
	    return 1;
	}

	// a.part == b.part
	if (a.fixptr || b.fixptr) {
	    if (!a.fixptr)
		return -1;
	    if (!b.fixptr)
		return 1;
	    return compare_fixptr(a.fixptr, b.fixptr);
	} else if (a.location || b.location) {
	    if (!a.location)
		return -1;
	    if (!b.location)
		return 1;
	    return compare_xpathlocations(a.location, b.location);
	}
	//if (!a.fixptr && !b.fixptr && !a.location && !b.location)
	return 0;
    });
}

function list_note(nid, iframe) {
    debug("note: " + nid);
    let note = g_nm.get_note(nid);

    // Main frame
    let div = document.createElementNS(XHTML_NS, "div");
    div.setUserData("luci", note.id, null);
    div.setAttribute("class", "noteitem");
    iframe.appendChild(div);

    // Comment and target div
    let div1 = document.createElementNS(XHTML_NS, "div");
    div.appendChild(div1);
    
    // opendiv
    let div2 = document.createElementNS(XHTML_NS, "div");
    div.appendChild(div2);

    make_link(div, note);
}

    /*
      <div class='navitem' data='nid'>
        <div>
	  <a href='#'>*</a> <span>...:</span><br/>
	  <a href="#">• ...</a><br/>
	  <a href="#">• ...</a>
	</div>
	<div>

	</div>
      </div>
     */
function make_link(notediv, note) {
    let container = notediv.childNodes[0];
    let opendiv = notediv.childNodes[1];
    while (container.hasChildNodes())
	container.removeChild(container.firstChild);

    let handle = document.createElementNS(XHTML_NS, "a");
    container.appendChild(handle);
    handle.setAttribute("href", "#");

    let span = document.createElementNS(XHTML_NS, "span");
    handle.appendChild(span);
    set_as_closed(span);

    handle.addEventListener("click", {
	opendiv: opendiv,
	nid: note.id,
	span: span,
	handleEvent: function(e) {
	    toggle_note(this.span, this.opendiv, this.nid);
	}
    }, false);

    handle.appendChild(document.createTextNode(" "));	// NBSP

    span = document.createElementNS(XHTML_NS, "span");
    span.setAttribute("class", "infoheader");
    let text = note.content;
    if (!text)
	text = "...";
    if (text.length > 60)
	text = text.substring(0, 55) + "...";
    //text += ":";
    span.appendChild(document.createTextNode(text));
    handle.appendChild(span);

    let sharedbundle = document.getElementById("lucifox-shared_strings");

    for (let target in loop(note.targets)) {
	let linkdiv = document.createElementNS(XHTML_NS, "div");
	linkdiv.setAttribute("class", "clear");
	container.appendChild(linkdiv);

	let link = document.createElementNS(XHTML_NS, "a");
	linkdiv.appendChild(link);

	link.setAttribute("class", "clear");
	/*
	link.setAttribute("href", "#");
	link.setAttribute("class", "blink clear");
	link.addEventListener("click", {
	    opendiv: notediv.childNodes[1],
	    nid: note.id,
	    target: target,
	    handleEvent: function(e) {
		view_note(this.opendiv, this.nid, this.target);
	    }
	}, false);
	*/

	text = target.text;
	if (!text)
	    text = "...";
	if (text.length > 50)
	    text = text.substring(0, 45) + "...";
	link.appendChild(document.createTextNode("– " + text));	// ↗ •
	link.setAttribute("title", note.content);

	if (!note.readonly) {
	    // Remove button
	    linkdiv.appendChild(document.createTextNode(" "));

	    link = document.createElementNS(XHTML_NS, "a");
	    linkdiv.appendChild(link);
	    link.setAttribute("href", "#");
	    link.setAttribute("class", "button right");
	    link.addEventListener("click", {
		pid: target.id,
		nid: note.id,
		opendiv: notediv.childNodes[1],
		handleEvent: function(e) {
		    remove_target(this.pid, this.nid, this.opendiv);
		}
	    }, false);
	    link.setUserData("luciopen", true, null);
	    let text = sharedbundle.getString("remove");
	    link.appendChild(document.createTextNode("[" + text + "]"));
	    link.hidden = true;
	}
    }
}

// ⊞⊟﹢﹣▸▾▹▿
function set_as_open(element) {
    if (element.firstChild)
	element.removeChild(element.firstChild);
    if (element.firstChild)
	element.insertBefore(document.createTextNode("▿"), element.firstChild);
    else
	element.appendChild(document.createTextNode("▿"));
}

function set_as_closed(element) {
    if (element.firstChild)
	element.removeChild(element.firstChild);
    if (element.firstChild)
	element.insertBefore(document.createTextNode("▹"), element.firstChild);
    else
	element.appendChild(document.createTextNode("▹"));
}

function toggle_note(handle, opendiv, nid) {
    if (opendiv.hasChildNodes()) {
	close_note(opendiv);
    } else {
	open_note(nid, opendiv, false);
    }
}

function close_note(opendiv) {
    while (opendiv.hasChildNodes())
	opendiv.removeChild(opendiv.firstChild);
    let notediv = opendiv.parentNode;
    for (let el in loop(notediv.getElementsByTagName("*"))) {
	if (el.getUserData("luciopen"))
	    el.hidden = true;
    }

    let grip = notediv.childNodes[0].childNodes[0].childNodes[0];
    set_as_closed(grip);
}

function open_note(nid, opendiv, remake) {
    debug("open_note");

    let bundle = document.getElementById("lucifox-strings");
    let sharedbundle = document.getElementById("lucifox-shared_strings");

    let notediv = opendiv.parentNode;
    for (let el in loop(notediv.getElementsByTagName("*"))) {
	if (el.getUserData("luciopen"))
	    el.hidden = false;
    }

    let grip = notediv.childNodes[0].childNodes[0].childNodes[0];
    set_as_open(grip);

    if (remake) {
	while (opendiv.hasChildNodes())
	    opendiv.removeChild(opendiv.firstChild);
    } else if (opendiv.hasChildNodes())
	return;

    let note = g_nm.get_note(nid);

    let div = document.createElementNS(XHTML_NS, "div");
    div.setAttribute("class", "note");

    opendiv.appendChild(div);

    /*
    if (!note.readonly) {
	// Add target selection
	let adiv = document.createElementNS(XHTML_NS, "div");
	adiv.setAttribute("class", "right clear");
	div.appendChild(adiv);

	let link = document.createElementNS(XHTML_NS, "a");
	adiv.appendChild(link);
	link.setAttribute("href", "#");
	link.setAttribute("class", "button");
	link.addEventListener("click", {
	    nid: note.id,
	    opendiv: opendiv,
	    handleEvent: function(e) { add_target(this.nid, this.opendiv); }
	}, false);
	let text = bundle.getString("add_pointer");
	link.appendChild(document.createTextNode("[" + text + "]"));
    }
    */

    // Space
    let br = document.createElementNS(XHTML_NS, "div");
    br.setAttribute("class", "clear");
    br.appendChild(document.createTextNode(" "));	// NBSP
    div.appendChild(br);

    // Modification date
    if (note.modified) {
	div.appendChild(document.createTextNode(note.modified.toLocaleString()));
	div.appendChild(document.createElementNS(XHTML_NS, "br"));
    }

    // Comment text
    let tb = document.createElementNS(XUL_NS, "textbox");
    tb.setAttribute("multiline", "true");
    tb.setAttribute("rows", "4");
    tb.setAttribute("cols", "80");
    tb.setAttribute("flex", "1");
    tb.setAttribute("value", note.content);
    if (note.readonly)
	tb.setAttribute("readonly", "true");
    //tb.value = note.content;
    div.appendChild(tb);

    div.appendChild(document.createElementNS(XHTML_NS, "br"));

    // Highlight checkbox
    let cb = document.createElementNS(XUL_NS, "checkbox");
    let text = bundle.getString("highlight");
    cb.setAttribute("label", text);
    if (note.readonly)
	cb.setAttribute("disabled", "true");
    else {
	cb.addEventListener("click", {
	    cb: cb,
	    handleEvent: function(e) {
		if (this.cb.checked)
		    this.cb.setAttribute("checked", "true");
		else
		    this.cb.removeAttribute("checked");
	    }
	}, false);
    }
    div.appendChild(cb);
    if (note.style == "highlight")
	cb.setAttribute("checked", "true");

    div.appendChild(document.createElementNS(XHTML_NS, "br"));

    // Buttons
    if (!note.readonly) {
	let adiv = document.createElementNS(XHTML_NS, "div");
	adiv.setAttribute("class", "right");
	div.appendChild(adiv);

	let link = document.createElementNS(XHTML_NS, "a");
	adiv.appendChild(link);
	link.setAttribute("href", "#");
	link.setAttribute("class", "button");
	link.addEventListener("click", {
	    nid: note.id,
	    handleEvent: function(e) { delete_note(this.nid); }
	}, false);
	text = sharedbundle.getString("delete");
	link.appendChild(document.createTextNode("[" + text + "]"));

	adiv.appendChild(document.createTextNode("  "));	// NBSP

	link = document.createElementNS(XHTML_NS, "a");
	adiv.appendChild(link);
	link.setAttribute("href", "#");
	link.setAttribute("class", "button");
	link.addEventListener("click", {
	    nid: nid,
	    opendiv: opendiv,
	    textbox: tb,
	    stylebox: cb,
	    handleEvent: function(e) {
		update_note(this.nid, this.opendiv, this.textbox,
			    this.stylebox.checked);
	    }
	}, false);
	text = sharedbundle.getString("update");
	link.appendChild(document.createTextNode("[" + text + "]"));
    }
}



var g_nm = null;
function load_notes() {
    g_nm = Components.classes["@lucidor.org/notemanagerService;1"].createInstance().wrappedJSObject;
    g_nm.open_new();

    let fp = Components.classes["@mozilla.org/filepicker;1"].
	createInstance(Components.interfaces.nsIFilePicker);
    let bundle = document.getElementById("lucifox-strings");
    let text = bundle.getString("import_annotations");
    fp.init(window, text, Components.interfaces.nsIFilePicker.modeOpen);

    fp.appendFilters(fp.filterAll | fp.filterText);

    let callback = {
	done: function(result) {
	    if (result == fp.returnOK) {
		let db = check_nook(fp.file);
		if (db) {
		    do_load_notes_nook(g_nm, db);
		    return;
		}

		NetUtil.asyncFetch(fp.file, function(inputStream, status) {
		    if (!Components.isSuccessCode(status)) {
			// Handle error!
			return;
		    }
		    let data = NetUtil.readInputStreamToString(inputStream, inputStream.available(), { charset: "UTF-8" });
		    let xml = check_ade(data);
		    if (xml) {
			do_load_notes_ade(g_nm, xml);
			return;
		    }
		    do_load_notes(g_nm, data);
		});
	    }
	}
    };

    if (fp.open != undefined)
	fp.open(callback);
    else
	callback.done(fp.show());
}

function force_to_current(note) {
    if (!note.publication)
	note.publication = {};
    note.publication.sha1 = g_book.sha1;
    note.publication.root = g_book.rootfile;
    note.publication.rootfile = g_book.rootfiles[g_book.rootfile];
    if (g_book.opf.uid)
	note.publication.uid = g_book.opf.uid;
    if (g_book.opf.metadata.meta.modified)
	note.publication.modified = g_book.opf.metadata.meta.modified;

    if (!note.publication.info)
	note.publication.info = {};
    if (g_book.opf.metadata.title)
	note.publication.info.title = g_book.opf.metadata.title;
    if (g_book.opf.metadata.author)
	note.publication.info.creator = g_book.opf.metadata.author.name;
    else if (g_book.opf.metadata.creator.length > 0)
	note.publication.info.creator = g_book.opf.metadata.creator[0].name;
    if (g_book.opf.metadata.publisher)
	note.publication.info.publisher = g_book.opf.metadata.publisher;
}

function note_check(note) {
    if (note.publication == undefined)
	note.publication = [];
    if (note.publication.sha1 == undefined)
	note.publication.sha1 = null;
    if (note.publication.root == undefined)
	note.publication.root = 0;
    if (note.publication.rootfile == undefined)
	note.publication.rootfile = null;
    if (note.publication.uid == undefined)
	note.publication.uid = null;
    if (note.publication.modified == undefined)
	note.publication.modified = null;
    if (note.publication.info == undefined)
	note.publication.info = null;

    if (note.uid == undefined)
	note.uid = null;
    if (note.annotator == undefined)
	note.annotator = null;
    if (note.source == undefined)
	note.source = null;
    if (note.targets == undefined)
	note.targets = [];
    for (let target in loop(note.targets)) {
	if (target.part == undefined)
	    target.part = null;
	if (target.endpart == undefined)
	    target.endpart = null;
	if (target.fixptr == undefined)
	    target.fixptr = null;
	if (target.epubcfi == undefined)
	    target.epubcfi = null;
	if (target.text == undefined)
	    target.text = null;
	if (target.endtext == undefined)
	    target.endtext = null;
	if (target.pretext == undefined)
	    target.pretext = null;
	if (target.posttext == undefined)
	    target.posttext = null;
    }
    if (note.content == undefined)
	note.content = null;
    if (note.type == undefined)
	note.type = null;
    if (note.color == undefined)
	note.color = null;
    if (note.style == undefined)
	note.style = null;
    if (note.extra == undefined)
	note.extra = null;
    if (note.position == undefined)
	note.position = null;
}

function done_loading(num) {
    let text = "Loaded " + num + " annotations.";	// TRANSLATE-ME
    let nb = document.getElementById("notification_box");
    let no = nb.appendNotification(text, "imp", null,
				   nb.PRIORITY_INFO_HIGH, [], null);
    autohide(no, true);

    list_notes();
}

function do_load_notes(nm, data) {
    let inp;
    try {
	inp = JSON.parse(data);
    } catch (e) {
	return null;
    }

    let num = 0;

    for (let entry in loop(inp.annotations)) {
	for (let note in loop(entry.annotations)) {
	    if (!note.publication)
		note.publication = {};
	    note.publication.sha1 = entry.publication.sha1;
	    note.publication.root = entry.publication.root;
	    note.publication.rootfile = entry.publication.rootfile;
	    note.publication.uid = entry.publication.uid;
	    note.publication.modified = entry.publication.modified;
	    note.publication.info = {};
	    if (entry.publication.title)
		note.publication.info.title = entry.publication.title;
	    if (entry.publication.creator)
		note.publication.info.creator = entry.publication.creator;
	    if (entry.publication.publisher)
		note.publication.info.publisher = entry.publication.publisher;

	    //note.source = "imported";
	    note_check(note);
	    nm.add(note);
	    num += 1;
	}
    }

    done_loading(num);
    return 0;

}

function check_ade(data) {
    debug("check_ade");
    let parser = Components.classes["@mozilla.org/xmlextras/domparser;1"].
	getService(Components.interfaces.nsIDOMParser);
    let doc = parser.parseFromString(data, "text/xml");
    let aset = doc.documentElement;

    if (!(aset.localName == "annotationSet" && aset.namespaceURI == ADE_NS))
	return null;

    return doc;
}

function do_load_notes_ade(nm, doc) {
    debug("do_load_notes_ade");

    let aset = doc.documentElement;

    // publication
    let publication = null;
    let info = { readonly: true };
    for (let elem in loop(aset.childNodes)) {
	if (!(elem.nodeType == elem.ELEMENT_NODE &&
	      elem.localName == "publication" && elem.namespaceURI == ADE_NS))
	    continue;
	debug(elem.localName);
	for (let cn in loop(elem.childNodes)) {
	    if (!(cn.nodeType == cn.ELEMENT_NODE && cn.namespaceURI == DC_NS))
		continue;
	    debug(cn.localName + ":" + get_text(cn));
	    if (cn.localName == "identifier") {
		publication = get_text(cn);
	    } else if (cn.localName == "title") {
		info.title = get_text(cn);
	    } else if (cn.localName == "creator") {
		info.author = get_text(cn);
	    } else if (cn.localName == "publisher") {
		info.publisher = get_text(cn);
	    }
	}
    }

    // annotations
    let num = 0;
    for (let elem in loop(aset.childNodes)) {
	if (!(elem.nodeType == elem.ELEMENT_NODE &&
	      elem.localName == "annotation" && elem.namespaceURI == ADE_NS))
	    continue;
	debug(elem.localName);

	let note = { pid: publication,
		     root: 0,	// FIXME
		     rootfile: null,
		     info: info,
		     source: "imported",
		     targets: [] };

	for (let ch in loop(elem.childNodes)) {
	    if (ch.nodeType != ch.ELEMENT_NODE)
		continue;
	    debug(ch.localName + ":" + get_text(ch));

	    if (ch.localName == "identifier" && ch.namespaceURI == DC_NS) {
		note.uid = get_text(ch);
	    } else if (ch.localName == "date" && ch.namespaceURI == DC_NS) {
		note.created = get_text(ch);
	    } else if (ch.localName == "creator" && ch.namespaceURI == DC_NS) {
		//
	    } else if (ch.localName == "target" && ch.namespaceURI == ADE_NS) {
		for (let cn in loop(ch.childNodes)) {
		    if (cn.nodeType != ch.ELEMENT_NODE ||
			cn.localName != "fragment")
			continue;
		    debug(cn.localName + ":" + get_text(cn));
		    if (!(cn.hasAttribute("start") && cn.hasAttribute("end")))
			continue;

		    // FIXME check for text element
		    let text = null;

		    let start = cn.getAttribute("start").split("#");
		    let end = cn.getAttribute("end").split("#");
		    if (start.length == 2 && end.length == 2) {
			let endpart = null;
			if (end[0] != start[0])
			    endpart = end[0];
			note.targets.push({ part: start[0],
					    endpart: endpart,
					    fixptr: null,
					    pstart: start[1],
					    pend: end[1],
					    text: text });
		    }
		}
	    } else if (ch.localName == "content" && ch.namespaceURI == ADE_NS) {
		for (let cn in loop(ch.childNodes)) {
		    if (cn.nodeType != cn.ELEMENT_NODE)
			continue;
		    debug(cn.localName + ":" + get_text(cn));
		    if (cn.localName == "date" && cn.namespaceURI == DC_NS) {
			note.modified = get_text(cn);
		    } else if (cn.localName == "text" &&
			       cn.namespaceURI == ADE_NS) {
			note.content = get_text(cn);
			note.type = "text";
		    }
		}
	    }
	}

	note_check(note);
	nm.add(note);
	num += 1;
    }

    done_loading(num);
    return 0;
}

function get_text(node) {
    let text = node.textContent;

    // Strip beginning and ending whitespace
    let i;
    for (i = 0; i < text.length; i++)
	if (!is_ws[text[i]])
	    break;
    let j;
    for (j = text.length - 1; j > i; j--)
	if (!is_ws[text[j]])
	    break;
    return text.substring(i, j + 1);
}

var is_ws = [];
is_ws[" "] = true;
is_ws["\n"] = true;
is_ws["\t"] = true;
is_ws["\r"] = true;
is_ws[String.fromCharCode(160)] = true;	// &nbsp;

function check_nook(filep) {
    let ss = Components.classes["@mozilla.org/storage/service;1"].
	getService(Components.interfaces.mozIStorageService);
    try {
	let db = ss.openDatabase(filep);
	return db;
    } catch (e) {
    }
    return null;
}

function do_load_notes_nook(nm, db) {
    let statement = db.createStatement("SELECT ean, titles, contributor_sort, local_filename, created, modified FROM bn_client_products ORDER BY id ASC");
    let pubs = [];
    try {
	while (statement.executeStep()) {
	    let pub = {};
	    pub.ean = statement.row.ean;
	    pub.titles = statement.row.titles;
	    pub.contributor_sort = statement.row.contributor_sort;
	    pub.local_filename = statement.row.local_filename;
	    pub.created = statement.row.created;
	    pub.modified = statement.row.modified;

	    do_load_notes_nook2(nm, db, pub);
	    //pubs.push(pub);
	}
    } finally {
	statement.reset();
    }
    //return pubs;
}

function do_load_notes_nook2(nm, db, pub) {
    let num = 0;
    let statement = db.createStatement("SELECT content, color, created, modified, position_rmsdk, position_rmsdk_end FROM bn_client_text_annotations WHERE product_ean=:x_product_ean ORDER BY id ASC");
    try {
	statement.params.x_product_ean = pub.ean;
	while (statement.executeStep()) {
	    let note = { uid: pub.ean,	/* not really the EPUB uid
					   must be changed when imported */
			 info: { readonly: true },
			 source: "imported",
			 targets: [] };
	    if (pub.titles)
		note.info.title = pub.titles;	// parse JSON?
	    if (pub.contributor_sort)
		note.info.author = pub.contributor_sort;
	    if (pub.local_filename)
		note.info.filename = pub.local_filename;

	    if (statement.row.content) {
		note.content = statement.row.content;
		note.type = "text";
	    } else
		note.style = "highlight";
	    note.color = statement.row.color.toString(16);
	    note.created = (new Date(statement.row.created * 1000)).toISOString();
	    note.modified = (new Date(statement.row.modified * 1000)).toISOString();

	    let start = statement.row.position_rmsdk.split("#");
	    let end = statement.row.position_rmsdk_end.split("#");
	    if (start.length == 2 && end.length == 2) {
		let endpart = null;
		if (end[0] != start[0])
		    endpart = end[0];
		note.targets.push({ part: start[0],
				    endpart: endpart,
				    fixptr: null,
				    pstart: start[1],
				    pend: end[1],
				    text: null });
	    }

	    note_check(note);
	    nm.add(note);
	    num += 1;
	}
    } finally {
	statement.reset();
    }
    db.asyncClose();

    done_loading(num);
    return 0;
}

function import_notes(nm, force, subset) {

}
