/*
 *  Lucidor - e-book reader
 *  Copyright © 2009-2013  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("chrome://lucidor/content/utils.jsm");

const XHTML_NS = "http://www.w3.org/1999/xhtml";
const XUL_NS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";

var is_extension = false;

function init() {
    let prefs = Components.classes["@mozilla.org/preferences-service;1"].
	getService(Components.interfaces.nsIPrefService).
	getBranch("extensions.lucifox.");

    enable_debug(prefs.getBoolPref("debug.enable"));

    let appinfo = Components.classes["@mozilla.org/xre/app-info;1"].
	getService(Components.interfaces.nsIXULAppInfo);
    if (appinfo.ID == "Lucidor@lucidor.org")
	is_extension = false;
    else
	is_extension = true;

    let ioService = Components.classes["@mozilla.org/network/io-service;1"].
	getService(Components.interfaces.nsIIOService);
    let uri = ioService.newURI(document.documentURI, "UTF-8", null);
    uri.QueryInterface(Components.interfaces.nsIURL);
    debug("QUERY: " + uri.query);

    let params = uri.query.split("&");
    let uris = [];
    let vals = {};
    for (let par in loop(params)) {
	let pair = par.split("=");
	if (pair.length != 2)
	    continue;
	pair[1] = decodeURIComponent(pair[1]);
	debug("PARAM: " + pair[0] + " = " + pair[1]);
	if (pair[0] == "uri") {
	    uris.push(pair[1]);
	} else	// author, title
	    vals[pair[0]] = pair[1];
    }

    for (let uri in loop(uris)) {
	let j = uri.indexOf(':');
	if (j > 0)
	    append_url(uri.substring(0, j), uri.substring(j + 1));
    }

    if (vals["title"]) {
	let tb = document.getElementById("title_textbox");
	tb.value = vals["title"];
    }

    update_sense();
}

function update_sense() {
    let list = document.getElementById("urllist");
    let button = document.getElementById("remove_feed");
    if (list.selectedIndex >= 0)
	button.disabled = false;
    else
	button.disabled = true;

    button = document.getElementById("open_new_book");
    if (list.getRowCount() > 0)
	button.disabled = false;
    else
	button.disabled = true;
}

function add_doc() {
    let tb = document.getElementById("feed_url_textbox");
    let location = tb.value;

    let ioService = Components.classes["@mozilla.org/network/io-service;1"].
	getService(Components.interfaces.nsIIOService);
    try {
	ioService.newURI(location, "UTF-8", null);
    } catch (e) {
	try {
	    location = 'http://' + location;
	    ioService.newURI(location, "UTF-8", null);
	} catch (e) {
	    //show_error();
	    debug("Incorrect URL");
	    return;
	}
    }

    let menu = document.getElementById("type_menu");

    append_url(menu.value, location);
    tb.value = "";
    update_sense();
}

function append_url(type, location) {
    let list = document.getElementById("urllist");
    list.appendItem('(' + type + ') ' + location, type + ':' + location);
}

function remove_url() {
    let list = document.getElementById("urllist");
    if (list.selectedIndex >= 0)
	list.removeItemAt(list.selectedIndex);

    update_sense();
}

function clear_fields() {
    let list = document.getElementById("urllist");
    while (list.getRowCount() > 0)
	list.removeItemAt(0);

    let tb = document.getElementById("title_textbox");
    tb.value = "";
    tb = document.getElementById("feed_url_textbox");
    tb.value = "";

    update_sense();
}

function generate_new_book() {
    debug("generate_new_book");
    let list = document.getElementById("urllist");
    let rows = list.getRowCount();
    let ls = new Array(rows);
    for (let i = 0; i < rows; i++)
	ls[i] = { uri: list.getItemAtIndex(i).value };
    let title = null;
    let tb = document.getElementById("title_textbox");
    if (tb.value)
	title = tb.value;
    let feed = { items: ls, version: 1 };
    generate2(feed, title);
}

function generate2(feed, title) {
    debug("generate2: " + feed.items + " : " + title);
    let location = "x-lucidor-convert://feed/?";

    for (let item in loop(feed.items))
	location += "uri=" + encodeURIComponent(item.uri) + "&";

    if (title)
	location += "title=" + encodeURIComponent(title);

    let uri = "chrome://lucidor/content/viewepub.xul?uri=" +
	encodeURIComponent(location);

    debug("location: " + location);
    if (is_extension) {
	let mainWindow = window.QueryInterface(Components.interfaces.nsIInterfaceRequestor)
            .getInterface(Components.interfaces.nsIWebNavigation)
            .QueryInterface(Components.interfaces.nsIDocShellTreeItem)
            .rootTreeItem
            .QueryInterface(Components.interfaces.nsIInterfaceRequestor)
            .getInterface(Components.interfaces.nsIDOMWindow);
	mainWindow.gBrowser.addTab(uri);
    } else {
	let main_control = Components.classes["@lucidor.org/mainControlService;1"].
	    getService().wrappedJSObject;
	main_control.add_tab(uri);
    }
}
