/*
 *  Lucidor - e-book reader
 *  Copyright © 2009-2016  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("chrome://lucidor/content/utils.jsm");

function install_ff_theme() {
    try {
	do_install_ff_theme();
    } catch (e) {
	debug("E:" + e);
    }
}

function do_install_ff_theme() {
    let fp = Components.classes["@mozilla.org/filepicker;1"].
	createInstance(Components.interfaces.nsIFilePicker);
    //let strbundle = document.getElementById("lucifox-strings");
    //let text = strbundle.getString("open_file");
    let text = "Install Firefox Theme";	// TRANSLATE-ME
    fp.init(window, text, Components.interfaces.nsIFilePicker.modeOpen);
    //let sharedbundle = document.getElementById("lucifox-shared_strings");
    //text = sharedbundle.getString("themes");
    //fp.appendFilter(text, "*");
    text = "Themes";	// TRANSLATE-ME
    fp.appendFilter(text, "*.jar ; *.xpi");

    let callback = {
	done: function(result) {
	    if (result != fp.returnOK)
		return;

	    let zipreader = Components.classes["@mozilla.org/libjar/zip-reader;1"].
		createInstance(Components.interfaces.nsIZipReader);
	    zipreader.open(fp.file);

	    if (!zipreader.hasEntry("install.rdf")) {
		debug("No install.rdf in theme file");
		return;
	    }
	    zipreader.close();

	    let ios = Components.classes["@mozilla.org/network/io-service;1"].
		getService(Components.interfaces.nsIIOService);
	    let fileuri = ios.newFileURI(fp.file);

	    function my_do_2(event) {
		do_install_ff_theme_2(event, fp.file);
	    }
	    let xhr = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"].
		createInstance(Components.interfaces.nsIXMLHttpRequest);
	    let jurl = "jar:" + fileuri.spec + "!/install.rdf";
	    xhr.onload = my_do_2;
	    xhr.open("GET", jurl);
	    xhr.send();
	}
    };

    if (fp.open != undefined)
	fp.open(callback);
    else
	callback.done(fp.show());
}

function do_install_ff_theme_2(event, fil) {
    let oXHR = event.target;
    let doc = oXHR.responseXML;
    if (!doc) {
	debug("incorrect theme file");
	return;
    }

    let EM_NS = "http://www.mozilla.org/2004/em-rdf#";
    let RDF_NS = "http://www.w3.org/1999/02/22-rdf-syntax-ns#";

    // FIXME Should check that typ.textContent == "2"
    // but almost all themes on amo have incorrect em:type
    // Also note that sometimes em:type is a property on Description
    let typ = doc.getElementsByTagNameNS(EM_NS, "type").item(0);
    //if (typ.textContent != "2") {
    if (typ && typ.textContent != "2" && typ.textContent != "4") {
	debug("Not a theme: " + typ.textContent);
	return;
    }

    let app = doc.getElementsByTagNameNS(EM_NS, "targetApplication").item(0);
    if (!app) {
	debug("No elem");
	return;
    }
    let node = doc.createElementNS(EM_NS, "targetApplication");
    app.parentNode.appendChild(node);
    let node2 = doc.createElementNS(RDF_NS, "Description");
    node.appendChild(node2);

    let node3 = doc.createElementNS(EM_NS, "id");
    let text = doc.createTextNode("Lucidor@lucidor.org");
    node3.appendChild(text);
    node2.appendChild(node3);

    node3 = doc.createElementNS(EM_NS, "minVersion");
    text = doc.createTextNode("0.7");
    node3.appendChild(text);
    node2.appendChild(node3);

    node3 = doc.createElementNS(EM_NS, "maxVersion");
    text = doc.createTextNode("1.*");
    node3.appendChild(text);
    node2.appendChild(node3);

    let bookman = Components.classes["@lucidor.org/bookmanagerService;1"].
	getService().wrappedJSObject;
    let outfil = bookman.getTempdir();
    outfil.append("lucidor.xpi");
    outfil.createUnique(Components.interfaces.nsIFile.NORMAL_FILE_TYPE,
			parseInt("0600", 8));
    outfil.remove(false);
    fil.copyTo(outfil.parent, outfil.leafName);

    let domser = Components.classes["@mozilla.org/xmlextras/xmlserializer;1"].
        createInstance(Components.interfaces.nsIDOMSerializer);
    let install_rdf = domser.serializeToString(doc);
    //dump(install_rdf + "\n");
    let sis = Components.classes["@mozilla.org/io/string-input-stream;1"].
        createInstance(Components.interfaces.nsIStringInputStream);
    sis.setData(install_rdf, install_rdf.length);

    let zipwriter = Components.classes["@mozilla.org/zipwriter;1"].
	createInstance(Components.interfaces.nsIZipWriter);
    zipwriter.open(outfil, 0x04);

    zipwriter.removeEntry("install.rdf", false);
    zipwriter.addEntryStream("install.rdf", null, 6, sis, false);
    zipwriter.close();

    let browser = document.getElementById("hidden_browser");
    let fph = Components.classes["@mozilla.org/network/protocol;1?name=file"].getService(Components.interfaces.nsIFileProtocolHandler);
    let uri = fph.getURLSpecFromFile(outfil);
    browser.loadURI(uri, null, null);
}
