/*
 *  Lucidor - e-book reader
 *  Copyright © 2008-2016  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("chrome://lucidor/content/utils.jsm");

var XHTML_NS = "http://www.w3.org/1999/xhtml";	// FIXME should be const

var bookcase;

function init() {
    bookcase = Components.classes["@lucidor.org/bookcaseSqlService;1"].
	getService().wrappedJSObject;
    bookcase.open();

    let uri = document.documentURI;
    if (uri.indexOf("startup") >= 0) {
	let frame = document.getElementById("legalinfo");
	frame.hidden = false;
    }

    let frame = document.getElementById("recent_box");
    let i = 0;
    if (uri.indexOf("newuser") >= 0) {
	let header = document.getElementById("lastread");
	header.style.display = "none";
	header = document.getElementById("toread");
	header.style.display = "block";

	let elem = document.getElementById("newuser");
	elem.style.display = "block";
	i = 1;
    } else {
	let where = "WHERE date_read IS NOT NULL ORDER BY date_read DESC";
	let frame_first = document.getElementById("first_box");
	let frame_more = document.getElementById("more_box");
	for (let bookinfo in bookcase.search(where, "where")) {
	    if (i++ == 10)
		break;
	    let div = document.createElementNS(XHTML_NS, "div");
	    if (i <= 4)
		frame_first.appendChild(div);
	    else
		frame_more.appendChild(div);
	    div.setAttribute("class", "link");
	    let link = document.createElementNS(XHTML_NS, "a");
	    div.appendChild(link);

	    link.setAttribute("href", "");
	    link.setAttribute("onclick", "open_book_location(event, '" +
			      bookinfo.location + "', false);");

	    div = document.createElementNS(XHTML_NS, "div");
	    link.appendChild(div);
	    let node = document.createTextNode(bookinfo.title);
	    div.appendChild(node);

	    div = document.createElementNS(XHTML_NS, "div");
	    div.setAttribute("class", "author");
	    link.appendChild(div);
	    node = document.createTextNode(bookinfo.author);
	    div.appendChild(node);
	}
    }

    if (i == 0) {
	let header = document.getElementById("lastread");
	header.style.display = "none";
	header = document.getElementById("toread");
	header.style.display = "block";

	let fph = Components.classes["@mozilla.org/network/protocol;1?name=file"].
	    getService(Components.interfaces.nsIFileProtocolHandler);
	let entries = null;
	try {
	    let ioService = Components.classes["@mozilla.org/network/io-service;1"].
		getService(Components.interfaces.nsIIOService);
	    let uri = ioService.newURI("resource://lucidor/samples/", null, null);
	    let location = Components.classes["@mozilla.org/network/protocol;1?name=resource"].
		getService(Components.interfaces.nsIResProtocolHandler).
		resolveURI(uri);
	    let dir = fph.getFileFromURLSpec(location);
	    entries = dir.directoryEntries;
	} catch (e) {
	}
	for (; entries && entries.hasMoreElements() && i < 8; i++) {
	    let entry = entries.getNext().QueryInterface(Components.interfaces.nsIFile);
	    let loc = fph.getURLSpecFromFile(entry);
	    if (loc.substring(loc.length - 5) != ".epub")
		continue;

	    let book = Components.classes["@lucidor.org/epubDocument;1"].
		createInstance().wrappedJSObject;
	    if (i < 4)
		book.init(entry, loc, addBookFirst);
	    else
	    	book.init(entry, loc, addBookMore);
	}
    }
    if (i > 0)
	frame.hidden = false;
    if (i > 4)
	document.getElementById("more_link").setAttribute("style", "display: block");
}

function open_book_location(event, location, newtab) {
    event.preventDefault();

    let main_control = Components.classes["@lucidor.org/mainControlService;1"].
	getService().wrappedJSObject;
    if (newtab)
	main_control.add_epub_tab(location, 0);
    else
	main_control.open_epub_location(location);
}

function more(event) {
    event.preventDefault();
    let box = document.getElementById("more_box");
    box.hidden = !box.hidden;
    document.getElementById("more_link").setAttribute("style", "display: none");
}

function addBookFirst(book, uri) {
    addBook(book, uri, "first_box");
}

function addBookMore(book, uri) {
    addBook(book, uri, "more_box");
}

function addBook(book, uri, frameid) {
    let frame = document.getElementById(frameid);

    let div = document.createElementNS(XHTML_NS, "div");
    frame.appendChild(div);
    div.setAttribute("class", "link");
    let link = document.createElementNS(XHTML_NS, "a");
    div.appendChild(link);

    link.setAttribute("href", "");
    link.setAttribute("onclick",
		      "open_book_location(event, '" + uri + "', false);");

    div = document.createElementNS(XHTML_NS, "div");
    link.appendChild(div);
    let node = document.createTextNode(book.opf.metadata.title);
    div.appendChild(node);

    div = document.createElementNS(XHTML_NS, "div");
    div.setAttribute("class", "author");
    link.appendChild(div);
    node = document.createTextNode(book.opf.metadata.author.name);
    div.appendChild(node);
}

function open_web_location(location) {
    let main_control = Components.classes["@lucidor.org/mainControlService;1"].
	getService().wrappedJSObject;
    main_control.add_web_tab(location);
}
