/*
 *  Lucidor - e-book reader
 *  Copyright © 2008-2019  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("chrome://lucidor/content/utils.jsm");

const XUL_NS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";

var progressListener = {
    onLocationChange: function(webProgress, request, location) {
    },

    onProgressChange: function(webProgress, request, curSelfProgress,
			       maxSelfProgress, curTotalProgress,
			       maxTotalProgress) {
    },

    onSecurityChange: function(webProgress, request, state) {
    },

    onStateChange: function(webProgress, request, stateFlags, status) {
	const CIWPL = Components.interfaces.nsIWebProgressListener;

	if ((stateFlags & CIWPL.STATE_STOP) &&
	    (stateFlags & CIWPL.STATE_IS_DOCUMENT)) {
	    update_tab_title();
	}
    },

    onStatusChange: function(webProgress, request, status, message) {
    },

    QueryInterface: function(aIID) {
	if (aIID.equals(Components.interfaces.nsIWebProgressListener) ||
	    aIID.equals(Components.interfaces.nsISupportsWeakReference))
	    return this;
	throw Components.results.NS_NOINTERFACE;
    }
};

var historyListener = {
    OnHistoryGoBack: function(backURI) {
	update_bar();
	return true;
    },

    OnHistoryGoForward: function(forwardURI) {
	update_bar();
	return true;
    },

    OnHistoryGotoIndex: function(index, gotoURI) {
	update_bar();
	return true;
    },

    OnHistoryNewEntry: function(newURI) {
	update_bar();
	return true;
    },

    OnHistoryPurge: function(numEntries) {
	update_bar();
	return true;
    },

    OnHistoryReload: function(reloadURI, reloadFlags) {
	update_bar();
	return true;
    },

    QueryInterface: function(aIID) {
	if (aIID.equals(Components.interfaces.nsISHistoryListener) ||
	    aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
	    aIID.equals(Components.interfaces.nsISupports))
	    return this;
	throw Components.results.NS_NOINTERFACE;
    }
};

function init() {
    let prefs = Components.classes["@mozilla.org/preferences-service;1"].
	getService(Components.interfaces.nsIPrefService).
	getBranch("extensions.lucifox.");

    enable_debug(prefs.getBoolPref("debug.enable"));

    /*
    let chromeRegistry = Components.classes["@mozilla.org/chrome/chrome-registry;1"].getService(Components.interfaces.nsIXULChromeRegistry);
    let currentLocale = chromeRegistry.getSelectedLocale("lucidor");
    debug("Locale: " + currentLocale);
    */

    // Must keep reference to contentListener when using this
    //let uriLoader = Components.classes["@mozilla.org/uriloader;1"].
    //	getService(Components.interfaces.nsIURILoader);
    //uriLoader.registerContentListener(contentListener);

    // Hide Mac OS X application menu from other OSes
    let appinfo = Components.classes["@mozilla.org/xre/app-info;1"].
	getService(Components.interfaces.nsIXULRuntime);
    if (appinfo.OS != "Darwin" && appinfo.OS != "MacOSX") {
	let mb = document.getElementById("osx-app-menu");
	mb.hidden = true;
	mb = document.getElementById("full_screen_button");
	mb.hidden = false;
    }

    let control = Components.classes["@lucidor.org/mainControlService;1"].
	getService().wrappedJSObject;
    control.init(window);

    // Register default stylesheets
    let sss = Components.classes["@mozilla.org/content/style-sheet-service;1"].
	getService(Components.interfaces.nsIStyleSheetService);
    let ios = Components.classes["@mozilla.org/network/io-service;1"].
	getService(Components.interfaces.nsIIOService);
    let uri = ios.newURI("chrome://lucidor/content/dtbook.css", null, null);
    if (!sss.sheetRegistered(uri, sss.AGENT_SHEET))
	sss.loadAndRegisterSheet(uri, sss.AGENT_SHEET);
    uri = ios.newURI("chrome://lucidor/content/xhtml.css", null, null);
    if (!sss.sheetRegistered(uri, sss.AGENT_SHEET))
	sss.loadAndRegisterSheet(uri, sss.AGENT_SHEET);
    uri = ios.newURI("chrome://lucidor/content/ordbrand.css", null, null);
    if (!sss.sheetRegistered(uri, sss.AGENT_SHEET))
	sss.loadAndRegisterSheet(uri, sss.AGENT_SHEET);

    // Initialize previewCreator service
    let previews = Components.classes["@lucidor.org/previewCreatorService;1"].
	getService().wrappedJSObject;
    let pbr = document.getElementById("preview_browser");
    previews.init(pbr);

    // Initialize languageMapper service
    let language_mapper = Components.classes["@lucidor.org/languageMapperService;1"].getService().wrappedJSObject;
    language_mapper.init();
    //language_mapper.debug();

    // Check and maybe update preferences
    if (!prefs.prefHasUserValue("prefsUpdated") &&
       !prefs.prefHasUserValue("collection.order")) {
	let branch = "lucidor.";
	let prefs2 = Components.classes["@mozilla.org/preferences-service;1"].
	    getService(Components.interfaces.nsIPrefService).
	    getBranch(branch);
	let chds = prefs2.getChildList("");
	for (let chd in loop(chds)) {
	    let typ = prefs2.getPrefType(chd);
	    if (typ == 32) {
		prefs.setCharPref(chd, prefs2.getCharPref(chd));
	    } else if (typ == 64) {
		prefs.setIntPref(chd, prefs2.getIntPref(chd));
	    } else if (typ == 128) {
		prefs.setBoolPref(chd, prefs2.getBoolPref(chd));
	    }
	}
	prefs.setBoolPref("prefsUpdated", true);
    }

    let new_user = false;
    let ver = null;
    if (!prefs.prefHasUserValue("lastVersion")) {
	new_user = true;
	ver = parse_version("0.0.0");
    } else
	ver = parse_version(prefs.getCharPref("lastVersion"));

    if (is_updated(ver, [0, 9, 8])) {
	let cman = Components.classes["@lucidor.org/collectionManagerService;1"].
	    getService().wrappedJSObject;
	cman.maybeAddCollection({ title: "Bokselskap.no",
				  type: "opds",
				  url: "http://www.bokselskap.no/wp-content/themes/bokselskap/tekster/opds/root.xml",
				  icon: null });
    }

    prefs.setCharPref("lastVersion", "0.9.15");

    // Make sure lucidor.bookcase.folder is set
    let bookcase = Components.classes["@lucidor.org/bookcaseSqlService;1"].
	getService().wrappedJSObject;
    bookcase.get_books_directory();

    // Handle command line arguments
    let clh = Components.classes["@lucidor.org/cmdlineHandler;1"].
	getService().wrappedJSObject;
    if (clh.defaultURIS.length > 0)
	for (let duri in loop(clh.defaultURIS))
	    add_epub_tab(duri.spec, 0);
    else {
	let url = "chrome://lucidor/content/homepage.xul?startup";
	if (new_user)
	    url += ";newuser";
	add_tab(url);
    }
}

function toggle_full_screen() {
    let mb = document.getElementById("full_screen_button");
    if (mb.getAttribute("checked"))
	setTimeout('window.fullScreen = true;', 1);
    else
	setTimeout('window.fullScreen = false;', 1);
}

function open_preferences() {
    window.openDialog("chrome://lucidor/content/preferences.xul",
		      "Preferences",
		      "chrome,titlebar,toolbar,close,resizable,scrollbars,centerscreen,dialog=yes");
}

function open_addons() {
    //window.open("chrome://mozapps/content/extensions/extensions.xul?type=extensions",
    //		"Add-ons", "chrome,titlebar,toolbar,centerscreen");
    return add_tab("chrome://mozapps/content/extensions/extensions.xul?type=extensions");
}

function add_epub_tab(location, root) {
    return add_tab("chrome://lucidor/content/viewepub.xul?uri=" +
		   encodeURIComponent(location) + "&root=" + root);
}

function add_web_tab(location) {
    return add_tab("chrome://lucidor/content/viewweb.xul?uri=" +
		   encodeURIComponent(location));
}

function add_web_tab_post(location, post) {
    return add_tab("chrome://lucidor/content/viewweb.xul?uri=" +
		   encodeURIComponent(location) + "&post=" +
		   encodeURIComponent(post));
}

function add_tab(location) {
    let tabs = document.getElementById("tabs");
    let tab = document.createElementNS(XUL_NS, "tab");
    tab.setAttribute("label", " ");
    //tab.setAttribute("style", "overflow: hidden; crop: start; width: 8em;");
    tabs.appendChild(tab);

    let tabpanels = document.getElementById("tabpanels");
    let tabpanel = document.createElementNS(XUL_NS, "tabpanel");
    //tabpanel.setAttribute("flex", "1");
    let topbox = document.createElementNS(XUL_NS, "vbox");
    topbox.setAttribute("flex", "1");
    let vbox = document.createElementNS(XUL_NS, "vbox");
    topbox.appendChild(vbox);
    let browser = document.createElementNS(XUL_NS, "browser");
    browser.setAttribute("src", location);
    browser.setAttribute("disablehistory", "true");
    //browser.setAttribute("type", "content");
    browser.setAttribute("flex", "1");
    //tabpanel.appendChild(browser);
    topbox.appendChild(browser);
    tabpanel.appendChild(topbox);
    tabpanels.appendChild(tabpanel);
    tabs.selectedItem = tab;

    browser.addProgressListener(progressListener, Components.interfaces.nsIWebProgress.NOTIFY_STATE_DOCUMENT);
    //browser.sessionHistory.addSHistoryListener(historyListener);

    /*
    let cl = Components.classes["@lucidor.org/myURIContentListener;1"].createInstance(Components.interfaces.nsIURIContentListener);
    //let browser = document.getElementById("book_view");
    let wnd = browser.docShell.QueryInterface(Components.interfaces.nsIInterfaceRequestor).getInterface(Components.interfaces.nsIURIContentListener);
    wnd.parentContentListener = cl;
    */

    let button = document.getElementById("close-tab-button");
    button.disabled = false;

    let box = document.getElementById("scrollbox");
    let sb;
    try {
	sb = box.boxObject.QueryInterface(Components.interfaces.nsIScrollBoxObject);
    } catch (e) {
	sb = box.boxObject;
    }
    let x = tab.boxObject.x - tabs.childNodes.item(0).boxObject.x;	// FIXME
    sb.scrollTo(x, 0);
    return vbox;
}

function close_tab() {
    let tabs = document.getElementById("tabs");
    let index = tabs.selectedIndex;
    if (index == -1)
	return;

    // This is now done with onbeforeunload
    /*
    let browser = get_browser();
    if (browser.contentWindow && browser.contentWindow.warn_unload)
	browser.contentWindow.warn_unload();
    */

    tabs.removeItemAt(index);
    let tabpanels = document.getElementById("tabpanels");
    tabpanels.removeChild(tabpanels.selectedPanel);

    if (index == tabs.childNodes.length)
	index--;
    tabs.selectedIndex = index;
    if (index == -1) {
	let button = document.getElementById("close-tab-button");
	button.disabled = true;
	update_bar();
    }

    let box = document.getElementById("scrollbox");
    let sb;
    try {
	sb = box.boxObject.QueryInterface(Components.interfaces.nsIScrollBoxObject);
    } catch (e) {
	sb = box.boxObject;
    }
    let xp = {};
    sb.getPosition(xp, {});
    sb.scrollTo(xp.value, 0);	// FIXME
}

function overflow() {
    //debug("overflow");
    let box = document.getElementById("scrollbox");
    let sb;
    try {
	sb = box.boxObject.QueryInterface(Components.interfaces.nsIScrollBoxObject);
    } catch (e) {
	sb = box.boxObject;
    }
    let wp = {};
    sb.getScrolledSize(wp, {});
    //debug("W: " + wp.value + " : " + box.boxObject.width);
    // NOTE Necessary because with some themes there are false overflow events
    if (wp.value <= box.boxObject.width)
	return underflow();

    let button = document.getElementById("tab_left");
    button.hidden = false;
    button = document.getElementById("tab_right");
    button.hidden = false;
    return null;
}

function underflow() {
    //debug("underflow");
    let button = document.getElementById("tab_left");
    button.hidden = true;
    button = document.getElementById("tab_right");
    button.hidden = true;

    let box = document.getElementById("scrollbox");
    let sb;
    try {
	sb = box.boxObject.QueryInterface(Components.interfaces.nsIScrollBoxObject);
    } catch (e) {
	sb = box.boxObject;
    }
    sb.scrollTo(0, 0);
}

function scroll_left() {
    let tabs = document.getElementById("tabs");
    let box = document.getElementById("scrollbox");
    let sb;
    try {
	sb = box.boxObject.QueryInterface(Components.interfaces.nsIScrollBoxObject);
    } catch (e) {
	sb = box.boxObject;
    }
    let xp = {};
    let yp = {};
    sb.getPosition(xp, yp);
    let off = tabs.childNodes.item(0).boxObject.x;
    let j = 0;
    for (; j < tabs.childNodes.length; j++) {
	let p = tabs.childNodes.item(j).boxObject.x - off;
	//debug(";; " + tabs.childNodes.item(j).boxObject.x + " " + p + " " + xp.value);
	if (xp.value < p + 5)
	    break;
    }
    j = Math.max(j - 1, 0);
    let x = tabs.childNodes.item(j).boxObject.x - off;
    sb.scrollTo(x, yp.value);

    if (j == 0 || x <= 0) {
	let button = document.getElementById("tab_left");
	button.disabled = true;
    }
    let button = document.getElementById("tab_right");
    button.disabled = false;
}

function scroll_right() {
    let tabs = document.getElementById("tabs");
    let box = document.getElementById("scrollbox");
    let sb;
    try {
	sb = box.boxObject.QueryInterface(Components.interfaces.nsIScrollBoxObject);
    } catch (e) {
	sb = box.boxObject;
    }
    let xp = {};
    sb.getPosition(xp, {});
    let off = tabs.childNodes.item(0).boxObject.x;
    let j = 0;
    for (; j < tabs.childNodes.length; j++) {
	let p = tabs.childNodes.item(j).boxObject.x - off;
	//debug(";; " + tabs.childNodes.item(j).boxObject.x + " " + p + " " + xp.value);
	if (xp.value < p + 5)
	    break;
    }
    j = Math.min(j + 1, tabs.childNodes.length - 1);
    let x = tabs.childNodes.item(j).boxObject.x - off;
    sb.scrollTo(x, 0);

    let last = tabs.childNodes.item(tabs.childNodes.length - 1);
    // FIXME
    if (x >= (last.boxObject.x + last.boxObject.width - box.boxObject.width)) {
	let button = document.getElementById("tab_right");
	button.disabled = true;
    }
    let button = document.getElementById("tab_left");
    button.disabled = false;
}

var dragObserver = {
    onDragOver: function(event, flavour, session) {
    },

    onDrop: function(event, transferData, session) {
	debug("DROP: " + transferData.data + " " + transferData.flavour.contentType);
	if (transferData.flavour.contentType == "application/x-moz-file") {
	    let uri = Components.classes["@mozilla.org/network/protocol;1?name=file"].getService(Components.interfaces.nsIFileProtocolHandler).
		getURLSpecFromFile(transferData.data);
	    add_epub_tab(uri, 0);
	} else if (transferData.flavour.contentType == "text/x-moz-url") {
	    let ss = transferData.data.split("\n");
	    add_web_tab(ss[0]);
	}
    },

    getSupportedFlavours: function() {
	let flavours = new FlavourSet();
	flavours.appendFlavour("application/x-moz-file","nsIFile");
	flavours.appendFlavour("text/x-moz-url");
	return flavours;
    }
}

function open_epub_location(location) {
    return open_location("chrome://lucidor/content/viewepub.xul?uri=" +
			 encodeURIComponent(location));
}

function open_location(uri) {
    debug("open_location: " + uri);
    let browser = get_browser();
    if (browser.loadURI)
	browser.loadURI(uri, null, null);
    else
	add_tab(uri);
}

function open_book_from_location() {
    let fp = Components.classes["@mozilla.org/filepicker;1"].
	createInstance(Components.interfaces.nsIFilePicker);
    let strbundle = document.getElementById("lucifox-strings");
    let text = strbundle.getString("open_file");
    fp.init(window, text, Components.interfaces.nsIFilePicker.modeOpen);
    let sharedbundle = document.getElementById("lucifox-shared_strings");
    text = sharedbundle.getString("books");
    fp.appendFilter(text, "*.epub ; *.EPUB");
    //fp.appendFilter(text, "*.epub ; *.EPUB ; *.pdf ; *.PDF");
    text = sharedbundle.getString("all_files");
    fp.appendFilter(text, "*");
    fp.addToRecentDocs = true;

    let callback = {
	done: function(result) {
	    if (result == fp.returnOK)
		open_epub_location(fp.fileURL.spec);
	}
    };

    if (fp.open != undefined)
	fp.open(callback);
    else
	callback.done(fp.show());
}

function save_file() {
    let browser = get_browser();

    if (browser.contentWindow.save_page) {
	browser.contentWindow.save_page();
	return;
    }

    let file;
    if (browser.contentWindow.get_save_file)
	file = browser.contentWindow.get_save_file();

    let fp = Components.classes["@mozilla.org/filepicker;1"].
	createInstance(Components.interfaces.nsIFilePicker);
    let sharedbundle = document.getElementById("lucifox-shared_strings");
    let text = sharedbundle.getString("save_book");
    fp.init(window, text, Components.interfaces.nsIFilePicker.modeSave);

    text = sharedbundle.getString("books");
    fp.appendFilter(text, "*.epub ; *.EPUB");
    //fp.appendFilter(text, "*.epub ; *.EPUB ; *.pdf ; *.PDF");
    text = sharedbundle.getString("all_files");
    fp.appendFilter(text, "*");

    let defaultString = browser.contentWindow.get_save_name();
    if (!defaultString && file)
	defaultString = file.leafName;
    if (!defaultString)
	defaultString = "book.epub";

    fp.defaultString = defaultString;
    fp.defaultExtension = "epub";

    function my_save_file(targetFile, replace) {
	debug("my_save_file");
	if (!replace && targetFile.exists())
	    return;
	// FIXME honour "replace", it is now OS specific, ok on GNU/Linux
	try {
	    file.copyTo(targetFile.parent, targetFile.leafName);
	} catch (e) {
	    let main_control = Components.classes["@lucidor.org/mainControlService;1"].getService().wrappedJSObject;
	    let strbundle = document.getElementById("lucifox-strings");
	    let text = strbundle.getString("save_error");
	    let nb = browser.contentWindow.document.getElementById("notification_box");
	    nb.appendNotification(text, "save_err", null,
				  nb.PRIORITY_CRITICAL_MEDIUM, [], null);
	}
    }
    let save_it = my_save_file;
    if (!file)
	save_it = browser.contentWindow.save_file;

    let callback = {
	done: function(result) {
	    if (result == fp.returnOK)
		save_it(fp.file, false);
	    else if (result == fp.returnReplace)
		save_it(fp.file, true);
	}
    };

    if (fp.open != undefined)
	fp.open(callback);
    else
	callback.done(fp.show());
}

function get_panel(win) {
    let tabpanels = document.getElementById("tabpanels");
    if (!win)
	return tabpanels.selectedPanel;
    let i = 0;
    for (let panel in loop(tabpanels.childNodes)) {
	if (panel.firstChild.firstChild.nextSibling.contentWindow == win)
	    return [panel, i];
	i++;
    }
    throw "Error";
}

function get_browser(panel) {
    if (panel)
	return panel.firstChild.firstChild.nextSibling;
    let tabpanels = document.getElementById("tabpanels");
    return tabpanels.selectedPanel.firstChild.firstChild.nextSibling;
}

function go_back() {
    let main_browser = get_browser();
    if (main_browser.contentWindow.has_browser) {
	let browser = main_browser.contentWindow.get_browser();
	if (browser.canGoBack) {
	    browser.goBack();
	    return;
	}
    }
    /*
    if (main_browser.canGoBack)
	main_browser.goBack();
    */
}

function go_forward() {
    let main_browser = get_browser();
    if (main_browser.contentWindow.has_browser) {
	let browser = main_browser.contentWindow.get_browser();
	if (browser.canGoForward) {
	    browser.goForward();
	    return;
	}
    }
    /*
    if (main_browser.canGoForward)
	main_browser.goForward();
    */
}

function zoom_in_text() {
    let browser = get_browser();
    if (browser.contentWindow.zoomable) {
	let mdv = browser.contentWindow.getMarkupDocumentViewer();
	mdv.textZoom += 0.05;
    }
}

function zoom_out_text() {
    let browser = get_browser();
    if (browser.contentWindow.zoomable) {
	let mdv = browser.contentWindow.getMarkupDocumentViewer();
	mdv.textZoom -= 0.05;
    }
}

function zoom_in() {
    let browser = get_browser();
    if (browser.contentWindow.zoomable) {
	let mdv = browser.contentWindow.getMarkupDocumentViewer();
	mdv.fullZoom += 0.05;
    }
}

function zoom_out() {
    let browser = get_browser();
    if (browser.contentWindow.zoomable) {
	let mdv = browser.contentWindow.getMarkupDocumentViewer();
	mdv.fullZoom -= 0.05;
    }
}

function update_bar() {
    //debug("update_bar");
    let main_browser = get_browser();
    let browser = { canGoBack: false, canGoForward: false };
    if (main_browser.contentWindow && main_browser.contentWindow.has_browser)
	browser = main_browser.contentWindow.get_browser();

    // For now
    main_browser = { canGoBack: false, canGoForward: false,
		     contentWindow: main_browser.contentWindow };

    let value = true;
    if (main_browser.contentWindow && main_browser.contentWindow.zoomable)
	value = false;
    let button = document.getElementById("zoom_in_text_button");
    button.disabled = value;
    button = document.getElementById("zoom_out_text_button");
    button.disabled = value;
    button = document.getElementById("zoom_in_button");
    button.disabled = value;
    button = document.getElementById("zoom_out_button");
    button.disabled = value;

    let item = document.getElementById("save_file_as_item");
    item.disabled = !(main_browser.contentWindow &&
		      (main_browser.contentWindow.get_save_file ||
		       main_browser.contentWindow.save_page));

    button = document.getElementById("forward-button");
    if (browser.canGoForward || main_browser.canGoForward)
	button.disabled = false;
    else
	button.disabled = true;

    button = document.getElementById("back-button");
    if (browser.canGoBack || main_browser.canGoBack)
	button.disabled = false;
    else
	button.disabled = true;

    //debug("update_bar OK");
}

function update_tab_title(win) {
    //debug("update_tab_title");
    let tabs = document.getElementById("tabs");
    let tab = tabs.selectedItem;
    let index = tabs.selectedIndex;
    let panel = null;
    if (win) {
	let pan = get_panel(win);
	panel = pan[0];
	index = pan[1];
	tab = tabs.childNodes.item(index);
    }
    if (tab) {
	let browser = get_browser(panel);
	let title = "[Untitled]";	// FIXME
	if (browser.contentDocument && browser.contentDocument.title)
	    title = browser.contentDocument.title;
	let shorttitle = title.substring(0, 25);
	if (title.length > 25)
	    shorttitle += "...";
	tab.setAttribute("label", shorttitle);
	tab.setAttribute("tooltiptext", title);
    }

    // Update tab buttons
    let box = document.getElementById("scrollbox");
    let sb;
    try {
	sb = box.boxObject.QueryInterface(Components.interfaces.nsIScrollBoxObject);
    } catch (e) {
	sb = box.boxObject;
    }
    let xp = {};
    sb.getPosition(xp, {});
    let x = xp.value;
    let off = tabs.childNodes.item(0).boxObject.x;
    let last = tabs.childNodes.item(tabs.childNodes.length - 1);

    let button = document.getElementById("tab_right");
    if (x >= (last.boxObject.x + last.boxObject.width - box.boxObject.width))
	button.disabled = true;
    else
	button.disabled = false;

    button = document.getElementById("tab_left");
    if (x > 0)
	button.disabled = false;
    else
	button.disabled = true;
}

function tab_onselect() {
    update_bar();
}

function toggle_tabbar() {
    let tabbar = document.getElementById("tabbar");
    let val = !tabbar.hidden;
    tabbar.hidden = val;
    return val;
}

function show_findbar() {
    let bar = document.getElementById("findbar");
    bar.hidden = false;
    let box = document.getElementById("find-box");
    box.focus();
}

function hide_findbar() {
    let bar = document.getElementById("findbar");
    bar.hidden = true;
}

function on_find_keypress(event) {
    if (event.keyCode == 13)
	find(false);
}

var search_start;
var search_string;
var finder;

function find(backwards) {
    let browser = get_browser();
    if (browser.contentWindow.has_browser)
	browser = browser.contentWindow.get_browser();
    let doc = browser.contentDocument;
    let root = doc.documentElement;
    let fbox = document.getElementById("find-box");

    let word = fbox.value;
    if (!word)
	return;

    let range = doc.createRange();
    range.selectNodeContents(root);

    let start;
    // FIXME should search continue from the same place even if search string has changed?
    if (search_start && (search_start.startContainer.ownerDocument == doc) &&
	(search_string == word)) {
	start = search_start;
	if (backwards)
	    start.setEnd(start.startContainer, start.startOffset);
	else
	    start.setStart(start.endContainer, start.endOffset);
    } else {
	start = doc.createRange();
	if (backwards) {
	    start.setStart(range.endContainer, range.endOffset);
	    start.setEnd(range.endContainer, range.endOffset);
	} else {
	    start.setStart(range.startContainer, range.startOffset);
	    start.setEnd(range.startContainer, range.startOffset);
	}
    }

    let end = doc.createRange();
    if (backwards) {
	end.setStart(range.startContainer, range.startOffset);
	end.setEnd(range.startContainer, range.startOffset);
    } else {
	end.setStart(range.endContainer, range.endOffset);
	end.setEnd(range.endContainer, range.endOffset);
    }

    if (!finder)
	finder = Components.classes["@mozilla.org/embedcomp/rangefind;1"].
	    createInstance(Components.interfaces.nsIFind);
    let match_case = document.getElementById("match-case");
    finder.caseSensitive = match_case.checked;
    finder.findBackwards = backwards;

    let res = finder.Find(word, range, start, end);
    if (res) {
	debug("FIND " + res + " " + res.startOffset);
	let sel = browser.contentWindow.getSelection();
	sel.removeAllRanges();
	sel.addRange(res);
	// FIXME must refresh view so that the old selection disappears
	try {
	    sel.QueryInterface(Components.interfaces.nsISelectionPrivate);
	    sel.scrollIntoView(Components.interfaces.nsISelectionController.SELECTION_ANCHOR_REGION, false, -1, -1);
	} catch (e) {
	    try {
		let selcon = browser.docShell
		    .QueryInterface(Components.interfaces.nsIInterfaceRequestor)
		    .getInterface(Components.interfaces.nsISelectionDisplay)
		    .QueryInterface(Components.interfaces.nsISelectionController);
		selcon.scrollSelectionIntoView(selcon.SELECTION_NORMAL,
					       selcon.SELECTION_ANCHOR_REGION,
					       selcon.SCROLL_CENTER_VERTICALLY);
	    } catch (e) {
		debug("scroll error: " + e);
	    }
	}
    }

    search_start = res;
    search_string = word;
}

function find3(backwards) {
    let browser = get_browser();
    if (browser.contentWindow.has_browser)
	browser = browser.contentWindow.get_browser();
    debug("title: " + browser.contentDocument.title);
    let wbfind = browser.webBrowserFind;
    let fbox = document.getElementById("find-box");
    wbfind.searchString = fbox.value;
    wbfind.findBackwards = backwards;
    wbfind.searchFrames = true;
    wbfind.entireWord = false;
    wbfind.matchCase = false;
    let i = wbfind.findNext();
    debug("FIND '" + wbfind.searchString + "' " + i);
}

function update_notes() {
    let tabpanels = document.getElementById("tabpanels");
    for (let i = 0; i < tabpanels.childNodes.length; i++) {
	let tabpanel = tabpanels.childNodes.item(i);
	let browser = tabpanel.firstChild.firstChild.nextSibling;
	if (browser.contentWindow && browser.contentWindow.list_notes)
	    browser.contentWindow.list_notes();
    }
}

function onclose() {
    let tabpanels = document.getElementById("tabpanels");
    for (let i = 0; i < tabpanels.childNodes.length; i++) {
	let tabpanel = tabpanels.childNodes.item(i);
	let browser = tabpanel.firstChild.firstChild.nextSibling;
	if (browser.contentWindow && browser.contentWindow.warn_unload)
	    browser.contentWindow.warn_unload();
    }
}

function onunload() {
    let bookman = Components.classes["@lucidor.org/bookmanagerService;1"].
	getService().wrappedJSObject;
    return bookman.removeTempdir();
}

function quit() {
    window.close();
    let appStartup = Components.classes["@mozilla.org/toolkit/app-startup;1"].
        getService(Components.interfaces.nsIAppStartup);
    try {
	appStartup.quit(appStartup.eAttemptQuit);
    } catch (e) {
	appStartup.quit(appStartup.eForceQuit);
    }
}

function parse_version(ver) {
    let major = 0;
    let minor = 0;
    let extra = 0;
    try {
	let parts = ver.split(".");
	major = Number(parts[0]);
	if (parts.length > 1)
	    minor = Number(parts[1]);
	if (parts.length > 2)
	    extra = Number(parts[2]);
    } catch (e) {
    }
    return [major, minor, extra];
}

function is_updated(prev, next) {
    if (prev[0] > next[0])			// major
	return true;
    else if (prev[0] == next[0]) {		// major
	if (prev[1] > next[1])		// minor
	    return true;
	else if (prev[1] == next[1]) {	// minor
	    if (prev[2] >= next[2])		// extra
		return true;
	}
    }
    return false;
}
