/*
 *  Lucidor - e-book reader
 *  Copyright © 2008-2016  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("chrome://lucidor/content/utils.jsm");

const XUL_NS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";

var prefs;
var old_bookcase_file;
var new_bookcase_file;

function init() {
    prefs = Components.classes["@mozilla.org/preferences-service;1"].
	getService(Components.interfaces.nsIPrefService).
	getBranch("extensions.lucifox.");

    try {
	let file = prefs.getComplexValue("bookcase.folder",
					 Components.interfaces.nsIFile);
	//old_bookcase_file = file;
    } catch (e) {
	// FIXME
    }

    let language_mapper = Components.classes["@lucidor.org/languageMapperService;1"].getService().wrappedJSObject;
    language_mapper.init();

    /* Font list */
    let scripts = [ "ar", "el", "he", "ja", "ko", "th", "ru" , "x-armn", "x-beng", "x-cans", "x-cyrillic", "x-devanagari", "x-ethi", "x-geor", "x-gujr", "x-guru", "x-khmr", "x-knda", "x-math", "x-mlym", "x-orya", "x-sinh", "x-tamil", "x-telu", "x-tibt", "x-unicode", "x-western", "zh-CN", "zh-HK", "zh-TW" ];
    for (let script in loop(scripts)) {
	let pref = document.createElementNS(XUL_NS, "preference");
	pref.setAttribute("id", "font.default." + script);
	pref.setAttribute("name", "font.default." + script);
	pref.setAttribute("type", "string");
	document.getElementById("font_prefs").appendChild(pref);

	pref = document.createElementNS(XUL_NS, "preference");
	pref.setAttribute("id", "font.name.serif." + script);
	pref.setAttribute("name", "font.name.serif." + script);
	pref.setAttribute("type", "string");
	document.getElementById("font_prefs").appendChild(pref);

	pref = document.createElementNS(XUL_NS, "preference");
	pref.setAttribute("id", "font.name.sans-serif." + script);
	pref.setAttribute("name", "font.name.sans-serif." + script);
	pref.setAttribute("type", "string");
	document.getElementById("font_prefs").appendChild(pref);

    	pref = document.createElementNS(XUL_NS, "preference");
	pref.setAttribute("id", "font.size.variable." + script);
	pref.setAttribute("name", "font.size.variable." + script);
	pref.setAttribute("type", "int");
	document.getElementById("font_prefs").appendChild(pref);

        let item = document.createElementNS(XUL_NS, "menuitem");
	item.setAttribute("id", "scriptitem-" + script);
	item.setAttribute("value", script);
	let lang = language_mapper.translate(script);
	item.setAttribute("label", lang);
	document.getElementById("script_popup").appendChild(item);
    }

    // Select script from start
    let flist = document.getElementById("script_list");
    let locales = [];
    let strbundle = document.getElementById("locale-strings");
    let group = strbundle.getString("font.language.group");
    locales.push(group);
    locales.push("x-western");
    for (let locale in loop(locales)) {
	let item = document.getElementById("scriptitem-" + locale);
	if (item) {
	    flist.selectedItem = item;
	    change_script();
	    break;
	}
    }
}

function change_script() {
    let list = document.getElementById("script_list");
    let script = list.selectedItem.value;

    let pref = document.getElementById("font.size.variable." + script);
    let current = pref.value;
    let tb = document.getElementById("size_box");
    tb.removeAttribute("preference");

    let fprefs = Components.classes["@mozilla.org/preferences-service;1"].
	getService(Components.interfaces.nsIPrefService).
	getBranch("font.");
    tb.value = fprefs.getIntPref("size.variable." + script);
    tb.setAttribute("preference", "font.size.variable." + script);

    update_font_lists(script);
}

function update_font_lists(script) {
    update_default_font_list(script);
    update_serif_font_list(script);
    update_sans_font_list(script);
}

function update_default_font_list(script) {
    let flist = document.getElementById("default_font_list");
    //flist.removeAttribute("preference");

    let fprefs = Components.classes["@mozilla.org/preferences-service;1"].
	getService(Components.interfaces.nsIPrefService).
	getBranch("font.");
    let current = fprefs.getCharPref("default." + script);
    current = decodeURIComponent(escape(current));

    flist.setAttribute("preference", "font.default." + script);
    flist.value = current;
}

function update_serif_font_list(script) {
    let flist = document.getElementById("serif_font_list");
    flist.removeAttribute("preference");
    let fpop = document.getElementById("serif_font_popup");
    while (fpop.firstChild)
	fpop.removeChild(fpop.firstChild);

    //let pref = document.getElementById("font.name.serif." + script);
    //let current = pref.value;
    let fprefs = Components.classes["@mozilla.org/preferences-service;1"].
	getService(Components.interfaces.nsIPrefService).
	getBranch("font.");
    let current = fprefs.getCharPref("name.serif." + script);
    current = decodeURIComponent(escape(current));

    flist.setAttribute("preference", "font.name.serif." + script);

    let fe = Components.classes["@mozilla.org/gfx/fontenumerator;1"]
	.getService(Components.interfaces.nsIFontEnumerator);
    let fonts = fe.EnumerateFonts(script, "serif", {});
    for (let f in loop(fonts)) {
	let item = document.createElementNS(XUL_NS, "menuitem");
	item.setAttribute("value", f);
	item.setAttribute("label", f);
	fpop.appendChild(item);
	if (f == current)
	    flist.value = f;
    }

    if (!flist.value && current) {
	let item = document.createElementNS(XUL_NS, "menuitem");
	item.setAttribute("value", current);
	item.setAttribute("label", current);
	fpop.appendChild(item);
	flist.value = current;
    }
}

function update_sans_font_list(script) {
    let flist = document.getElementById("sans_font_list");
    flist.removeAttribute("preference");
    let fpop = document.getElementById("sans_font_popup");
    while (fpop.firstChild)
	fpop.removeChild(fpop.firstChild);

    //let pref = document.getElementById("font.name.sans-serif." + script);
    //let current = pref.value;
    let fprefs = Components.classes["@mozilla.org/preferences-service;1"].
	getService(Components.interfaces.nsIPrefService).
	getBranch("font.");
    let current = fprefs.getCharPref("name.sans-serif." + script);
    current = decodeURIComponent(escape(current));
    flist.setAttribute("preference", "font.name.sans-serif." + script);

    let fe = Components.classes["@mozilla.org/gfx/fontenumerator;1"]
	.getService(Components.interfaces.nsIFontEnumerator);
    let fonts = fe.EnumerateFonts(script, "sans-serif", {});
    for (let f in loop(fonts)) {
	let item = document.createElementNS(XUL_NS, "menuitem");
	item.setAttribute("value", f);
	item.setAttribute("label", f);
	fpop.appendChild(item);
	if (f == current)
	    flist.value = f;
    }

    if (!flist.value && current) {
	let item = document.createElementNS(XUL_NS, "menuitem");
	item.setAttribute("value", current);
	item.setAttribute("label", current);
	fpop.appendChild(item);
	flist.value = current;
    }
}

var reload_time = -1;

function reload_style_later(event) {
    reload_time = (new Date()).getTime();
    setTimeout(maybe_reload_style, 1100);
}

function maybe_reload_style() {
    if (reload_time < (new Date()).getTime() - 1000)
	reload_style();
}

function reload_style() {
    let sss = Components.classes["@mozilla.org/content/style-sheet-service;1"].
	getService(Components.interfaces.nsIStyleSheetService);
    let ios = Components.classes["@mozilla.org/network/io-service;1"].
	getService(Components.interfaces.nsIIOService);
    let uri = ios.newURI("x-lucidor://css/", null, null);
    if (sss.sheetRegistered(uri, sss.USER_SHEET))
	sss.unregisterSheet(ios.newURI("x-lucidor://css/", null, null),
			    sss.USER_SHEET);
    sss.loadAndRegisterSheet(uri, sss.USER_SHEET);
}

function background_dec() {
    let picker = document.getElementById("backgroundPicker");
    picker.color = color_dec(picker.color);
}

function background_inc() {
    let picker = document.getElementById("backgroundPicker");
    picker.color = color_inc(picker.color);
}

function foreground_dec() {
    let picker = document.getElementById("foregroundPicker");
    picker.color = color_dec(picker.color);
}

function foreground_inc() {
    let picker = document.getElementById("foregroundPicker");
    picker.color = color_inc(picker.color);
}

function color_dec(color) {
    let red = parseInt(color.substring(1, 3), 16);
    let green = parseInt(color.substring(3, 5), 16);
    let blue = parseInt(color.substring(5, 7), 16);
    red -= 2;
    green -= 2;
    blue -= 2;
    red = Math.min(0xff, Math.max(0, red));
    green = Math.min(0xff, Math.max(0, green));
    blue = Math.min(0xff, Math.max(0, blue));
    return "#" + (0x1000000 + red * 0x10000 + green * 0x100 + blue).toString(16).substring(1);
}

function color_inc(color) {
    let red = parseInt(color.substring(1, 3), 16);
    let green = parseInt(color.substring(3, 5), 16);
    let blue = parseInt(color.substring(5, 7), 16);
    red += 2;
    green += 2;
    blue += 2;
    red = Math.min(0xff, Math.max(0, red));
    green = Math.min(0xff, Math.max(0, green));
    blue = Math.min(0xff, Math.max(0, blue));
    return "#" + (0x1000000 + red * 0x10000 + green * 0x100 + blue).toString(16).substring(1);
}

function syncfrom_use_doc_fonts() {
    let pref = document.getElementById("browser.display.use_document_fonts");
    return pref.value != 0;
}

function syncto_use_doc_fonts() {
    if (document.getElementById("useDocumentFonts").checked)
	return 1;
    return 0;
}

function clear_bookmarks() {
    let bookcase = Components.classes["@lucidor.org/bookcaseLiteService;1"].
	getService().wrappedJSObject;
    if (bookcase.open())
	bookcase.clear();
}

function clear_annotations() {
    let note_manager = Components.classes["@lucidor.org/notemanagerService;1"].
	getService().wrappedJSObject;
    if (note_manager.open())
	note_manager.clear();

    var is_extension = true;
    let appinfo = Components.classes["@mozilla.org/xre/app-info;1"].
	getService(Components.interfaces.nsIXULAppInfo);
    if (appinfo.ID == "Lucidor@lucidor.org")
	is_extension = false;

    if (!is_extension) {
	var main_control = Components.classes["@lucidor.org/mainControlService;1"].
	    getService().wrappedJSObject;
	main_control.update_notes();
    }
}

function accept() {
    /*
    if (old_bookcase_file && new_bookcase_file &&
	!new_bookcase_file.equals(old_bookcase_file)) {
	let files = [];
	for (let book in bookcase) {
	    file = book...;
	    if (old_bookcase_file.contains(file, false))
		files.push(file);
	}
	if (files) {
	    let prompts = ...;
	    let text = "Your bookcase contains books that are stored in the old bookcase directory. Do you want to move them to the new bookcase directory?";
	    let migrate = prompts.confirm();
	    if (migrate) {
		for (let file in loop(files)) {
		    let name = "FIXME";
		    try {
			file.moveTo(new_bookcase_file, name);
		    } catch (e) {

		    }
		    // update bookcase book location
		}
	    }
	}
    }
    */
}

function select_directory() {
    let fp = Components.classes["@mozilla.org/filepicker;1"].
	createInstance(Components.interfaces.nsIFilePicker);
    let strbundle = document.getElementById("lucifox-strings");
    let title = strbundle.getString("select_path.title");
    fp.init(window, title, Components.interfaces.nsIFilePicker.modeGetFolder);

    let dir;
    try {
	dir = prefs.getComplexValue("bookcase.folder",
				    Components.interfaces.nsIFile);
	fp.displayDirectory = dir;
    } catch (e) {
    }

    let callback = {
	done: function(result) {
	    if (result != fp.returnOK)
		return;
	    prefs.setComplexValue("bookcase.folder",
				  Components.interfaces.nsIFile, fp.file);
	    //new_bookcase_file = fp.file;
	}
    };

    if (fp.open != undefined)
	fp.open(callback);
    else
	callback.done(fp.show());
}
