/*
 *  Lucidor - e-book reader
 *  Copyright © 2008-2015  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("chrome://lucidor/content/utils.jsm");
Components.utils.import("chrome://lucidor/content/epubcfi.jsm");
Components.utils.import("resource://gre/modules/Downloads.jsm");
Components.utils.import("resource://gre/modules/NetUtil.jsm");
Components.utils.import("resource://gre/modules/PrivateBrowsingUtils.jsm");

const DC_NS = "http://purl.org/dc/elements/1.1/";
const XHTML_NS = "http://www.w3.org/1999/xhtml";
const XUL_NS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";
const DTB_NS = "http://www.daisy.org/z3986/2005/dtbook/";
const OPS_NS = "http://www.idpf.org/2007/ops";
const MATHML_NS = "http://www.w3.org/1998/Math/MathML";
const SVG_NS = "http://www.w3.org/2000/svg";
const MUSICXML_NS = "http://www.musicxml.org/xsd/MusicXML";
const CML_NS = "http://www.xml-cml.org/schema";

const CIWPL = Components.interfaces.nsIWebProgressListener;

var enable_musicxml = false;
var enable_cml = false;
var enable_speech = false;

var is_extension = false;
var zoomable = true;
var has_browser = true;
var browser;
var tempfile = null;
var book = null;
var bookid;
var bookcase_id = -1;
var spine;
var partinfo;
var old_part = -1;
var view_mode = "fixed";
var view_border = false;
var column_adjustment = "center";
var ppdm;	// Pixels Per Decimeter (x-axis)
var allow_remote = false;
var shown_errors = false;
var shown_encryption_warning = false;
var g_scroll_ratio = 0;
var g_scroll_target = null;
var g_scroll_location = null;
var g_extra_stylesheet = null;
var scroll_time = 0;
var scroll_count = 0;
var language_mapper;
var note_manager = null;
var main_control;
var rootfile = 0;
var g_indirectionResolver = null;

function fixup_dom(doc) {
    //debug("fixup_dom: " + doc.documentElement.localName);
    let oldroot = doc.documentElement;
    if (oldroot.localName == "x-ordbrand-root")
	return;

    //debug("1");
    /* Set root elements; glue is necessary for columns */
    /* should not use XHTML_NS here, but needed to get .style */
    let root = doc.createElementNS(XHTML_NS, "x-ordbrand-root");
    let glue = doc.createElementNS(XHTML_NS, "x-ordbrand-glue");
    doc.replaceChild(root, oldroot);
    root.appendChild(glue);
    glue.appendChild(oldroot);
    //debug("2");

    if (g_extra_stylesheet) {
	//debug("3");
	let pi = doc.createProcessingInstruction("xml-stylesheet",
						 g_extra_stylesheet);
	doc.insertBefore(pi, doc.firstChild);
	g_extra_stylesheet = null;
    } else {
	//debug("4");
	/* If <body> has background-color set, then <html>/<dtbook> should
	   inherit the same value unless they set a different value */
	/* FIXME does not yet work correctly if <html>/<dtbook> has
	   background-color set to a different value (rare case) */
	let doc_elem = doc.documentElement;
	let du = Components.classes["@mozilla.org/inspector/dom-utils;1"].
	    createInstance(Components.interfaces.inIDOMUtils);
	let background = null;
	let body;
	if (oldroot.localName == "dtbook")
	    body = doc_elem.getElementsByTagNameNS(DTB_NS, "book").item(0);
	else
	    body = doc_elem.getElementsByTagNameNS(XHTML_NS, "body").item(0);
	if (!body) {
	    // SVG page?
	    return;
	}
	let rules;
	try {
	    rules = du.getCSSStyleRules(body);
	} catch (e) {
	    debug("error: " + e);
	    throw e;
	}
	for (let i = 0; i < rules.Count(); i++) {
	    let el = rules.GetElementAt(i);
	    if (el.parentStyleSheet.href == "x-lucidor://css/")
		continue;
	    for (let j = 0; j < el.style.length; j++) {
		let item = el.style.item(j);
		if (item == "background-color") {
		    let val = el.style.getPropertyValue("background-color");
		    if (true || val != "rgb(255, 255, 255)") {
			background = val;
			break;
		    }
		}
	    }
	    if (background)
		break;
	}
	if (background) {
	    debug("background color set: " + background);
	    body.parentNode.style.backgroundColor = background;
	}
    }

    //debug("5");
    // Handle opf:switch element
    let elems = doc.getElementsByTagNameNS(OPS_NS, "switch");
    for (let elem in loop(elems)) {
	for (let child in loop(elem.childNodes))
	    if (child.nodeType == child.ELEMENT_NODE) {
		child.hidden = true;
	    }
	for (let child in loop(elem.childNodes)) {
	    if (child.nodeType == child.ELEMENT_NODE &&
		child.namespaceURI == OPS_NS) {
		if (child.localName == "case") {
		    let ns = child.getAttribute("required-namespace");
		    if (ns == MATHML_NS || ns == SVG_NS) {
			child.hidden = false;
			break;
		    } else if (enable_musicxml && ns == MUSICXML_NS) {
			add_music(elem, child);
		    } else if (enable_cml && ns == CML_NS) {
			add_chemistry(elem, child);
		    }
		} else if (child.localName == "default") {
		    child.hidden = false;
		    break;
		}
	    }
	}
	for (let child in loop(elem.childNodes))
	    if (child.nodeType == child.ELEMENT_NODE && child.hidden)
		child.parentNode.removeChild(child);
    }

	/*
	if (enable_music) {
	    elems = doc.getElementsByTagNameNS(XHTML_NS, "object");
	    for (let elem in loop(elems)) {
		if (elem.hasAttribute("type")) {
		    if (elem.getAttribute("type") == "application/vnd.recordare.musicxml+xml") {
			add_music_object(elem);
		    }
		}
	    }
	}
	*/

    // add event handlers for links
    let prefs = Components.classes["@mozilla.org/preferences-service;1"].
	getService(Components.interfaces.nsIPrefService).
	getBranch("extensions.lucifox.");
    let preview = prefs.getBoolPref("viewer.linkPreview");
    if (preview) {
	elems = doc.getElementsByTagNameNS(XHTML_NS, "a");
	for (let elem in loop(elems)) {
	    elem.addEventListener("mouseover", link_handler, false);
	    elem.addEventListener("mouseout", link_out_handler, false);
	}
    }

	/*
	elems = doc.getElementsByTagNameNS(DTB_NS, "a");
	for (let elem in loop(elems))
	    elem.addEventListener("mouseover", link_handler, false);
	*/
	/*
	elems = doc.getElementsByTagNameNS(DTB_NS, "noteref");
	for (let elem in loop(elems))
	    elem.addEventListener("mouseover", noteref_handler, false);
	*/

    //debug("9");
}

function on_page_load(event) {
    let doc = event.originalTarget;
    debug("on_page_load: " + doc.location);
}

var progressListener = {
    onLocationChange: function(webProgress, request, location) {
	debug("onLocationChange: " + location.spec);

	let part = find_part(location.spec);
	if (part < 0) {
	    //debug("FIXME Error: unknown spine location");
	    return;
	}
	debug("PART: " + part);

	if (part != old_part) {
	    update_buttons(part);
	    update_navigation_bar();

	    for (let i = 0; i < 2 && i < partinfo[part].length; i++)
		partinfo[part][i].setAttribute("class", "first-current-navitem");
	    for (let i = 2; i < partinfo[part].length; i++)
		partinfo[part][i].setAttribute("class", "current-navitem");

	    if (old_part >= 0)
		for (let link in loop(partinfo[old_part]))
		    link.removeAttribute("class");
	    old_part = part;
	}
    },

    onProgressChange: function(webProgress, request, curSelfProgress,
			       maxSelfProgress, curTotalProgress,
			       maxTotalProgress) {
    },

    onSecurityChange: function(webProgress, request, state) {
    },

    onStateChange: function(webProgress, request, stateFlags, status) {
	debug("on_state_change: " + request.name);

	maybe_display_errors();

	if ((request.name.substring(0, 11) == "resource://") ||
	    (request.name.substring(0, 6) == "about:") ||
	    (request.name == ""))
	    return;

	if (!allow_remote && (stateFlags & CIWPL.STATE_START) &&
	    !(stateFlags & CIWPL.STATE_IS_DOCUMENT) &&
	    (request.name.substring(0, 12) != "x-lucidor://")) {
	    // Cancel remote requests
	    debug("on_state_change cancel");
	    request.cancel(404);
	    return;
	}

	debug("stateFlags: " + stateFlags);

	let doc = webProgress.DOMWindow.document;
	if ((stateFlags & CIWPL.STATE_STOP) &&
	    (stateFlags & CIWPL.STATE_IS_DOCUMENT)) {
	    // This tests whether the loaded document is a top-level document
	    // or included as an object (or subframe?)
	    /*
	    if (webProgress.DOMWindow != webProgress.DOMWindow.parent)
		return;
	    */
	    if (webProgress.DOMWindow != browser.contentWindow)
		return;

	    if (!doc.documentElement) {
		debug("Error: no document");
		// FIXME
	    } else if (doc.documentElement.tagName == "parsererror") {
		book.add_error("ill-formed document");
	    } else {
		fixup_dom(doc);

		set_view_mode(null);
		if (view_mode == "flow") {
		    let button = document.getElementById("flow_button");
		    button.checked = true;
		} else if (view_mode == "columns") {
		    let button = document.getElementById("columns_button");
		    button.checked = true;
		}

		// Scroll window
		let loca = request.name.split("#");
		if (loca.length == 2) {
		    debug("fragment identifier: " + loca[1]);
		    let el = doc.getElementById(loca[1]);
		    if (el) {
			//debug("id");
			let rect = el.getBoundingClientRect();
			webProgress.DOMWindow.scrollTo(rect.left, rect.top);
		    } else {
			//debug("epubcfi?");
			let dr = epubcfi_to_domrange(loca[1]);
			if (dr)
			    scroll_to_domrange(dr, window);
		    }
		} else if (g_scroll_location == request.name) {
		    if (g_scroll_ratio > 0) {
			debug("scroll_ratio: " + g_scroll_ratio);
			let wind = webProgress.DOMWindow;
			if (view_mode == "columns")
			    wind.scrollTo(wind.scrollMaxX * g_scroll_ratio,
					  wind.scrollY);
			else
			    wind.scrollTo(wind.scrollX,
					  wind.scrollMaxY * g_scroll_ratio);
		    } else if (g_scroll_target) {
			scroll_to_target(g_scroll_target, doc,
					 webProgress.DOMWindow);
		    }
		}
		if (g_scroll_location == request.name) {
		    g_scroll_ratio = 0;
		    g_scroll_target = null;
		    g_scroll_location = null;
		}

		if (note_manager) {
		    try {
			if (update_targets(true))	// must come before mark_notes
			    list_notes();

			if (should_mark_notes()) {
			    mark_notes();
			    highlight_notes();
			}
		    } catch (e) {
			debug("annotation error: " + e);
		    }
		}
	    }
	    update_last_date_and_position();
	    update_bar();	// FIXME should not be necessary
	}
    },

    onStatusChange: function(webProgress, request, status, message) {
	set_status_message(message);
    },

    QueryInterface: function(aIID) {
	if (aIID.equals(Components.interfaces.nsIWebProgressListener) ||
	    aIID.equals(Components.interfaces.nsISupportsWeakReference))
	    return this;
	throw Components.results.NS_NOINTERFACE;
    }
};

function scroll_to_target(target, doc, window) {
    debug("scroll_to_target");
    let dr = target_to_domrange(target, get_root(doc));
    if (!dr) {
	debug("target error");
	return;
    }
    //debug(dr + ":" + dr.startOffset);
    scroll_to_domrange(dr, window);
}

function scroll_to_domrange(range, window) {
    let sel = window.getSelection();
    sel.removeAllRanges();
    sel.addRange(range);
    try {
	sel.QueryInterface(Components.interfaces.nsISelectionPrivate);
	sel.scrollIntoView(Components.interfaces.nsISelectionController.SELECTION_ANCHOR_REGION, false, 50, -1);
    } catch (e) {
	try {
	    let selcon = get_selection_controller();
	    selcon.scrollSelectionIntoView(selcon.SELECTION_NORMAL,
					   selcon.SELECTION_ANCHOR_REGION,
					   selcon.SCROLL_CENTER_VERTICALLY);
	} catch (e) {
	    debug("scroll error: " + e);
	}
    }
}

function maybe_display_errors() {
    if (book && book.uses_encryption && !shown_encryption_warning) {
	let bundle = document.getElementById("lucifox-strings");
	let text = bundle.getString("uses_encryption");
	let nb = document.getElementById("notification_box");
	nb.appendNotification(text, "enc", null,
			      nb.PRIORITY_CRITICAL_MEDIUM, [], null);
	shown_encryption_warning = true;
    }

    if (book && book.has_major_errors && !shown_errors) {
	let bundle = document.getElementById("lucifox-strings");
	let text = bundle.getString("has_errors");
	let details = bundle.getString("show_details");
	let nb = document.getElementById("notification_box");

	if (book.errors.length > 0) {
	    let no = nb.appendNotification(text, "err", null,
					   nb.PRIORITY_WARNING_MEDIUM,
					   [{ accessKey: null,
					      callback: show_details,
					      label: details,
					      popup: null }], null);
	    autohide(no, true);
	} else {
	    let no = nb.appendNotification(text, "err", null,
					   nb.PRIORITY_WARNING_MEDIUM,
					   [], null);
	    autohide(no, true);
	}
	shown_errors = true;
    }

    let prefs = Components.classes["@mozilla.org/preferences-service;1"].
	getService(Components.interfaces.nsIPrefService).
	getBranch("extensions.lucifox.");
    let ignore = prefs.getBoolPref("viewer.ignoreWarnings");
    if (!ignore && book && book.errors.length > 0) {
	let item = document.getElementById("details_item");
	item.disabled = false;
	item = document.getElementById("details_button");
	item.hidden = false;
    }
}

function update_bar() {
    main_control.update_bar();
}

var historyListener = {
    timeout: 200,	// FIXME hack

    OnHistoryGoBack: function(backURI) {
	setTimeout(function() { main_control.update_bar(); }, this.timeout);
	return true;
    },

    OnHistoryGoForward: function(forwardURI) {
	setTimeout(function() { main_control.update_bar(); }, this.timeout);
	return true;
    },

    OnHistoryGotoIndex: function(index, gotoURI) {
	setTimeout(function() { main_control.update_bar(); }, this.timeout);
	return true;
    },

    OnHistoryNewEntry: function(newURI) {
	setTimeout(function() { main_control.update_bar(); }, this.timeout);
	return true;
    },

    OnHistoryPurge: function(numEntries) {
	setTimeout(function() { main_control.update_bar(); }, this.timeout);
	return true;
    },

    OnHistoryReload: function(reloadURI, reloadFlags) {
	setTimeout(function() { main_control.update_bar(); }, this.timeout);
	return true;
    },

    QueryInterface: function(aIID) {
	if (aIID.equals(Components.interfaces.nsISHistoryListener) ||
	    aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
	    aIID.equals(Components.interfaces.nsISupports))
	    return this;
	throw Components.results.NS_NOINTERFACE;
    }
};

var uriContentListener = {
    loadCookie: null,
    parentContentListener: null,

    QueryInterface: function(aIID) {
	if (aIID.equals(Components.interfaces.nsIURIContentListener) ||
	    aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
	    aIID.equals(Components.interfaces.nsISupports))
	    return this;
	throw Components.results.NS_NOINTERFACE;
    },

    canHandleContent: function(aContentType, aIsContentPreferred) {
	if (this.parentContentListener)
	    return this.parentContentListener.canHandleContent(aContentType,
							       aIsContentPreferred);
	return false;
    },

    doContent: function(aContentType, aIsContentPreferred, aRequest, aContentHandler) {
	if (this.parentContentListener)
	    return this.parentContentListener.doContent(aContentType,
							aIsContentPreferred,
							aRequest,
							aContentHandler);
	throw Components.results.NS_ERROR_NOT_IMPLEMENTED;	// FIXME
    },

    isPreferred: function(aContentType, desiredContentType) {
	if (this.parentContentListener)
	    return this.parentContentListener.isPreferred(aContentType,
							  desiredContentType);
	return false;
    },

    onStartURIOpen: function(aURI) {
	debug("onStartURIOpen: " + aURI.scheme);
	if (aURI.scheme == "x-lucidor")
	    return false;
	if (aURI.spec == "about:blank")
	    return false;
	if (aURI.spec == "file:///FIXMEqwerty")
	    return true;
	// if BUSY_FLAGS_PAGE_LOADING
	if (browser.docShell.busyFlags & 4)
	    return true;

	let ok = true;
	let prefs = Components.classes["@mozilla.org/preferences-service;1"].
	    getService(Components.interfaces.nsIPrefService).
	    getBranch("extensions.lucifox.");
	let prompt_on = prefs.getBoolPref("viewer.promptOnExternalLink");
	if (prompt_on) {
	    let bundle = document.getElementById("lucifox-strings");
	    let title = bundle.getString("external_link");
	    let text = bundle.getString("external_link_question");
	    text += "\n(" + aURI.spec + ")";
	    let prompts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
	    ok = prompts.confirm(window, title, text);
	}
	if (ok)
	    main_control.add_web_tab(aURI.spec);
	return true;
    }
};

var noExternalURIContentListener = {
    loadCookie: null,
    parentContentListener: null,

    QueryInterface: function(aIID) {
	if (aIID.equals(Components.interfaces.nsIURIContentListener) ||
	    aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
	    aIID.equals(Components.interfaces.nsISupports))
	    return this;
	throw Components.results.NS_NOINTERFACE;
    },

    canHandleContent: function(aContentType, aIsContentPreferred) {
	if (this.parentContentListener)
	    return this.parentContentListener.canHandleContent(aContentType,
							       aIsContentPreferred);
	return false;
    },

    doContent: function(aContentType, aIsContentPreferred, aRequest, aContentHandler) {
	if (this.parentContentListener)
	    return this.parentContentListener.doContent(aContentType,
							aIsContentPreferred,
							aRequest,
							aContentHandler);
	throw Components.results.NS_ERROR_NOT_IMPLEMENTED;	// FIXME
    },

    isPreferred: function(aContentType, desiredContentType) {
	if (this.parentContentListener)
	    return this.parentContentListener.isPreferred(aContentType,
							  desiredContentType);
	return false;
    },

    onStartURIOpen: function(aURI) {
	debug("ONSTARTURIOPEN: " + aURI.spec);
	if (aURI.scheme == "x-lucidor")
	    return false;
	return true;
    }
};

var contextHandler = {
    handleEvent: function(event) {
	if (event.type == "contextmenu") {
	    debug("E: " + event.type);
	    let popup = document.getElementById("contextmenu");
	    popup.showPopup(null, event.screenX + 2, event.screenY + 2,
			    "context");
	    event.stopPropagation();
	}
    }
};

var sidebarContextHandler = {
    handleEvent: function(event) {
	if (event.type == "contextmenu") {
	    debug("E: " + event.type);
	    /*
	    let popup = document.getElementById("contextmenu");
	    popup.showPopup(null, event.screenX + 2, event.screenY + 2,
			    "context");
	     */
	    event.stopPropagation();
	}
    }
};

function init() {
    let prefs = Components.classes["@mozilla.org/preferences-service;1"].
	getService(Components.interfaces.nsIPrefService).
	getBranch("extensions.lucifox.");

    enable_debug(prefs.getBoolPref("debug.enable"));

    debug("viewepub init: " + document.documentURI);

    let appinfo = Components.classes["@mozilla.org/xre/app-info;1"].
	getService(Components.interfaces.nsIXULAppInfo);
    if (appinfo.ID == "Lucidor@lucidor.org")
	is_extension = false;
    else
	is_extension = true;

    allow_remote = prefs.getBoolPref("viewer.allowRemote");

    const CIWP = Components.interfaces.nsIWebProgress;
    browser = document.getElementById("epub_browser");
    //browser.addEventListener("load", on_page_load, true);
    browser.addProgressListener(progressListener,
				CIWP.NOTIFY_STATE_ALL | CIWP.NOTIFY_LOCATION);
    browser.docShell.allowAuth = false;
    browser.docShell.allowJavascript = prefs.getBoolPref("viewer.javascript");
    browser.docShell.allowMetaRedirects = false;
    //browser.docShell.allowPlugins = false;
    browser.docShell.useErrorPages = true;
    let wnd = browser.docShell.QueryInterface(Components.interfaces.nsIInterfaceRequestor).getInterface(Components.interfaces.nsIURIContentListener);
    uriContentListener.parentContentListener = wnd.parentContentListener;
    wnd.parentContentListener = uriContentListener;

    if (!is_extension) {
	// FIXME Why is it necessary to manually create a sessionHistory?
	// It is not necessary for viewweb.
	try {
	    // FIXME Multiprocess
	    browser.webNavigation.sessionHistory = Components.classes["@mozilla.org/browser/shistory;1"].createInstance(Components.interfaces.nsISHistory);
	    browser.webNavigation.sessionHistory.addSHistoryListener(historyListener);
	} catch (e) {
	}
    }

    if (is_extension) {
	browser.contentWindow.addEventListener("contextmenu", contextHandler,
					       true);
	let el = document.getElementById("ve_deck");
	el.addEventListener("contextmenu", sidebarContextHandler, true);
    }

    // Register default stylesheets
    let sss = Components.classes["@mozilla.org/content/style-sheet-service;1"].
	getService(Components.interfaces.nsIStyleSheetService);
    let ios = Components.classes["@mozilla.org/network/io-service;1"].
	getService(Components.interfaces.nsIIOService);
    let uri = ios.newURI("chrome://lucidor/content/dtbook.css", null, null);
    if (!sss.sheetRegistered(uri, sss.AGENT_SHEET))
	sss.loadAndRegisterSheet(uri, sss.AGENT_SHEET);
    uri = ios.newURI("chrome://lucidor/content/xhtml.css", null, null);
    if (!sss.sheetRegistered(uri, sss.AGENT_SHEET))
	sss.loadAndRegisterSheet(uri, sss.AGENT_SHEET);
    uri = ios.newURI("chrome://lucidor/content/ordbrand.css", null, null);
    if (!sss.sheetRegistered(uri, sss.AGENT_SHEET))
	sss.loadAndRegisterSheet(uri, sss.AGENT_SHEET);
    uri = ios.newURI("x-lucidor://css/", null, null);
    if (!sss.sheetRegistered(uri, sss.USER_SHEET))
	sss.loadAndRegisterSheet(uri, sss.USER_SHEET);

    let hbrowser = document.getElementById("hidden_browser_tt");
    hbrowser.docShell.allowAuth = false;
    hbrowser.docShell.allowImages = false;
    hbrowser.docShell.allowMedia = false;
    hbrowser.docShell.allowJavascript = false;
    hbrowser.docShell.allowMetaRedirects = false;
    hbrowser.docShell.allowPlugins = false;
    hbrowser.docShell.useErrorPages = true;
    hbrowser.addProgressListener(tooltipProgressListener,
				 CIWP.NOTIFY_STATE_DOCUMENT);
    wnd = hbrowser.docShell.QueryInterface(Components.interfaces.nsIInterfaceRequestor).getInterface(Components.interfaces.nsIURIContentListener);
    noExternalURIContentListener.parentContentListener = wnd.parentContentListener;
    wnd.parentContentListener = noExternalURIContentListener;

    hbrowser = document.getElementById("hidden_browser_s");
    hbrowser.docShell.allowAuth = false;
    hbrowser.docShell.allowImages = false;
    hbrowser.docShell.allowMedia = false;
    hbrowser.docShell.allowJavascript = false;
    hbrowser.docShell.allowMetaRedirects = false;
    hbrowser.docShell.allowPlugins = false;
    hbrowser.docShell.useErrorPages = true;
    hbrowser.addProgressListener(searchProgressListener,
				 CIWP.NOTIFY_STATE_DOCUMENT);
    wnd = hbrowser.docShell.QueryInterface(Components.interfaces.nsIInterfaceRequestor).getInterface(Components.interfaces.nsIURIContentListener);
    wnd.parentContentListener = noExternalURIContentListener;

    let box = document.getElementById("hidden_box");
    ppdm = box.boxObject.width;

    language_mapper = Components.classes["@lucidor.org/languageMapperService;1"].getService().wrappedJSObject;
    language_mapper.init();

    note_manager = Components.classes["@lucidor.org/notemanagerService;1"].getService().wrappedJSObject;
    if (note_manager.open()) {
	let el = document.getElementById("annotations_button");
	el.hidden = false;
	el = document.getElementById("export_notes_button");
	el.hidden = false;
	el = document.getElementById("print_notes_button");
	el.hidden = false;
    } else
	note_manager = null;

    if (is_extension) {
	fakeControlService.init(null);
	main_control = fakeControlService;
    } else {
	main_control = Components.classes["@lucidor.org/mainControlService;1"].
	    getService().wrappedJSObject;
    }
    view_border = prefs.getBoolPref("viewer.border");
    view_mode = prefs.getCharPref("viewer.viewMode");
    column_adjustment = prefs.getCharPref("viewer.columnAdjustment");

    try {
	let textbox = document.getElementById("w-box");
	textbox.value = prefs.getCharPref("viewer.columnWidth");
	textbox = document.getElementById("h-box");
	textbox.value = prefs.getCharPref("viewer.columnHeight");
    } catch (e) {
    }

    myPrefObserver.register();

    document.addEventListener("LucidorEvent", lucidor_event, false, true);

    if (is_extension) {
	enable_musicxml = false;
	enable_cml = false;
	enable_speech = false;
    } else
	enable_speech = false;	// FIXME

    if (enable_speech)
	document.getElementById("play-button").hidden = false;

    update_visibility();

    init1();
};

function init1() {
    let bundle = document.getElementById("lucifox-strings");
    let text = bundle.getString("opening_book");
    set_open_message(text);
    document.getElementById("load_box").hidden = false;

    let ioService = Components.classes["@mozilla.org/network/io-service;1"].
	getService(Components.interfaces.nsIIOService);
    let docuri = ioService.newURI(document.documentURI, "UTF-8", null);
    docuri.QueryInterface(Components.interfaces.nsIURL);

    let urispec = null;
    let params = docuri.query.split("&");
    for (let par in loop(params)) {
	let pair = par.split("=");
	if (pair[0] == "uri")
	    urispec = decodeURIComponent(pair[1]);
	else if (pair[0] == "id")
	    bookcase_id = decodeURIComponent(pair[1]);
	else if (pair[0] == "root") {
	    rootfile = parseInt(decodeURIComponent(pair[1]), 10);
	    if (isNaN(rootfile))
		rootfile = 0;
	}
    }

    debug("URI: " + urispec);
    let link = document.getElementById("link");
    link.setAttribute("href", urispec);
    if (link.firstChild)
	link.removeChild(link.firstChild);
    link.appendChild(document.createTextNode(urispec));

    let nsuri;
    try {
	nsuri = ioService.newURI(urispec, "UTF-8", null);
    } catch (e) {
	debug("strange uri: " + e);
	error_opening(0);
	return null;
    }

    // Open already copied files directly
    let bookman = Components.classes["@lucidor.org/bookmanagerService;1"].
	getService().wrappedJSObject;
    let filepair = bookman.getFile(window);
    if (filepair) {
	let furispec = filepair[0]
	let filename = filepair[1]
	debug("furispec: " + furispec);
	let file = Components.classes["@mozilla.org/network/protocol;1?name=file"].
	    getService(Components.interfaces.nsIFileProtocolHandler).
	    getFileFromURLSpec(furispec);
	try {
	    tempfile = file;
	    return init2(file, nsuri.spec, filename);
	} catch (e) {
	    debug("init2 failed: " + e);
	    error_opening(0);
	    return null;
	}
    }

    // Open local resource files directly
    if (nsuri.schemeIs("resource")) {
	let urispec2 = Components.classes["@mozilla.org/network/protocol;1?name=resource"].
	    getService(Components.interfaces.nsIResProtocolHandler).
	    resolveURI(nsuri);
	let file = Components.classes["@mozilla.org/network/protocol;1?name=file"].
	    getService(Components.interfaces.nsIFileProtocolHandler).
	    getFileFromURLSpec(urispec2);
	try {
	    tempfile = null;
	    return init2(file, nsuri.spec, null);
	} catch (e) {
	    debug("init2 failed: " + e);
	    error_opening(0);
	    return null;
	}
    }

    // Maybe open local files directly
    let prefs = Components.classes["@mozilla.org/preferences-service;1"].
	getService(Components.interfaces.nsIPrefService).
	getBranch("extensions.lucifox.");
    let alwaysCopy = prefs.getBoolPref("alwaysCopy");
    if (!alwaysCopy && nsuri.schemeIs("file")) {
	let file = Components.classes["@mozilla.org/network/protocol;1?name=file"].
	    getService(Components.interfaces.nsIFileProtocolHandler).
	    getFileFromURLSpec(nsuri.spec);
	try {
	    tempfile = null;
	    return init2(file, nsuri.spec, null);
	} catch (e) {
	    debug("init2 failed: " + e);
	    error_opening(0);
	    return null;
	}
    }

    // Make local copies of remote files before opening them
    let file = bookman.getTempfile(".epub")
    try {
	file.createUnique(Components.interfaces.nsIFile.NORMAL_FILE_TYPE,
			  parseInt("0600", 8));
    } catch (e) {
	error_opening(2);
	return null;
    }

    document.getElementById("open_box").hidden = false;

    let name = file.path;
    let pm = document.getElementById("progress");

    // Unfortunately this does not seem to work with redirects
    // (archive.org not working)
    let downloadp = Downloads.createDownload({ source: nsuri, target: file });
    downloadp.then(function (download) {
	download.onchange = function () {
	    debug("onchange: " + download.progress);
	    if (download.hasProgress) {
		pm.mode = "determined";
		pm.value = Math.floor(100.0 * download.progress);
	    }
	};

	let prom = download.start();
	prom.then(function (value) {
	    let fn = null;
	    tempfile = file;
	    init2(file, nsuri.spec, fn);
	},
	function (reason) {
	    download.finalize(true);
	    Components.utils.reportError(reason);
	    debug("Retrieving book failed");
	    error_opening(1);
	});
    },
    Components.utils.reportError);

    return null;
}

function unload() {
    debug("unload");
    myPrefObserver.unregister();

    let bookman = Components.classes["@lucidor.org/bookmanagerService;1"].
	getService().wrappedJSObject;
    bookman.removeBook(bookid);

    if (book)
	book.close();

    if (tempfile) {
	try {
	    let tempdir = bookman.getTempdir();
	    if (tempfile && tempfile.isFile() &&
		tempdir.contains(tempfile, false)) {
		tempfile.remove(false);
	    }
	} catch (e) {
	}
    }

    if (note_manager)
	note_manager.close();

    if (is_extension)
	bookman.tryRemoveTempdir();
}

function before_unload() {
    debug("before_unload");
    update_last_date_and_position();
    return null;
}

function open_bookcase() {
    let bookcase = null;
    if (is_extension) {
	let prefs = Components.classes["@mozilla.org/preferences-service;1"].
	    getService(Components.interfaces.nsIPrefService).
	    getBranch("extensions.lucifox.");
	if (prefs.getBoolPref("lucifox.bookcaseLite")) {
	    bookcase = Components.classes["@lucidor.org/bookcaseLiteService;1"].
		getService().wrappedJSObject;
	    bookcase.open();
	    let id = bookcase.get_or_add_from_sha1(book.sha1);
	}
    } else {
	bookcase = Components.classes["@lucidor.org/bookcaseSqlService;1"].
	    getService().wrappedJSObject;
	bookcase.open();
    }
    return bookcase;
}

function update_date_read() {
    // PrivateBrowsingUtils.isContentWindowPrivate(document.defaultView)
    if (PrivateBrowsingUtils.isWindowPrivate(document.defaultView))
	return;
    if (!book)
	return;

    let bookcase = open_bookcase();
    if (bookcase) {
	let id = bookcase.get_id_from_sha1(book.sha1);
	if (id >= 0)
	    bookcase.update_date_read(id);
	bookcase.close();
    }
}

function update_last_date_and_position() {
    debug("update_last_date_and_position");
    if (PrivateBrowsingUtils.isWindowPrivate(document.defaultView))
	return;
    if (!book)
	return;

    try {
	let bookcase = open_bookcase();
	if (bookcase) {
	    let id = bookcase.get_id_from_sha1(book.sha1);
	    if (id >= 0) {
		bookcase.update_date_read(id);
		bookcase.update_last_position(id, get_current_position());
	    }
	    bookcase.close();
	}
    } catch (e) {
    }
}

var book_file;

function init2(file, uri, fn) {
    debug("viewepub init2");
    //enable_debug(false);

    book = Components.classes["@lucidor.org/epubDocument;1"].
	createInstance().wrappedJSObject;
    if (fn)
	book.set_filename(fn);
    book.init_rootfile(file, uri, init3, rootfile);
}

function init3(bok, uri) {
    //enable_debug(true);
    debug("viewepub init3");

    if (!bok) {
	debug("Error opening book: " + uri);
	return error_opening(2);
    }

    let bundle = document.getElementById("lucifox-strings");
    let sharedbundle = document.getElementById("lucifox-shared_strings");

    let main = document.getElementById("open_box");
    main.hidden = true;
    main = document.getElementById("main_box");
    main.hidden = false;

    book.original_location = uri;	// FIXME

    if (!is_extension && book.original_location !=
	"resource://lucidor/manual/lucidor-manual.epub") {
	let bookcase = Components.classes["@lucidor.org/bookcaseSqlService;1"].
	    getService().wrappedJSObject;
	bookcase.open();
	if (bookcase.get_id_from_sha1(book.sha1) < 0) {
	    let prefs = Components.classes["@mozilla.org/preferences-service;1"].
		getService(Components.interfaces.nsIPrefService).
		getBranch("extensions.lucifox.");
	    let prompt_before = prefs.getBoolPref("bookcase.promptBeforeAdd");
	    if (prompt_before) {
		book_file = book.file;	// FIXME

		let text = bundle.getString("addQuestion");
		let yestext = sharedbundle.getString("yes");
		let notext = sharedbundle.getString("no");

		let nb = document.getElementById("notification_box");
		let no = nb.appendNotification(text, "add", null,
					       nb.PRIORITY_INFO_MEDIUM,
					       [{ accessKey: null,
						  callback: answer_yes,
						  label: yestext,
						  popup: null },
						{ accessKey: null,
						  callback: close_callback,
						  label: notext,
						  popup: null }], null);
		autohide(no, true);
	    } else {
		try_copy_and_add_book(bookcase, book, book.file);
	    }
	    book.file = null;
	}
    }

    let bookman = Components.classes["@lucidor.org/bookmanagerService;1"].
	getService().wrappedJSObject;
    bookid = bookman.addBook("epub", book);
    bookman.setBundle(bundle);

    //enable_debug(false);

    document.title = book.opf.metadata.title;
    main_control.update_tab_title(window);

    partinfo = new Array(book.opf.spine.length);

    g_indirectionResolver = new indirectionResolver(book, true);

    // Info sidebar
    let box = document.getElementById("ve_info");

    function add_line(entity_id, value) {
	add_line1(entity_id, value, null);
    }

    function add_line1(entity_id, value, extra) {
	if (!value)
	    return;
	let label = sharedbundle.getString(entity_id);
	if (extra)
	    label += " (" + extra + ")";
	add_line2(label, value);
    }

    let last_header = null;

    function add_line2(label, value) {
	if (!value)
	    return;
	if (label == last_header && last_header)
	    label = null;
	else if (label)
	    last_header = label;
	add_line3(label, value);
    }

    function add_line3(label, value) {
	if (label) {
	    let div = document.createElementNS(XHTML_NS, "div");
	    div.setAttribute("class", "infoheader");
	    box.appendChild(div);
	    let node = document.createTextNode(label);
	    div.appendChild(node);
	}
	if (value) {
	    let div = document.createElementNS(XHTML_NS, "div");
	    div.setAttribute("class", "infoitem");
	    box.appendChild(div);
	    let node;
	    if (value.length >= 100)
		node = document.createTextNode(trim_end(value.substring(0, 100), 10) + "\u2026");	// "..."
	    else
		node = document.createTextNode(value);
	    div.appendChild(node);
	    if (value.length >= 100)
		div.setAttribute("tooltiptext", value);
	}
    }

    function add_header(entity_id) {
	let label = sharedbundle.getString(entity_id);
	add_line3(label, null);
    }

    function add_role(person, default_role) {
	let label = null;
	if (person.role)
	    try {
		label = bundle.getString("role_" + person.role);
	    } catch (e) {
	    }
	if (!label) {
	    label = bundle.getString(default_role);
	    if (person.role)
		label += " (" + person.role + ")";
	}
	add_line2(label, person.name);
    }

    function add_date(date) {
	let label = null;
	if (date.event)
	    try {
		label = bundle.getString("date_" + date.event);
	    } catch (e) {
	    }
	if (!label) {
	    label = sharedbundle.getString("date");
	    if (date.event)
		label += " (" + date.event + ")";
	}
	add_line2(label, get_datestring(date.date));
    }

    function get_datestring(date) {
	if (date instanceof Date)
	    return date.toLocaleString();

	let d = new Date(date);
	if (d.getYear() > 0)
	    return d.toLocaleString();

	return date;
    }

    add_line("title", book.opf.metadata.title);
    if ((book.opf.metadata.creator.length == 1) &&
	!book.opf.metadata.creator[0].role)
	add_line("author", book.opf.metadata.creator[0].name);
    else
	for (let creator in loop(book.opf.metadata.creator))
	    add_role(creator, "creator");
    if (book.opf.metadata.contributor != undefined)
	for (let contrib in loop(book.opf.metadata.contributor))
	    add_role(contrib, "contributor");
    if (book.opf.metadata.publisher != undefined)
	add_line("publisher", book.opf.metadata.publisher);
    if (book.opf.metadata.rights != undefined)
	add_line("rights", book.opf.metadata.rights);
    if (book.opf.metadata.date != undefined)
	for (let date in loop(book.opf.metadata.date))
	    add_date(date);
    if (book.opf.metadata.modified != undefined)
	add_date({ date: book.opf.metadata.modified, event: 'modification' });
    if (book.opf.metadata.languages != undefined) {
	for (let language in loop(book.opf.metadata.languages)) {
	    let lang = language_mapper.translate(language);
	    add_line("language", lang);
	}
    }
    if (book.opf.metadata.description != undefined)
	add_line("description", book.opf.metadata.description);
    if (book.opf.metadata.subject != undefined)
	for (let subject in loop(book.opf.metadata.subject))
	    add_line("subject", subject);
    if (book.opf.metadata.coverage != undefined)
	for (let coverage in loop(book.opf.metadata.coverage))
	    add_line("coverage", coverage);
    if (book.opf.metadata.type != undefined)
	for (let type in loop(book.opf.metadata.type))
	    add_line("type", type);
    if (book.opf.metadata.format != undefined)
	for (let format in loop(book.opf.metadata.format))
	    add_line("format", format);
    if (book.opf.metadata.source != undefined)
	for (let source in loop(book.opf.metadata.source))
	    add_line("source", source);
    if (book.opf.metadata.relation != undefined)
	for (let relation in loop(book.opf.metadata.relation))
	    add_line("relation", relation);
    if (book.opf.uid != undefined)
	add_line("uid", book.opf.uid);
    for (let id in loop(book.opf.metadata.identifier))
	if (!id.uid)
	    add_line1("identifier", id.id, id.scheme);
    add_line("location", book.original_location);

    // Spine sidebar
    spine = book.getSpine();
    box = document.getElementById("ve_spine");
    for (let i = 0; i < spine.length; i++) {
	let item = spine[i];
	let div = document.createElementNS(XHTML_NS, "div");
	div.setAttribute("class", "navitem");
	box.appendChild(div);
	let link = document.createElementNS(XHTML_NS, "a");
	div.appendChild(link);

	if (book.is_encrypted(item.href)) {
	    link.setAttribute("style", "text-decoration: line-through");
	} else {
	    link.setAttribute("href", "#");
	    /*link.addEventListener("click", {
		uri: bookid + "/" + item.href,
		handleEvent: function(e) { view_uri(this.uri); }
	    }, false);*/
	    link.addEventListener("click", {
		part: i,
		handleEvent: function(e) { view_part(this.part, false, true); }
	    }, false);
	}

	let span = document.createElementNS(XHTML_NS, "span");
	span.setAttribute("class", "infoheader");
	span.appendChild(document.createTextNode(item.idref + ":"));
	link.appendChild(span);
	let text = " " + item.href;
	if (!item.linear) {
	    let nl_text = bundle.getString("auxiliary");
	    text += " (" + nl_text + ")";
	}
	link.appendChild(document.createTextNode(text));

	if (!partinfo[i])
	    partinfo[i] = [];
	partinfo[i].push(link);
    }

    // Navigation sidebar

    function traverse_items(node, level, container) {
	debug("*: " + node + ";" + container);
	let div = document.createElementNS(XHTML_NS, "div");
	div.setAttribute("class", "navitem");
	div.setAttribute("level", level);
	container.appendChild(div);
	let handle = null;

	if (node.list) {
	    let spa = document.createElementNS(XHTML_NS, "span");
	    div.appendChild(spa);
	    handle = document.createElementNS(XHTML_NS, "a");
	    spa.appendChild(handle);
	    handle.setAttribute("href", "#");
	    handle.addEventListener("click", function(event) {
		toggle_it(event.currentTarget);
	    }, false);
	    set_as_open(handle);
	    spa.appendChild(document.createTextNode(" "));	// NBSP

	    if (!node.href) {
		let text = node.head;
		if (!text)
		    text = '-';
		handle.appendChild(document.createTextNode(text));
	    }
	}
	if (node.href) {
	    debug("*: " + node.href);
	    let link = document.createElementNS(XHTML_NS, "a");
	    div.appendChild(link);
	    if (node.broken || book.is_encrypted(node.href)) {
		link.setAttribute("style", "text-decoration: line-through");
	    } else {
		link.setAttribute("href", "#");
		link.addEventListener("click", {
		    uri: bookid + "/" + node.href,
		    handleEvent: function(e) { view_uri(this.uri); }
		}, false);
	    }
	    let text = node.text;
	    link.appendChild(document.createTextNode(text));
	    let part = find_part(node.href);
	    if (part >= 0) {
		if (!partinfo[part])
		    partinfo[part] = [];
		partinfo[part].push(link);
	    }
	}
	if (node.list) {
	    let child_div = document.createElementNS(XHTML_NS, "div");
	    child_div.setAttribute("class", "navchild");
	    hide_element(child_div, true);
	    div.appendChild(child_div);
	    for (let n in loop(node.list))
		traverse_items(n, level + 1, child_div);
	}
    }

    if (book.navs.toc) {
	debug("TOC navigation sidebar");
	let nav = book.navs.toc;
	let button = document.getElementById("content_button");
	button.hidden = false;
	let box = document.getElementById("ve_toc");

	// Link to show the navigation page
	let div = document.createElementNS(XHTML_NS, "div");
	div.setAttribute("class", "navheader navitem");
	let link = document.createElementNS(XHTML_NS, "a");
	div.appendChild(link);
	link.setAttribute("href", "#");
	link.addEventListener("click", {
	    uri: bookid + "/" + book.opf.path + book.opf.navitem.href,
	    handleEvent: function(e) { view_uri(this.uri); }
	}, false);
	let text = bundle.getString("type_toc");
	link.appendChild(document.createTextNode(text));
	box.appendChild(div);

	if (nav.head) {
	    let div = document.createElementNS(XHTML_NS, "div");
	    div.setAttribute("class", "navheader");
	    div.appendChild(document.createTextNode(nav.head));
	    box.appendChild(div);
	}

	for (let n in loop(nav.list))
	    traverse_items(n, 0, box);

	// FIXME Add preference for this?
	if (false) {
	    expand_all(true);
	} else if (false) {
	    collapse_all(true);
	} else {
	    show_all(true);
	}
    }

    if (book.navs.pagelist) {
	debug("Pagelist navigation sidebar");
	let nav = book.navs.pagelist;
	let button = document.getElementById("pages_button");

	button.hidden = false;
	let box = document.getElementById("ve_pages");

	if (nav.head) {
	    let div = document.createElementNS(XHTML_NS, "div");
	    div.setAttribute("class", "navheader");
	    div.appendChild(document.createTextNode(nav.head));
	    box.appendChild(div);
	}

	for (let n in loop(nav.list))
	    traverse_items(n, 0, box);

	// FIXME Add preference for this?
	if (false) {
	    expand_all(true);
	} else if (false) {
	    collapse_all(true);
	} else {
	    show_all(true);
	}
    }

    if (book.navs.landmarks) {
	debug("Landmarks navigation sidebar");
	let nav = book.navs.landmarks;
	let button = document.getElementById("guide_button");

	button.hidden = false;
	let box = document.getElementById("ve_guide");

	if (nav.head) {
	    let div = document.createElementNS(XHTML_NS, "div");
	    div.setAttribute("class", "navheader");
	    div.appendChild(document.createTextNode(nav.head));
	    box.appendChild(div);
	}

	for (let n in loop(nav.list))
	    traverse_items(n, 0, box);

	// FIXME Add preference for this?
	if (false) {
	    expand_all(true);
	} else if (false) {
	    collapse_all(true);
	} else {
	    show_all(true);
	}
    }

    // Contents sidebar
    function traverse_points(node, level, container) {
	debug("*: " + node + ";" + container);
	if (node.label && node.content) {
	    debug("*: " + node.label + ";" + node.content);
	    let div = document.createElementNS(XHTML_NS, "div");
	    div.setAttribute("class", "navitem");
	    div.setAttribute("level", level);
	    container.appendChild(div);
	    if (node.children) {
		let spa = document.createElementNS(XHTML_NS, "span");
		div.appendChild(spa);
		let handle = document.createElementNS(XHTML_NS, "a");
		spa.appendChild(handle);
		handle.setAttribute("href", "#");
		handle.addEventListener("click", function(event) {
		    toggle_it(event.currentTarget);
		}, false);
		set_as_open(handle);
		spa.appendChild(document.createTextNode(" "));	// NBSP
	    }
	    let link = document.createElementNS(XHTML_NS, "a");
	    div.appendChild(link);
	    if (node.broken || book.is_encrypted(node.content)) {
		link.setAttribute("style", "text-decoration: line-through");
	    } else {
		link.setAttribute("href", "#");
		link.addEventListener("click", {
		    uri: bookid + "/" + node.content,
		    handleEvent: function(e) { view_uri(this.uri); }
		}, false);
	    }
	    let text = node.label;
	    link.appendChild(document.createTextNode(text));
	    let part = find_part(node.content);
	    if (part >= 0) {
		if (!partinfo[part])
		    partinfo[part] = [];
		partinfo[part].push(link);
	    }

	    if (node.children) {
		let child_div = document.createElementNS(XHTML_NS, "div");
		child_div.setAttribute("class", "navchild");
		hide_element(child_div, true);
		div.appendChild(child_div);
		for (let n in loop(node.children))
		    traverse_points(n, level + 1, child_div);
	    }
	} else {
	    if (level > 0)
		debug("Error: No label");
	    for (let n in loop(node.children))
		traverse_points(n, level, container);
	}
    }

    if (book.ncx && book.ncx.navmap) {
	let button = document.getElementById("content_button");
	button.hidden = false;
	let navmap = book.ncx.navmap;
	let box = document.getElementById("ve_toc");

	if (navmap.label) {
	    let div = document.createElementNS(XHTML_NS, "div");
	    div.setAttribute("class", "navheader");
	    div.appendChild(document.createTextNode(navmap.label));
	    box.appendChild(div);
	}

	traverse_points(navmap, 0, box);

	// FIXME Add preference for this?
	if (false) {
	    expand_all(true);
	} else if (false) {
	    collapse_all(true);
	} else {
	    show_all(true);
	}
    }

    // Contents sidebar FIXME
    function traverse_points_alt(node, treeparent) {
	let treeitem;
	debug("*: " + node.label + ";" + node.content);
	if (treeparent) {
	    treeitem = document.createElementNS(XUL_NS, "treeitem");
	    if (node.label && node.content) {
		let row = document.createElementNS(XUL_NS, "treerow");
		treeitem.appendChild(row);
		let cell = document.createElementNS(XUL_NS, "treecell");
		cell.setAttribute("label", node.label);
		row.appendChild(cell);
	    }
	    treeparent.appendChild(treeitem);
	}
	let treechildren;
	if (node.children) {
	    treechildren = document.createElementNS(XUL_NS, "treechildren");
	    if (treeitem) {
		treeitem.setAttribute("container", "true");
		treeitem.setAttribute("open", "true");
		treeitem.appendChild(treechildren);
	    }
	    for (let n in loop(node.children))
		traverse_points_alt(n, treechildren);
	}
	return treechildren;
    }
    if (false && book.ncx && book.ncx.navmap) {
	let button = document.getElementById("content_button");
	button.hidden = false;
	let navmap = book.ncx.navmap;
	let box = document.getElementById("ve_toc");

	if (navmap.label) {
	    let div = document.createElementNS(XHTML_NS, "div");
	    div.setAttribute("class", "navheader");
	    div.appendChild(document.createTextNode(navmap.label));
	    box.appendChild(div);
	}

	let tree = document.createElementNS(XUL_NS, "tree");
	tree.setAttribute("seltype", "single");
	//tree.setAttribute("onselect", "FIXME");
	tree.setAttribute("flex", "1");
	box.appendChild(tree);
	let treecols = document.createElementNS(XUL_NS, "treecols");
	//treecols.setAttribute("hidden", "true");
	//treecols.setAttribute("collapsed", "true");
	//treecols.setAttribute("style", "display: none");
	tree.appendChild(treecols);
	let treecol = document.createElementNS(XUL_NS, "treecol");
	treecol.setAttribute("primary", "true");
	treecol.setAttribute("flex", "1");
	treecols.appendChild(treecol);

	let childs = traverse_points_alt(navmap, null);
	tree.appendChild(childs);
    }

    // Secondary content sidebar
    if (book.ncx && book.ncx.navlists) {
	let button = document.getElementById("list_button");
	button.hidden = false;
	box = document.getElementById("ve_list");
	for (let navlist in loop(book.ncx.navlists)) {
	    if (navlist.label) {
		let div = document.createElementNS(XHTML_NS, "div");
		div.setAttribute("class", "navheader");
		div.appendChild(document.createTextNode(navlist.label));
		box.appendChild(div);
	    }
	    for (let node in loop(navlist.list)) {
		if (node.label && node.content) {
		    debug("$: " + node.label + ";" + node.content);
		    let div = document.createElementNS(XHTML_NS, "div");
		    div.setAttribute("class", "navitem");
		    box.appendChild(div);
		    let link = document.createElementNS(XHTML_NS, "a");
		    div.appendChild(link);
		    link.setAttribute("href", "#");
		    link.addEventListener("click", {
			uri: bookid + "/" + node.content,
			handleEvent: function(e) { view_uri(this.uri); }
		    }, false);
		    let text = node.label;
		    link.appendChild(document.createTextNode(text));
		}
	    }
	}
    }

    // Pages sidebar
    if (book.ncx && book.ncx.pagelist) {
	let button = document.getElementById("pages_button");
	button.hidden = false;
	box = document.getElementById("ve_pages");

	if (book.ncx.pagelist.label) {
	    let div = document.createElementNS(XHTML_NS, "div");
	    div.setAttribute("class", "navheader");
	    div.appendChild(document.createTextNode(book.ncx.pagelist.label));
	    box.appendChild(div);
	}
	for (let node in loop(book.ncx.pagelist.list)) {
	    if (node.label && node.content) {
		debug("$: " + node.label + ";" + node.content);
		let div = document.createElementNS(XHTML_NS, "div");
		div.setAttribute("class", "navitem");
		box.appendChild(div);
		let link = document.createElementNS(XHTML_NS, "a");
		div.appendChild(link);
		link.setAttribute("href", "#");
		link.addEventListener("click", {
		    uri: bookid + "/" + node.content,
		    handleEvent: function(e) { view_uri(this.uri); }
		}, false);
		let text = node.label;
		link.appendChild(document.createTextNode(text));
	    }
	}
    }

    // Guide sidebar
    if (book.opf.guide) {
	let button = document.getElementById("guide_button");
	button.hidden = false;
	box = document.getElementById("ve_guide");
	for (let item in loop(book.opf.guide)) {
	    let div = document.createElementNS(XHTML_NS, "div");
	    div.setAttribute("class", "navitem");
	    box.appendChild(div);
	    let link = document.createElementNS(XHTML_NS, "a");
	    div.appendChild(link);
	    if (item.broken || book.is_encrypted(item.href)) {
		link.setAttribute("style", "text-decoration: line-through");
	    } else {
		link.setAttribute("href", "#");
		link.addEventListener("click", {
		    uri: bookid + "/" + item.href,
		    handleEvent: function(e) { view_uri(this.uri); }
		}, false);
	    }
	    let text = item.title;

	    let type_text;
	    try {
		type_text = bundle.getString("type_" + item.type);
	    } catch (e) {
		if (item.type.substring(0, 6) == "other.")
		    type_text = bundle.getString("type_other");
		else
		    book.add_error("invalid guide item type: " + item.type);
	    }
	    //if (type_text)
	    //	link.setAttribute("tooltiptext", type_text);
	    if (!text) {
		if (type_text)
		    text = type_text;
		else if (item.type)
		    text = item.type;
		else
		    text = "";
	    } else if (type_text &&
		       (type_text.toLowerCase() != text.toLowerCase())) {
		text += " (" + type_text + ")";
	    }
	    link.appendChild(document.createTextNode(text));
	}
    }

    // Notes sidebar
    if (note_manager) {
	try {
	    update_target_parts();
	    list_notes();
	} catch (e) {
	    debug("annotation error: " + e);
	}
    }

    // Tools sidebar
    let show_tools = false;

    //enable_debug(true);

    debug("Original location: " + book.original_location)
    let uri2 = null;
    try {
	let ioService = Components.classes["@mozilla.org/network/io-service;1"].
	    getService(Components.interfaces.nsIIOService);
	uri2 = ioService.newURI(book.original_location, "UTF-8", null);
    } catch (e) {
    }

    /*
    if (uri2 && (uri2.schemeIs("http") || uri2.schemeIs("https"))) {
	let button = document.getElementById("epubcat");
	button.hidden = false;
	show_tools = true;
    }
    */
    debug("AUTOPUBS: " + book.opf.metadata.autopubs.length);
    if (book.opf.metadata.autopubs.length > 0) {
	let button = document.getElementById("autopub");
	button.hidden = false;
	show_tools = true;

	if (uri2 && !uri2.schemeIs("x-lucidor-convert")) {
	    let text = bundle.getString("updateQuestion");
	    let yestext = sharedbundle.getString("yes");
	    let notext = sharedbundle.getString("no");
	    let nb = document.getElementById("notification_box");
	    let no = nb.appendNotification(text, "upd", null,
					   nb.PRIORITY_INFO_MEDIUM,
					   [{ accessKey: null,
					      callback: update_yes,
					      label: yestext,
					      popup: null },
					    { accessKey: null,
					      callback: close_callback,
					      label: notext,
					      popup: null }], null);
	    autohide(no, true);
	}
    }
    if (is_extension) {
	let button = document.getElementById("save_button");
	button.hidden = false;
	show_tools = true;
    }
    if (book.rootfiles.length > 1) {
	let box = document.getElementById("rootfiles");
	let rfn = 0;
	for (let rf in loop(book.rootfiles)) {
	    let div = document.createElementNS(XHTML_NS, "div");
	    div.setAttribute("class", "navitem");
	    box.appendChild(div);
	    let link = document.createElementNS(XHTML_NS, "a");
	    div.appendChild(link);
	    link.appendChild(document.createTextNode("Open rendition: " + rf));
	    link.setAttribute("href", "#");
	    link.addEventListener("click", {
		rf: rfn,
		handleEvent: function(e) {
		    main_control.add_epub_tab(uri, this.rf);
		}
	    }, false);
	    rfn++;
	}
	box.hidden = false;
	show_tools = true;
    }

    if (show_tools) {
	let button = document.getElementById("tools_button");
	button.hidden = false;
    }

    let mlist = document.getElementById("sidelist");
    for (let cld in loop(mlist.getElementsByTagName("menuitem")))
	if (!cld.hidden) {
	    cld.doCommand();
	    break;
	}

    if (!show_last_position()) {
	let part = book.get_first_part();
	if (part >= 0)
	    view_uri_x(bookid + "/" + spine[part].href, spine[part].stylesheet,
		       0, null);
	else
	    book.add_error("has no content");
    }
    setTimeout(function() {
	if (hide_toolbar_ok) {
	    let toolbar = document.getElementById("toolbar");
	    if (!toolbar.hidden)
		toggle_toolbar();
	}
    }, 15000);

    update_date_read();
    return null;
}

var hide_toolbar_ok = true;

function hide_element(element, fast) {
    /*
    if (fast)
	$(element).hide("fast");
    else
	$(element).hide("slow");
    */
    //element.style.display = "none";
    element.hidden = true;
}

function show_element(element, fast) {
    /*
    if (fast)
	$(element).show("fast");
    else
	$(element).show("slow");
    */
    //element.style.display = "inline";
    element.hidden = false;
}

// ⊞⊟﹢﹣▸▾▹▿
function set_as_open(element) {
    if (element.firstChild)
	element.removeChild(element.firstChild);
    if (element.firstChild)
	element.insertBefore(document.createTextNode("▿"), element.firstChild);
    else
	element.appendChild(document.createTextNode("▿"));
}

function set_as_closed(element) {
    if (element.firstChild)
	element.removeChild(element.firstChild);
    if (element.firstChild)
	element.insertBefore(document.createTextNode("▹"), element.firstChild);
    else
	element.appendChild(document.createTextNode("▹"));
}

function toggle_it(link) {
    let child_div = link.parentNode.parentNode.lastChild;
    //if (child_div.style.display == "none") {
    if (child_div.hidden) {
	show_element(child_div, true);
	set_as_open(link);
    } else {
	hide_element(child_div, true);
	set_as_closed(link);
    }
}

// Default Contents state: Expanded, Collapsed, Show all
function expand_all(imm) {
    let box = document.getElementById("ve_toc");
    /*
    $("div:has(div.navchild) > span:first-child > a:first-child", box).text("▿");
    $("div:has(div.navchild) > span:first-child", box).show();
    if (imm)
	$(".navchild", box).show();
    else
	$(".navchild", box).show("slow");
    */

    /*
      <div.navitem>
      <span><a>▹</a> </span><a>...</a>
      <div.navchild>
      ...
    */
    let xpe = new XPathEvaluator();
    // 6 = UNORDERED_NODE_SNAPSHOT_TYPE
    let result = xpe.evaluate("//*[@class='navitem']", box,
			      xpe.createNSResolver(document), 6, null);
    for (let i = 0; i < result.snapshotLength; i++) {
	debug("EXPAND: " + i);
	let node = result.snapshotItem(i);
	//node.hidden = false;
	if (node.firstChild.localName == "span") {
	    node.firstChild.hidden = false;
	    set_as_open(node.firstChild.firstChild);
	}
    }

    result = xpe.evaluate("//*[@class='navchild']", box,
			  xpe.createNSResolver(document), 6, null);
    for (let i = 0; i < result.snapshotLength; i++) {
	let node = result.snapshotItem(i);
	node.hidden = false;
    }
}

function collapse_all(imm) {
    let box = document.getElementById("ve_toc");
    /*
    $("div:has(div.navchild) > span:first-child > a:first-child", box).text("▹");
    $("div:has(div.navchild) > span:first-child", box).show();
    if (imm)
	$(".navchild", box).hide();
    else
	$(".navchild", box).hide("slow");
    */
    let xpe = new XPathEvaluator();
    // 6 = UNORDERED_NODE_SNAPSHOT_TYPE
    let result = xpe.evaluate("//*[@class='navitem']", box,
			      xpe.createNSResolver(document), 6, null);
    for (let i = 0; i < result.snapshotLength; i++) {
	let node = result.snapshotItem(i);
	//node.hidden = false;
	if (node.firstChild.localName == "span") {
	    node.firstChild.hidden = false;
	    set_as_closed(node.firstChild.firstChild);
	}
    }

    result = xpe.evaluate("//*[@class='navchild']", box,
			  xpe.createNSResolver(document), 6, null);
    for (let i = 0; i < result.snapshotLength; i++) {
	let node = result.snapshotItem(i);
	node.hidden = true;
    }
}

function show_all(imm) {
    let box = document.getElementById("ve_toc");
    /*
    $("div:has(div.navchild) > span:first-child", box).hide();
    if (imm)
	$(".navchild", box).show();
    else
	$(".navchild", box).show("slow");
    */
    let xpe = new XPathEvaluator();
    // 6 = UNORDERED_NODE_SNAPSHOT_TYPE
    let result = xpe.evaluate("//*[*/@class='navchild']", box,
			      xpe.createNSResolver(document), 6, null);
    for (let i = 0; i < result.snapshotLength; i++) {
	let node = result.snapshotItem(i);
	//node.hidden = false;
	if (node.firstChild.localName == "span") {
	    node.firstChild.hidden = true;
	}
    }

    result = xpe.evaluate("//*[@class='navchild']", box,
			  xpe.createNSResolver(document), 6, null);
    for (let i = 0; i < result.snapshotLength; i++) {
	let node = result.snapshotItem(i);
	node.hidden = false;
    }
}

function show_last_position() {
    try {
	let bookcase = null;
	if (is_extension) {
	    bookcase = Components.classes["@lucidor.org/bookcaseLiteService;1"].
		getService().wrappedJSObject;
	} else {
	    bookcase = Components.classes["@lucidor.org/bookcaseSqlService;1"].
		getService().wrappedJSObject;
	}
	bookcase.open();
	let id = bookcase.get_id_from_sha1(book.sha1);
	let bm_text = bookcase.get_last_position(id);
	if (bm_text) {
	    let bm = JSON.parse(bm_text);
	    if (bm.version == 1) {
		let part = find_part(bm.href);
		if (part >= 0) {
		    view_uri_x(bookid + "/" + spine[part].href,
			       spine[part].stylesheet, bm.scroll, null);
		    return true;
		}
	    }
	}
    } catch (e) {
    }

    return false;
}

function get_current_position() {
    let bm = { version: 1,
	       href: href_to_local_href(browser.currentURI.spec) };

    let wind = browser.contentWindow;
    let doc = browser.contentDocument;
    if (false && doc.caretPositionFromPoint) {
	let caret = doc.caretPositionFromPoint(wind.scrollX, wind.scrollY);
	if (caret) {
	    let range = doc.createRange();
	    range.setStart(caret.offsetNode, caret.offset);
	    range.collapse(true);
	    let fixptr = domrange_to_fixptr(range);
	    bm.fixptr = fixptr;
	}
    }

    // Scroll ratio
    if (view_mode == "columns")
	bm.scroll = (1.0 * wind.scrollX) / wind.scrollMaxX;
    else
	bm.scroll = (1.0 * wind.scrollY) / wind.scrollMaxY;

    return JSON.stringify(bm);
}

function retry() {
    document.getElementById("retry_link").hidden = true;
    //webNavigation.reload(LOAD_FLAGS_NONE);
    init1();
}

function set_open_message(text) {
    let msg = document.getElementById("open_message");
    for (let ch in loop(msg.childNodes))
	msg.removeChild(ch);
    msg.appendChild(document.createTextNode(text));
}

function set_status_message(text) {
    let msg = document.getElementById("status_message");
    for (let ch in loop(msg.childNodes))
	msg.removeChild(ch);
    msg.appendChild(document.createTextNode(text));
}

function error_opening(error) {
    let text = "";

    let bundle = document.getElementById("lucifox-strings");
    if (error == 1) {
	// Retrieving file failed
	text = bundle.getString("retrieve_error");
	document.getElementById("retry_link").hidden = false;
    } else if (error == 2) {
	// Opening EPUB file failed
	text = bundle.getString("open_error");
	document.getElementById("save_link").hidden = false;
    } else {
	// Undetermined failure
	text = bundle.getString("other_error");
    }
    set_open_message(text);
    document.getElementById("load_box").hidden = true;
}

function answer_yes(nf, button) {
    try_copy_and_add_book(null, book, book_file);
    book_file = null;

    update_last_date_and_position();
    return false;
}

function update_yes(nf, button) {
    update_autopub();
    return false;
}

function close_callback(nf, button) {
    return false;
}

function show_details() {
    let text = "";
    for (let w in loop(book.errors))
	text += "• " + w + "\n";
    let prompts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
    let bundle = document.getElementById("lucifox-strings");
    let title = bundle.getString("errors");
    prompts.alert(window, title, text);
}

function show_source() {
    main_control.add_web_tab("view-source:" + browser.currentURI.spec);
}

function lookup_word() {
    let prefs = Components.classes["@mozilla.org/preferences-service;1"].
	getService(Components.interfaces.nsIPrefService).
	getBranch("extensions.lucifox.");
    let uri = prefs.getCharPref("viewer.lookup");

    let sel = browser.contentWindow.getSelection();
    let word = sel.toString();
    word = encodeURIComponent(word);
    uri = uri.replace("%%", word);
    uri = uri.replace("{W}", word);
    uri = uri.replace("{w}", word);

    let lang = "en";
    try {
	lang = book.opf.metadata.langcode.language;
    } catch (e) {
    }
    uri = uri.replace("{L}", lang);
    uri = uri.replace("{l}", lang);

    main_control.add_web_tab(uri);
}

function copy_text() {
    let clipboard = Components.classes["@mozilla.org/widget/clipboardhelper;1"].
	getService(Components.interfaces.nsIClipboardHelper);
    let sel = browser.contentWindow.getSelection();
    clipboard.copyString(sel.toString());

    //command="cmd_copy"
    //webshell.copySelection();
}

function contextmenu_onpopup() {
    let sel = browser.contentWindow.getSelection();
    if (!(sel.toString().length < 1)) {
	let item = document.getElementById("lookup_item");
	item.hidden = false;
	if (sel.toString().length > 0)
	    item.disabled = false;
	else
	    item.disabled = true;

	item = document.getElementById("copy_item");
	item.hidden = false;
	//if (webshell.canCopySelection())
	if (sel.toString().length > 0)
	    item.disabled = false;
	else
	    item.disabled = true;

	if (note_manager) {
	    item = document.getElementById("create_note_item");
	    item.hidden = false;
	    item.disabled = false;

	    item = document.getElementById("create_highlight_item");
	    item.hidden = false;
	    item.disabled = false;
	}

	item = document.getElementById("next_item");
	item.hidden = true;
	item = document.getElementById("previous_item");
	item.hidden = true;
    } else {
	let item = document.getElementById("next_item");
	item.hidden = false;
	item = document.getElementById("previous_item");
	item.hidden = false;

	item = document.getElementById("copy_item");
	item.hidden = true;
	item = document.getElementById("create_note_item");
	item.hidden = true;
	item = document.getElementById("create_highlight_item");
	item.hidden = true;
	item = document.getElementById("lookup_item");
	item.hidden = true;
    }
}

function update_autopub() {
    let location = "x-lucidor-convert://feed/?";
    for (let pub in loop(book.opf.metadata.autopubs))
	location += "uri=" + encodeURIComponent(pub) + "&";
    location += "title=" + encodeURIComponent(book.opf.metadata.title);
    debug("update_autopub: " + location);
    main_control.add_epub_tab(location, 0);
}

function edit_autopub() {
    let location = "chrome://lucidor/content/autopub.xul?";
    for (let pub in loop(book.opf.metadata.autopubs))
	location += "uri=" + encodeURIComponent(pub) + "&";
    location += "title=" + encodeURIComponent(book.opf.metadata.title);
    main_control.add_tab(location, 0);
}

function send_to_epubcatalog() {
    let url = "http://www.epubcatalog.org/epubcat/addcat";
    let dataString = "url=" + book.original_location;
    main_control.add_web_tab_post(url, dataString);
}

function try_copy_and_add_book(bookcase, book, file) {
    if (is_extension)
	return;

    if (!bookcase) {
	bookcase = Components.classes["@lucidor.org/bookcaseSqlService;1"].
	    getService().wrappedJSObject;
	bookcase.open();
    }

    try {
	bookcase.add_open_book(book, file);
    } catch (e) {
	// FIXME make more detailed
	debug(e);
	let bundle = document.getElementById("lucifox-strings");
	let text = sharedbundle.getString("add_error");
	let nb = document.getElementById("notification_box");
	nb.appendNotification(text, "add_err", null,
			      nb.PRIORITY_CRITICAL_MEDIUM, [], null);
    }
}

function get_browser() {
    return browser;
    //return document.getElementById("epub_browser");
}

function getMarkupDocumentViewer() {
    return browser.markupDocumentViewer;
    //return document.getElementById("epub_browser").markupDocumentViewer;
}

function get_save_file() {
    return book.zipreader.file;
}

function get_save_name() {
    return book.get_filename();
}

function switch_deck(i) {
    document.getElementById("ve_deck").selectedIndex = i;
}

function view_part(part, endscroll, focus = true) {
    // FIXME Add pref for endscroll
    let es = 0;
    if (endscroll)
	es = 1;
    view_uri_x(bookid + "/" + spine[part].href, spine[part].stylesheet,
	       es, null, focus);
}

function view_uri(location) {
    view_uri_x(location, null, 0, null, true);
}

function view_uri_x(location, stylesheet, endscroll, targetscroll,
		    focus = false) {
    debug("view_uri_x: " + location);

    if (location.indexOf("x-lucidor://") == 0)
	location = location.substring(12);

    let loca = location.split("#");
    if (loca.length == 2)
	location = encodeURI(loca[0]) + "#" +  loca[1];
    else
	location = encodeURI(location);
    if (stylesheet)
	stylesheet = encodeURI(stylesheet);

    let tmploc = location.substring(bookid.length + 1);
    //let item = book.opf.manifest_href[tmploc];
    let item = book.hrefToItem(tmploc);
    if (item &&	item.encryption) {
	let bundle = document.getElementById("lucifox-strings");
	let text = bundle.getString("encrypted_error");
	let nb = document.getElementById("notification_box");
	nb.appendNotification(text, "enc_err", null,
			      nb.PRIORITY_WARNING_MEDIUM, [], null);
	//update_buttons(-1);
	update_col_buttons();
	return;
    }

    location = "x-lucidor://" + location;

    debug("old: (" + browser.currentURI.spec + ")");

    g_scroll_location = location;
    g_scroll_ratio = endscroll;
    g_scroll_target = targetscroll;
    g_extra_stylesheet = stylesheet;

    try {
	debug("Location: " + location);
	browser.loadURI(location, null, null);
    } catch (e) {
	let bundle = document.getElementById("lucifox-strings");
	let text = bundle.getString("load_error");
	let nb = document.getElementById("notification_box");
	nb.appendNotification(text, "load_err", null,
			      nb.PRIORITY_CRITICAL_MEDIUM, [], null);
	//update_buttons(-1);
	update_col_buttons();
    }

    if (focus)
	document.getElementById("epub_browser").focus();
}

function reload_page() {
    debug("reload_page");
    let path = bookid + "/" + get_part_path();
    let pos = JSON.parse(get_current_position());
    view_uri_x(path, null, pos.scroll, null);
}

function href_to_local_href(href) {
    if (href.indexOf("x-lucidor://") == 0)
	href = href.substring(12);
    // FIXME there is a rare bug here: a name of a folder may be equal to bookid
    if (href.indexOf(bookid + "/") == 0)
	href = href.substring(bookid.length + 1);
    let i = href.indexOf("#");
    if (i >= 0)
	href = href.substring(0, i);
    return href;
}

function get_manifest_item() {
    return book.hrefToItem(href_to_local_href(browser.currentURI.spec));
}

function get_spine_item() {
    let n = book.hrefToPart(href_to_local_href(browser.currentURI.spec));
    if (n >= 0)
	return book.opf.spine[n];
    return null;
}

function find_part(href) {
    return find_part_x(href);
    //return book.hrefToPart(href_to_local_href(decodeURI(href)));
}

// Like find_part, but also handles Intra-Publication EPUB CFIs
function find_part_x(href) {
    //debug("find_part_x: " + href);
    let nhref = href_to_local_href(decodeURI(href));
    debug("find_part_x: " + nhref + " " + book.rootfiles[book.rootfile]);
    if (nhref == book.rootfiles[book.rootfile]) {
	let sp = href.split("#");
	if (sp.length == 2)
	    nhref = epubcfi_to_part(sp[1]);
    }
    return book.hrefToPart(nhref);
}

function show_next_part(focus = true) {
    let part = find_part(browser.currentURI.spec);
    let next = book.get_next_part(part);
    if (next < 0)
	return;
    view_part(next, false, focus);
}

function show_previous_part(focus = true) {
    let part = find_part(browser.currentURI.spec);
    let prev = book.get_previous_part(part);
    if (prev < 0)
	return;
    view_part(prev, true, focus);
}

function go_back() {
    browser.goBack();
}

function go_forward() {
    browser.goForward();
}

function update_buttons(part) {
    /*if (part < 0)
	part = find_part(location.spec);*/
    let linear = false;
    if (part >= 0 && spine[part])
	linear = spine[part].linear;

    let button = document.getElementById("prev_button");
    let item = document.getElementById("previous_item");
    if (linear && (book.get_previous_part(part) >= 0)) {
	button.disabled = false;
	item.disabled = false;
    } else {
	button.disabled = true;
	item.disabled = true;
    }

    button = document.getElementById("next_button");
    item = document.getElementById("next_item");
    if (linear && (book.get_next_part(part) >= 0)) {
	button.disabled = false;
	item.disabled = false;
    } else {
	button.disabled = true;
	item.disabled = true;
    }
}

function update_col_buttons() {
    let val = !(view_mode == "columns");
    let button = document.getElementById("prev_col_button");
    button.hidden = val;
    button = document.getElementById("next_col_button");
    button.hidden = val;
}

function update_navigation_bar() {
    if (is_extension)
	return;

    let buttonbar = document.getElementById("nav_buttons");
    let item = document.getElementById("navbar_item");
    let hideitem = document.getElementById("hide_navbar_item");
    let margin = 20;
    if (!buttonbar.hidden)
	margin += 20;
    if (!item.hasAttribute("checked")) {
	buttonbar.hidden = true;
    } else if (item.hasAttribute("checked") &&
	       !hideitem.hasAttribute("checked")) {
	buttonbar.hidden = false;
    } else if (item.hasAttribute("checked") &&
	       hideitem.hasAttribute("checked") &&
	       ((browser.contentWindow.scrollY < margin) ||
		(browser.contentWindow.scrollY >
		 (browser.contentWindow.scrollMaxY - margin)))) {
	buttonbar.hidden = false;
    } else {
	buttonbar.hidden = true;
    }
}

function update_visibility() {
    let prefs = Components.classes["@mozilla.org/preferences-service;1"].
	getService(Components.interfaces.nsIPrefService).
	getBranch("extensions.lucifox.");
    set_extra_bar(prefs.getBoolPref("viewer.enableExtraTools"));
    set_navigation_bar(prefs.getBoolPref("viewer.enableNavigationBar"));
}

function toggle_extra_bar() {
    let enable = document.getElementById("extra_checkbox").hasAttribute("checked");

    let prefs = Components.classes["@mozilla.org/preferences-service;1"].
	getService(Components.interfaces.nsIPrefService).
	getBranch("extensions.lucifox.");
    prefs.setBoolPref("viewer.enableExtraTools", enable);
}

function set_extra_bar(enable) {
    let item = document.getElementById("extra_checkbox");
    if (enable) {
	item.setAttribute("checked", "true");
	item.setAttribute("aria-pressed", "true");
    } else {
	item.removeAttribute("checked");
	item.setAttribute("aria-pressed", "false");
    }

    document.getElementById("extra_bar").hidden = !enable;
}

function toggle_navigation_bar() {
    let enable = document.getElementById("navbar_item").hasAttribute("checked");

    let prefs = Components.classes["@mozilla.org/preferences-service;1"].
	getService(Components.interfaces.nsIPrefService).
	getBranch("extensions.lucifox.");
    prefs.setBoolPref("viewer.enableNavigationBar", enable);
}

function set_navigation_bar(enable) {
    let item = document.getElementById("navbar_item");
    if (enable)
	item.setAttribute("checked", "true");
    else
	item.removeAttribute("checked");

    if (is_extension)
	document.getElementById("nav_buttons").hidden = !enable;
    else
	update_navigation_bar();
}

function toggle_sidebar() {
    let sidebar = document.getElementById("mysidebar");
    let splitter = document.getElementById("epub_splitter");
    let val = !sidebar.hidden;
    sidebar.hidden = val;
    splitter.hidden = val;

    let item = document.getElementById("sidebar_item");
    if (val)
	item.removeAttribute("checked");
    else
	item.setAttribute("checked", "true");
}

function toggle_toolbar() {
    hide_toolbar_ok = false;
    let toolbar = document.getElementById("toolbar");
    let val = !toolbar.hidden;
    toolbar.hidden = val;

    let item = document.getElementById("toolbar_item");
    if (val)
	item.removeAttribute("checked");
    else
	item.setAttribute("checked", "true");
}

function toggle_tabbar() {
    let val = main_control.toggle_tabbar();
    let item = document.getElementById("tabbar_item");
    if (val)
	item.removeAttribute("checked");
    else
	item.setAttribute("checked", "true");
}

function set_view_mode(mode) {
    if (mode)
	view_mode = mode;
    let root = browser.contentDocument.documentElement;
    let glue = root.firstChild;
    let wbox = document.getElementById("w-box");
    let hbox = document.getElementById("h-box");
    let autosize = document.getElementById("auto-button");

    let defclass = "";
    if (view_border)
	defclass = "border ";

    if (view_mode == "columns") {
	root.setAttribute("class", defclass + "columns");
	glue.setAttribute("class", defclass + "columns");
	wbox.disabled = false;
	hbox.disabled = false;
	autosize.disabled = false;
	document.getElementById("columns_button").setAttribute("checked", "true");
	document.getElementById("flow_button").removeAttribute("checked");
	document.getElementById("fixed_button").removeAttribute("checked");
    } else if (view_mode == "flow") {
	root.setAttribute("class", defclass + "flow");
	root.style.width = "";
	glue.setAttribute("class", defclass + "flow");
	glue.style.height = "";
	glue.style.MozColumnWidth = "";
	wbox.disabled = true;
	hbox.disabled = true;
	autosize.disabled = true;
	document.getElementById("columns_button").removeAttribute("checked");
	document.getElementById("flow_button").setAttribute("checked", "true");
	document.getElementById("fixed_button").removeAttribute("checked");
    } else {	// view_mode == "fixed"
	root.setAttribute("class", defclass + "fixed");
	glue.setAttribute("class", defclass + "fixed");
	glue.style.height = "";
	glue.style.MozColumnWidth = "";
	wbox.disabled = false;
	hbox.disabled = true;
	autosize.disabled = false;
	document.getElementById("columns_button").removeAttribute("checked");
	document.getElementById("flow_button").removeAttribute("checked");
	document.getElementById("fixed_button").setAttribute("checked", "true");
    }
    update_col_buttons();
    size_change();
}

function w_scroll(event) {
    let textbox = document.getElementById("w-box");
    if (event.detail < 0)
	textbox.increase();
    else
	textbox.decrease();
    size_change();
}

function h_scroll(event) {
    let textbox = document.getElementById("h-box");
    if (event.detail < 0)
	textbox.increase();
    else
	textbox.decrease();
    size_change();
}

function auto_size() {
    let val = browser.boxObject.width * 10 / ppdm - 1;
    // FIXME make two columns visible at once if enough space is available?
    if (view_mode == "columns" && val > 30)
	val = val / 2 - 1;
    //val = Math.min(val, 30);
    let textbox = document.getElementById("w-box");
    textbox.value = val;

    val = browser.boxObject.height * 10 / ppdm - 2;
    textbox = document.getElementById("h-box");
    textbox.value = val;

    size_change();
}

function size_change() {
    if (view_mode == "flow")
	return;

    let prefs = Components.classes["@mozilla.org/preferences-service;1"].
	getService(Components.interfaces.nsIPrefService).
	getBranch("extensions.lucifox.");

    let textbox = document.getElementById("w-box");
    let width = textbox.value + "cm";
    let root = browser.contentDocument.documentElement;
    root.style.width = width;
    prefs.setCharPref("viewer.columnWidth", textbox.value);

    textbox = document.getElementById("h-box");
    prefs.setCharPref("viewer.columnHeight", textbox.value);
    if (view_mode == "columns") {
	let height = textbox.value + "cm";
	let glue = root.firstChild;
	glue.style.MozColumnWidth = width;
	glue.style.height = height;
    }
}

function mouse_scroll(event) {
    //debug("::: " + event.detail);
    if (event.timeStamp > scroll_time + 600) {
	scroll_time = event.timeStamp;
	scroll_count = 0;
	return;
    }

    /*
    if (scroll_count == -1) {
	debug("DDD");
	event.preventDefault();
	return;
    }
    */

    if (event.detail > 0 && browser.contentWindow.scrollY >
	(browser.contentWindow.scrollMaxY - 2)) {
	scroll_count++;
	scroll_time = event.timeStamp;
	if (scroll_count > 8) {
	    scroll_count = -1;
	    //event.preventDefault();
	    if (view_mode != "columns" || browser.contentWindow.scrollX >
		(browser.contentWindow.scrollMaxX - 2)) {
		show_next_part();
		// This is to work around the problem that you cannot continue scrolling without moving mouse after changing part
		let utils = window.QueryInterface(Components.interfaces.nsIInterfaceRequestor).getInterface(Components.interfaces.nsIDOMWindowUtils);
		utils.sendMouseEvent("mousemove", 1, 0, 0, 0, 0);
		utils.sendMouseEvent("mousemove", -1, 0, 0, 0, 0);
	    } else	// view_mode == columns && not at right end
		next_column();
	    // FIXME must wait until part is loaded before scrolling
	    //browser.contentWindow.scrollY = 30;
	}
    } else if (browser.contentWindow.scrollY == 0 && event.detail < 0) {
	scroll_count++;
	scroll_time = event.timeStamp;
	if (scroll_count > 8) {
	    scroll_count = -1;
	    //event.preventDefault();
	    if (view_mode != "columns" || browser.contentWindow.scrollX == 0) {
		show_previous_part();
		// This is to work around the problem that you cannot continue scrolling without moving mouse after changing part
		let utils = window.QueryInterface(Components.interfaces.nsIInterfaceRequestor).getInterface(Components.interfaces.nsIDOMWindowUtils);
		utils.sendMouseEvent("mousemove", 1, 0, 0, 0, 0);
		utils.sendMouseEvent("mousemove", -1, 0, 0, 0, 0);
	    } else	// view_mode == columns && not at left end
		previous_column();
	}
    } else {
	scroll_count = 0;
    }
}

function next_column() {
    change_column(1);
}

function previous_column() {
    change_column(-1);
}

function change_column(dir) {
    if (view_mode != "columns")
	return;

    let textbox = document.getElementById("w-box");
    let w = browser.contentWindow.innerWidth + browser.contentWindow.scrollMaxX;
    let cw = textbox.value * ppdm * 0.1;
    let margin = 0.4 * ppdm * 0.1;	// FIXME not exact
    cw += margin;
    let cols = Math.floor(w / cw);
    let rcw = w / cols;
    debug(w + " " + cw + " " + cols + " " + rcw + " " + margin + " " + w / cw);

    if (column_adjustment == "none") {
	browser.contentWindow.scrollBy(dir * rcw,
				       dir * -browser.contentWindow.scrollMaxY);
	return;
    }

    let cx = browser.contentWindow.scrollX;
    let offset = 0;
    if (column_adjustment == "center" &&
	browser.contentWindow.innerWidth > rcw) {
	offset = (browser.contentWindow.innerWidth - rcw) / 2 - margin / 2;
	cx += offset;
	//debug("OFFSET: " + offset);
    }

    let current_col = Math.floor(cx / rcw);
    if (current_col < 0)	// This should not happen, but it does
	current_col = 0;
    if (dir == 1) {
	if ((cx % rcw) / rcw > 0.9)
	    current_col++;
    } else {
	if ((cx % rcw) / rcw > 0.1)
	    current_col++;
    }

    let next = (current_col + dir) * rcw - offset;
    debug(current_col + " " + next);

    if (dir == 1)
	browser.contentWindow.scrollTo(next, 0);
    else
	browser.contentWindow.scrollTo(next, browser.contentWindow.scrollMaxY);
}

function open_annotations() {
    if (document.getElementById("mysidebar").hidden)
	toggle_sidebar();
    document.getElementById("annotations_button").doCommand();
}

function create_new_note() {
    open_annotations();
    new_note();
    document.getElementById("note_text").focus();
}

function new_note() {
    document.getElementById("create_note_button").hidden = true;
    document.getElementById("new_note").hidden = false;
}

function get_note_position(note) {
    return null;

    /*
    let ios = Components.classes["@mozilla.org/network/io-service;1"].
	getService(Components.interfaces.nsIIOService);
    let uri = ios.newURI(doc.location, "UTF-8", null).
	QueryInterface(Components.interfaces.nsIURL);
    let part = uri.filePath.substring(1);
    let poss = note_manager.get_part_positions(book.sha1, book.rootfile, part);
    if (poss.length < 1)
	return 1000;
    let prev = -1;
    let next = -1;
    let i = Marh.floor(poss.length / 2);
    while (true) {
	;
    }
    */
}

function get_part_path() {
    let doc = browser.contentDocument;
    let ios = Components.classes["@mozilla.org/network/io-service;1"].
	getService(Components.interfaces.nsIIOService);
    let uri = ios.newURI(doc.location, "UTF-8", null).
	QueryInterface(Components.interfaces.nsIURL);
    let path = uri.filePath.substring(1);
    return path;
}

function selection_to_target() {
    let sel = browser.contentWindow.getSelection();
    if (sel.rangeCount == 0)
	return null;

    let range = sel.getRangeAt(0);

    if (is_luci_node(range.startContainer))
	range.setStartBefore(range.startContainer);
    else if (is_luci_node(range.startContainer.parentNode))
	range.setStartBefore(range.startContainer.parentNode);
    if (is_luci_node(range.endContainer))
	range.setEndBefore(range.endContainer);
    else if (is_luci_node(range.endContainer.parentNode))
	range.setEndBefore(range.endContainer.parentNode);

    debug("Range: " + range.startContainer + " " +
	  range.startOffset + " " + range.endOffset);
    let target = { part: get_part_path(),
		   endpart: null };
    let doc = range.startContainer.ownerDocument;
    let root = get_root(doc);
    fill_target(target, root, range);

    /*
    debug("toString=" + range.toString());
    debug("pretext: " + target.pretext);
    debug("text: " + target.text);
    debug("endtext: " + target.endtext);
    debug("posttext: " + target.posttext);
    */

    return target;
}

/*
function next_node(node, root) {
    if (!is_luci_node(node) && node.firstChild) {
	return node.firstChild;
    }
    if (node.nextSibling)
	return node.nextSibling;
    for (let n = node.parentNode; n != root; n = n.parentNode)
	if (n.nextSibling)
	    return n.nextSibling;
    return null;
}
*/

// end = -1 means "to the end"
function node_to_string(node, start, end) {
    if (node.nodeType == node.TEXT_NODE) {
	if (end >= 0)
	    return node.nodeValue.substring(start, end);
	return node.nodeValue.substring(start);
    }

    if (node.nodeType == node.ELEMENT_NODE) {
	if (is_luci_node(node))
	    return "";

	let s = "";
	if (end < 0)
	    end = node.childNodes.length;
	for (let i = start; i < end; i++)
	    s += node_to_string(node.childNodes[i], 0, -1);
	return s;
    }

    return "";
}

function trim_end(str, len) {
    let i = str.lastIndexOf(" ");
    if (i >= 0 && i >= (str.length - len))
	return str.substring(0, i);
    return str;
}

function trim_start(str, len) {
    let i = str.indexOf(" ");
    if (i >= 0 && i < len)
	return str.substring(i + 1);
    return str;
}

function collapse_whitespace(str) {
    return str.replace(/[ \t\n\r]+/g, " ");
}

function collapse_whitespace_strong(str) {
    return str.replace(/\s+/g, " ");
}

// Keep newlines
function collapse_whitespace_x(str) {
    str = str.replace(/\s*\n\s*/g, "\n");
    return str.replace(/[ \f\r\t\v\u00a0\u2028\u2029]+/g, " ");
}

function get_posttext_x(range, max) {
    let doc = range.startContainer.ownerDocument;
    let root = get_root(doc);

    // Do not start in luci_node
    for (let n = range.endContainer; n != root; n = n.parentNode) {
	if (is_luci_node(n)) {
	    range.setEndAfter(n);
	    break;
	}
    }
    return range_to_string({ startContainer: range.endContainer,
			     startOffset: range.endOffset,
			     endContainer: null,
			     endOffset: null }, max);
}

function get_rangetext(range, max) {
    let doc = range.startContainer.ownerDocument;
    let root = get_root(doc);

    // Do not start or end in luci_node
    for (let n = range.startContainer; n != root; n = n.parentNode) {
	if (is_luci_node(n)) {
	    range.setStartAfter(n);
	    break;
	}
    }
    for (let n = range.endContainer; n != root; n = n.parentNode) {
	if (is_luci_node(n)) {
	    range.setEndBefore(n);
	    break;
	}
    }

    //return range.toString();
    return range_to_string(range, max);
}

function append_string(str, val, space) {
    if (space && (str == "" || !is_ws[str[str.length - 1]]))
	str += " ";

    //debug(";" + str + ";" + val + ";");
    let text = collapse_whitespace(val);
    //debug("(" + text + ")");
    if (str && is_ws[str[str.length - 1]] && text && is_ws[text[0]])
	return str + text.substring(1);
    return str + text;
}

function range_to_string(range, max) {
    let doc = range.startContainer.ownerDocument;
    let root = get_root(doc);

    // Single node
    if (range.startContainer == range.endContainer) {
	let s = node_to_string(range.startContainer,
			       range.startOffset, range.endOffset);
	s = collapse_whitespace(s);
	return s.substring(0, max);
    }

    // Start node
    let s = "";
    let node = range.startContainer;
    if (node.nodeType == node.TEXT_NODE)
	s = collapse_whitespace(node_to_string(node, range.startOffset, -1));

    // Middle nodes
    let space = false;
    while (node) {
	debug("node: " + node.localName);
	// Find next node
	if (node.nodeType == node.ELEMENT_NODE) {
	    if (!space && (is_blocklike(node) || is_br(node)))
		space = true;
	}

	if (node.nodeType == node.ELEMENT_NODE && node.firstChild &&
	    !is_luci_node(node)) {
	    node = node.firstChild;
	} else if (node.nextSibling) {
	    node = node.nextSibling;
	} else {
	    for (node = node.parentNode; node != root; node = node.parentNode) {
		if (!space && is_blocklike(node))
		    space = true;
		if (node == range.endContainer)
		    return s.substring(0, max);
		if (node.nextSibling) {
		    node = node.nextSibling;
		    break;
		}
	    }
	}

	if (node == range.endContainer || node == root)
	    break;

	if (node.parentNode == range.endContainer &&
	    node == range.endContainer.childNodes[range.endOffset])
	    return s.substring(0, max);

	// Check node
	if (node.nodeType == node.TEXT_NODE) {
	    s = append_string(s, node.nodeValue, space);
	    space = false;
	    if (max > -1 && s.length >= max)
		return s.substring(0, max);
	}
    }

    // End node
    if (range.endContainer) {
	s = append_string(s, node_to_string(range.endContainer, 0,
					    range.endOffset), space);
    }

    return s.substring(0, max);
}

function get_pretext_x(range, max) {
    let doc = range.startContainer.ownerDocument;
    let root = get_root(doc);

    // Do not end in luci_node
    for (let n = range.startContainer; n != root; n = n.parentNode) {
	if (is_luci_node(n)) {
	    range.setStartBefore(n);
	    break;
	}
    }
    return range_to_endstring({ startContainer: null,
				startOffset: null,
				endContainer: range.startContainer,
				endOffset: range.startOffset }, max);
}

function get_endtext(range, max) {
    let doc = range.startContainer.ownerDocument;
    let root = get_root(doc);

    // Do not start or end in luci_node
    for (let n = range.startContainer; n != root; n = n.parentNode) {
	if (is_luci_node(n)) {
	    range.setStartAfter(n);
	    break;
	}
    }
    for (let n = range.endContainer; n != root; n = n.parentNode) {
	if (is_luci_node(n)) {
	    range.setEndBefore(n);
	    break;
	}
    }

    //return range.toString();
    return range_to_endstring(range, max);
}

function prepend_string(str, val, space) {
    if (space && (str == "" || !is_ws[str[0]]))
	str = " " + str;

    let text = collapse_whitespace(val);
    if (str && is_ws[str[0]] && text && is_ws[text[text.length - 1]])
	return text.substring(0, text.length - 1) + str;
    return text + str;
}

function range_to_endstring(range, max) {
    let doc = range.endContainer.ownerDocument;
    let root = get_root(doc);

    // Single node
    if (range.startContainer == range.endContainer) {
	let s = node_to_string(range.startContainer,
			       range.startOffset, range.endOffset);
	s = collapse_whitespace(s);
	return s.substring(s.length - max);
    }

    // End node
    let s = "";
    let node = range.endContainer;
    if (node.nodeType == node.TEXT_NODE)
	s = collapse_whitespace(node_to_string(node, 0, range.endOffset));

    // Middle nodes
    let space = false;
    while (node) {
	// Find next node
	if (node.nodeType == node.ELEMENT_NODE) {
	    if (!space && (is_blocklike(node) || is_br(node)))
		space = true;
	}

	if (node.nodeType == node.ELEMENT_NODE && node.lastChild &&
	    !is_luci_node(node)) {
	    node = node.lastChild;
	} else if (node.previousSibling) {
	    node = node.previousSibling;
	} else {
	    for (node = node.parentNode; node != root; node = node.parentNode) {
		if (!space && is_blocklike(node))
		    space = true;
		if (node == range.startContainer)
		    return s.substring(s.length - max);
		if (node.previousSibling) {
		    node = node.previousSibling;
		    break;
		}
	    }
	}

	if (node == range.startContainer || node == root)
	    break;

	if (node.parentNode == range.startContainer &&
	    node == range.startContainer.childNodes[range.startOffset])
	    return s.substring(s.length - max);

	// Check node
	if (node.nodeType == node.TEXT_NODE) {
	    s = prepend_string(s, node.nodeValue, space);
	    space = false;
	    if (max > -1 && s.length >= max)
		return s.substring(s.length - max);
	}
    }

    // Start node
    if (range.startContainer) {
	s = prepend_string(s, node_to_string(range.startContainer,
					     range.startOffset, -1), space);
    }

    return s.substring(s.length - max);
}

function bar_add_note() {
    let box = document.getElementById("bar_textbox");
    let nid;
    if (box.value != "")
	nid = do_add_note(box.value, null);
    else
	nid = do_add_note(null, "highlight");
    go_open_note(nid);
}

function add_highlight_note() {
    do_add_note(null, "highlight");
}

function add_note() {
    let style = null;
    let comment = null;
    let tb = document.getElementById("note_text");
    if (tb.value)
	comment = tb.value;
    tb.value = "";
    tb = document.getElementById("note_highlight");
    if (tb.checked)
	style = "highlight";

    do_add_note(comment, style);
}

function do_add_note(comment, style) {
    debug("do_add_note");

    let note = { publication: { },
		 annotator: null,
		 source: "user",
		 targets: [],
		 content: comment,
		 type: null,
		 color: null,
		 style: style,
		 position: null };

    if (note.content)
	note.type = "text";

    let target = selection_to_target();
    //if (target.fixptr)
    //	note.position = get_note_position(target.fixptr);
    note.targets.push(target);

    force_to_current(note);
    note_check(note);
    note.id = note_manager.add(note);

    document.getElementById("new_note").hidden = true;
    document.getElementById("create_note_button").hidden = false;

    list_notes();

    // mark note
    if (should_mark_notes())
	mark_note(note);

    // highlight note
    if (note.style == "highlight")
	highlight_note(note);

    return note.id;
}

function cancel_note() {
    document.getElementById("note_text").value = "";
    document.getElementById("new_note").hidden = true;
    document.getElementById("create_note_button").hidden = false;
}

function toggle_note_order() {
    list_notes();
}

function list_notes() {
    if (!note_manager)
	return;

    let iframe = document.getElementById("notes");
    while (iframe.childNodes.length > 0)
	iframe.removeChild(iframe.firstChild);

    //let nids = note_manager.find_notes(book.sha1, book.rootfile);
    let notes = note_manager.get_notes_path(book.sha1, book.rootfile);

    let date_order = document.getElementById("order_annotations").checked;
    if (!date_order) {
	let sord = book.get_spine_order();
	notes.sort(function(na, nb) {
	    if (na.targets.length == 0 && nb.targets.length == 0)
		return 0;
	    if (na.targets.length == 0)
		return 1;
	    if (nb.targets.length == 0)
		return -1;

	    let a = na.targets[0];
	    let b = nb.targets[0];

	    if (a.part != b.part) {
		if (sord[a.part] == sord[b.part])	// One is fallback item
		    return 0;
		if (sord[a.part] == -1 || sord[a.part] == undefined)
		    return 1;
		if (sord[b.part] == -1 || sord[b.part] == undefined)
		    return -1;
		if (sord[a.part] < sord[b.part])
		    return -1;
		return 1;
	    }

	    // a.part == b.part
	    if (a.fixptr || b.fixptr) {
		if (!a.fixptr)
		    return -1;
		if (!b.fixptr)
		    return 1;
		return compare_fixptr(a.fixptr, b.fixptr);
	    } else if (a.epubcfi || b.epubcfi) {
		if (!a.epubcfi)
		    return -1;
		if (!b.epubcfi)
		    return 1;
		return compare_epubcfi(a.epubcfi, b.epubcfi);
	    }
	    //if (!a.fixptr && !b.fixptr && !a.epubcfi && !b.epubcfi)
	    return 0;
	});
    }

    for (let nid in loop(notes)) {
	debug("note: " + nid);
	let note = note_manager.get_note(nid.id);
	//let note = note_manager.get_note(nid);

	// Main frame
	let div = document.createElementNS(XHTML_NS, "div");
	div.setAttribute("data-luci", "" + note.id);
	div.setAttribute("class", "noteitem");
	iframe.appendChild(div);

	// Comment and target div
	let div1 = document.createElementNS(XHTML_NS, "div");
	div.appendChild(div1);

	// opendiv
	let div2 = document.createElementNS(XHTML_NS, "div");
	div.appendChild(div2);

	make_link(div, note);
    }

    if (notes.length > 0) {
	document.getElementById("export_notes_button").hidden = false;
	document.getElementById("print_notes_button").hidden = false;
    } else {
	document.getElementById("export_notes_button").hidden = true;
	document.getElementById("print_notes_button").hidden = true;
    }
}

    /*
      <div class='navitem' data='nid'>
	<div>
	  <a href='#'>*</a> <span>...:</span><br/>
	  <a href="#">• ...</a><br/>
	  <a href="#">• ...</a>
	</div>
	<div>

	</div>
      </div>
     */
function make_link(notediv, note) {
    let container = notediv.childNodes[0];
    let opendiv = notediv.childNodes[1];
    while (container.hasChildNodes())
	container.removeChild(container.firstChild);

    let handle = document.createElementNS(XHTML_NS, "a");
    container.appendChild(handle);
    handle.setAttribute("href", "#");

    let span = document.createElementNS(XHTML_NS, "span");
    handle.appendChild(span);
    set_as_closed(span);

    handle.addEventListener("click", {
	opendiv: opendiv,
	nid: note.id,
	span: span,
	handleEvent: function(e) {
	    toggle_note(this.span, this.opendiv, this.nid);
	}
    }, false);

    handle.appendChild(document.createTextNode(" "));	// NBSP

    span = document.createElementNS(XHTML_NS, "span");
    span.setAttribute("class", "infoheader");
    let text = note.content;
    if (!text)
	text = "\u2026";	// "..."
    if (text.length > 50)
	text = trim_end(text.substring(0, 50), 7) + "\u2026";	// "..."
    //text += ":";
    span.appendChild(document.createTextNode(text));
    handle.appendChild(span);

    let sharedbundle = document.getElementById("lucifox-shared_strings");

    for (let target in loop(note.targets)) {
	let linkdiv = document.createElementNS(XHTML_NS, "div");
	linkdiv.setAttribute("class", "clear");
	container.appendChild(linkdiv);

	let link = document.createElementNS(XHTML_NS, "a");
	linkdiv.appendChild(link);

	link.setAttribute("href", "#");
	link.setAttribute("class", "blink clear");
	link.addEventListener("click", {
	    opendiv: notediv.childNodes[1],
	    nid: note.id,
	    target: target,
	    handleEvent: function(e) {
		view_note(this.opendiv, this.nid, this.target);
	    }
	}, false);

	text = target.text;
	if (!text)
	    text = "\u2026";	// "..."
	if (text.length > 45)
	    text = trim_end(text.substring(0, 45), 7) + "\u2026";	// "..."
	link.appendChild(document.createTextNode("– " + text));	// ↗ •
	link.setAttribute("title", note.content);

	// Remove button
	linkdiv.appendChild(document.createTextNode(" "));

	link = document.createElementNS(XHTML_NS, "a");
	linkdiv.appendChild(link);
	link.setAttribute("href", "#");
	link.setAttribute("class", "button right");
	link.addEventListener("click", {
	    pid: target.id,
	    nid: note.id,
	    opendiv: notediv.childNodes[1],
	    handleEvent: function(e) {
		remove_target(this.pid, this.nid, this.opendiv);
	    }
	}, false);
	link.setAttribute("data-luciopen", "true");
	text = sharedbundle.getString("remove");
	link.appendChild(document.createTextNode("[" + text + "]"));
	link.hidden = true;
    }
}

function toggle_note(handle, opendiv, nid) {
    if (opendiv.hasChildNodes()) {
	close_note(opendiv);
    } else {
	open_note(nid, opendiv, false);
    }
}

function go_open_note(nid) {
    debug("go_open_note: " + nid);
    let el = document.getElementById("notes");
    let div = null;
    for (let ch in loop(el.childNodes)) {
	if (ch.getAttribute("data-luci") == "" + nid) {
	    div = ch;
	    break;
	}
    }

    if (div) {
	//open_note(nid, div.childNodes[1], false);
	set_selected_note(div);
	div.scrollIntoView();
	open_annotations();
    }
}

function set_selected_note(div) {
    for (let ch in loop(div.parentNode.childNodes))
	if (ch != div)
	    ch.setAttribute("class", "noteitem");
    div.setAttribute("class", "noteitem selected");
}

function view_note(opendiv, nid, target) {
    debug("view_note: " + target.fixptr);
    if (target.part == get_part_path()) {
	scroll_to_target(target, browser.contentDocument,
			 browser.contentWindow);
    } else {
	let path = bookid + "/" + target.part;
	if (target.fixptr || target.epubcfi || target.start ||
	    target.pstart || target.xpointer) {
	    view_uri_x(path, null, 0, target);
	} else {
	    view_uri_x(path, null, 0, null);
	}
    }
    set_selected_note(opendiv.parentNode);
}

function close_note(opendiv) {
    while (opendiv.hasChildNodes())
	opendiv.removeChild(opendiv.firstChild);
    let notediv = opendiv.parentNode;
    for (let el in loop(notediv.getElementsByTagName("*"))) {
	if (el.hasAttribute("data-luciopen"))
	    el.hidden = true;
    }

    let grip = notediv.childNodes[0].childNodes[0].childNodes[0];
    set_as_closed(grip);
}

function open_note(nid, opendiv, remake) {
    debug("open_note");

    let notediv = opendiv.parentNode;
    for (let el in loop(notediv.getElementsByTagName("*"))) {
	if (el.hasAttribute("data-luciopen"))
	    el.hidden = false;
    }

    let grip = notediv.childNodes[0].childNodes[0].childNodes[0];
    set_as_open(grip);

    if (remake) {
	while (opendiv.hasChildNodes())
	    opendiv.removeChild(opendiv.firstChild);
    } else if (opendiv.hasChildNodes())
	return;

    let note = note_manager.get_note(nid);

    let div = document.createElementNS(XHTML_NS, "div");
    div.setAttribute("class", "note");

    opendiv.appendChild(div);

    // Add target selection
    let adiv = document.createElementNS(XHTML_NS, "div");
    adiv.setAttribute("class", "right clear");
    div.appendChild(adiv);

    let link = document.createElementNS(XHTML_NS, "a");
    adiv.appendChild(link);
    link.setAttribute("href", "#");
    link.setAttribute("class", "button");
    link.addEventListener("click", {
	nid: note.id,
	opendiv: opendiv,
	handleEvent: function(e) { add_target(this.nid, this.opendiv); }
    }, false);
    let bundle = document.getElementById("lucifox-strings");
    let text = bundle.getString("add_pointer");
    link.appendChild(document.createTextNode("[" + text + "]"));

    // Space
    let br = document.createElementNS(XHTML_NS, "div");
    br.setAttribute("class", "clear");
    br.appendChild(document.createTextNode(" "));	// NBSP
    div.appendChild(br);

    // Modification date
    if (note.modified) {
	div.appendChild(document.createTextNode(note.modified.toLocaleString()));
	div.appendChild(document.createElementNS(XHTML_NS, "br"));
    }

    // Comment text
    let tb = document.createElementNS(XUL_NS, "textbox");
    tb.setAttribute("multiline", "true");
    tb.setAttribute("rows", "4");
    tb.setAttribute("cols", "80");
    tb.setAttribute("flex", "1");
    if (note.content) {
	tb.setAttribute("value", note.content);
	//tb.value = note.content;
    }
    div.appendChild(tb);

    div.appendChild(document.createElementNS(XHTML_NS, "br"));

    // Edit comment
    link = document.createElementNS(XHTML_NS, "a");
    div.appendChild(link);
    link.setAttribute("href", "#");
    link.setAttribute("class", "button right");
    link.addEventListener("click", {
	note: note,
	opendiv: opendiv,
	handleEvent: function(e) {
	    edit_comment(this.note, this.opendiv);
	}
    }, false);
    text = bundle.getString("edit_comment");
    link.appendChild(document.createTextNode("[" + text + "]"));

    div.appendChild(document.createElementNS(XHTML_NS, "br"));

    // Highlight checkbox
    let cb = document.createElementNS(XUL_NS, "checkbox");
    text = bundle.getString("highlight");
    cb.setAttribute("label", text);

    cb.addEventListener("click", {
	cb: cb,
	handleEvent: function(e) {
	    if (this.cb.checked)
		this.cb.setAttribute("checked", "true");
	    else
		this.cb.removeAttribute("checked");
	}
    }, false);
    div.appendChild(cb);
    if (note.style == "highlight")
	cb.setAttribute("checked", "true");

    div.appendChild(document.createElementNS(XHTML_NS, "br"));

    // Buttons
    adiv = document.createElementNS(XHTML_NS, "div");
    adiv.setAttribute("class", "right");
    div.appendChild(adiv);

    link = document.createElementNS(XHTML_NS, "a");
    adiv.appendChild(link);
    link.setAttribute("href", "#");
    link.setAttribute("class", "button");
    link.addEventListener("click", {
	nid: note.id,
	handleEvent: function(e) { delete_note(this.nid); }
    }, false);
    let sharedbundle = document.getElementById("lucifox-shared_strings");
    text = sharedbundle.getString("delete");
    link.appendChild(document.createTextNode("[" + text + "]"));

    adiv.appendChild(document.createTextNode("  "));	// NBSP

    link = document.createElementNS(XHTML_NS, "a");
    adiv.appendChild(link);
    link.setAttribute("href", "#");
    link.setAttribute("class", "button");
    link.addEventListener("click", {
	nid: nid,
	opendiv: opendiv,
	textbox: tb,
	stylebox: cb,
	handleEvent: function(e) {
	    update_note(this.nid, this.opendiv, this.textbox.value,
			this.stylebox.checked);
	}
    }, false);
    text = sharedbundle.getString("update");
    link.appendChild(document.createTextNode("[" + text + "]"));
}

function edit_comment(note, opendiv) {
    window.openDialog("chrome://lucidor/content/editnote.xul",
		      "editNote", "chrome,resizable,centerscreen",
		      { note: note, opendiv: opendiv });
}

function refresh_note(nid, opendiv) {
    let note = note_manager.get_note(nid);
    make_link(opendiv.parentNode, note);
    open_note(nid, opendiv, true);
}

function remove_target(pid, nid, opendiv) {
    note_manager.remove_target(pid);
    note_manager.update_modified(nid);
    refresh_note(nid, opendiv);
    refresh_marks(nid);
}

function refresh_marks(nid) {
    if (should_mark_notes()) {
	unmark_notes();
	mark_notes();
	if (true)
	    highlight_notes();
    }
}

function add_target(nid, opendiv) {
    let target = selection_to_target();
    if (target) {
	note_manager.add_target(nid, target);
	note_manager.update_modified(nid);
	refresh_note(nid, opendiv);
	refresh_marks(nid);
    }
}

function update_note(nid, opendiv, text, highlight) {
    let oldnote = null;
    let style = null;
    if (highlight)
	style = "highlight";
    note_manager.update(nid, text, "text", style);

    refresh_note(nid, opendiv);
    refresh_marks(nid);
}

// update == false: only test if there are targets to update on this page
function update_targets(update) {
    debug("update_targets: " + update);
    let modified = false;
    let part = get_part_path();
    let nids = note_manager.find_part_notes(book.sha1, book.rootfile, part);
    for (let nid in loop(nids)) {
	let note = note_manager.get_note(nid);
	let doc = browser.contentDocument;
	for (let target in loop(note.targets)) {
	    if (target.part == part) {
		let root = get_root(doc);
		if (fill_target(target, root, null)) {
		    if (!update)
			return true;
		    note_manager.update_target(target);
		    modified = true;
		}
	    }
	}
    }
    return modified;
}

function update_target_parts() {
    debug("update_target_parts");
    let nids = note_manager.find_notes(book.sha1, book.rootfile);
    for (let nid in loop(nids)) {
	let note = note_manager.get_note(nid);
	for (let target in loop(note.targets)) {
	    if (!target.part) {
		if (target.epubcfi) {
		    let part = epubcfi_to_part(target.epubcfi);
		    if (part) {
			target.part = part;
			note_manager.update_target(target);
		    }
		}
	    }
	}
    }
}

function fill_target(target, root, range) {
    let res = false;
    if (!range) {
	range = target_to_domrange(target, root);
	if (!range)
	    return res;
    }
    if (!target.fixptr) {
	target.fixptr = domrange_to_fixptr(range);
	if (target.fixptr)
	    res = true;
    }
    if (!target.epubcfi) {
	target.epubcfi = domrange_to_epubcfi(range);
	if (target.epubcfi)
	    res = true;
    }
    /*
    if (!target.start) {
	let loc = domrange_to_xpathlocation(range);
	if (loc && loc.start) {
	    target.start = loc.start;
	    target.startoffset = loc.startoffset;
	    if (loc.end) {
		target.end = loc.end;
		target.endoffset = loc.endoffset;
	    }
	    res = true;
	}
    }
    */
    /*
    if (!target.pstart) {
	let loc = domrange_to_adelocation(range);
	if (loc && loc.pstart) {
	    target.pstart = loc.pstart;
	    if (loc.pend)
		target.pend = loc.pend;
	    res = true;
	}
    }
    */
    /*
    if (!target.xpointer) {
	target.xpointer = domrange_to_xpointer(range);
	if (target.xpointer)
	    res = true;
    }
    */
    if (!target.text) {
	//let text = range_to_string(range, 150);
	let text = get_rangetext(range, 150);
	if (text.length < 150) {
	    target.text = text;
	    target.endtext = null;
	} else {
	    target.text = trim_end(text.substring(0, 100), 10);
	    target.endtext = trim_start(get_endtext(range, 45), 10);
	}
	target.pretext = trim_start(get_pretext_x(range, 27), 7);
	target.posttext = trim_end(get_posttext_x(range, 27), 7);

	if (target.text || target.pretext || target.posttext)
	    res = true;
    }

    return res;
}

function delete_note(nid) {
    note_manager.remove_note(nid);
    list_notes();
    refresh_marks(nid);
}

function unhighlight_notes() {
    let sel = get_highlight_selection();
    sel.removeAllRanges();
}

function highlight_notes() {
    if (!note_manager)
	return;

    let sel = get_highlight_selection();
    sel.removeAllRanges();
    let part = get_part_path();
    let nids = note_manager.find_highlighted_notes(book.sha1, book.rootfile,
						   part);
    for (let nid in loop(nids)) {
	debug("highlight note: " + nid);
	let note = note_manager.get_note(nid);
	for (let target in loop(note.targets)) {
	    if (target.part != part)
		continue;
	    let dr = target_to_domrange(target, null);
	    if (dr)
		sel.addRange(dr);
	}
    }
}

function highlight_note(note) {
    debug("highlight_note");
    if (!note.targets)
	return;

    let sel = get_highlight_selection();
    let part = get_part_path();
    for (let target in loop(note.targets)) {
	if (target.part != part)
	    continue;
	let dr = target_to_domrange(target, null);
	if (dr)
	    sel.addRange(dr);
    }
}

function get_selection_controller() {
    /*let dv = browser.contentDocument.defaultView;
    let selcon = dv.QueryInterface(Components.interfaces.nsIInterfaceRequestor).
	getInterface(Components.interfaces.nsIWebNavigation).
	QueryInterface(Components.interfaces.nsIInterfaceRequestor).
	getInterface(Components.interfaces.nsISelectionDisplay).
	QueryInterface(Components.interfaces.nsISelectionController);*/
    let selcon = browser.docShell
	.QueryInterface(Components.interfaces.nsIInterfaceRequestor)
	.getInterface(Components.interfaces.nsISelectionDisplay)
	.QueryInterface(Components.interfaces.nsISelectionController);
    return selcon;
}

function get_highlight_selection() {
    let selcon = get_selection_controller();
    return selcon.getSelection(selcon.SELECTION_SPELLCHECK);
    //return selcon.getSelection(selcon.SELECTION_FIND);
}

function should_mark_notes() {
    return document.getElementById("mark_annotations").checked;
}

function toggle_mark_notes() {
    debug("mark_notes");
    if (should_mark_notes()) {
	mark_notes();
	highlight_notes();
    } else {
	unmark_notes();
	unhighlight_notes();
    }
}

function mark_notes() {
    let part = get_part_path();
    let nids = note_manager.find_part_notes(book.sha1, book.rootfile, part);
    for (let nid in loop(nids)) {
	debug("mark_notes: " + nid);
	let note = note_manager.get_note(nid);
	mark_note(note);
    }
}

function mark_note(note) {
    if (!note.content)
	return;

    let doc = browser.contentDocument;
    let part = get_part_path();
    for (let target in loop(note.targets)) {
	if (target.part != part)
	    continue;
	let dr = target_to_domrange(target, null);
	if (!dr) {
	    let n = doc.getElementsByTagNameNS(XHTML_NS, "body");
	    if (n) {
		dr = doc.createRange();
		dr.setStart(n[0], 0);
	    } else
		continue;
	}

	dr.collapse(true);
	//debug(dr + ":" + dr.startOffset);

	let span = doc.createElementNS(XHTML_NS, "abbr");
	set_luci_node(span);
	span.setAttribute("title", note.content);
	span.setAttribute("style", "background-color: #ffdc85; color: #000000; border: 1px solid -moz-use-text-color; margin-left: 1px; margin-right: 1px; padding-left: 2px; padding-right: 2px; border-radius: 4px; cursor: pointer;");
	//span.appendChild(doc.createTextNode(" ✎ "));	// nnbsp
	span.appendChild(doc.createTextNode("★"));
	// ◆◇●⬤☥☛☞❦✗✎★☆ ▶▷▼▽
	span.addEventListener("click", {
	    nid: note.id,
	    target: target,
	    handleEvent: function(e) { go_open_note(this.nid, this.target); }
	}, false);

	dr.insertNode(span);
    }
}

function unmark_notes() {
    let doc = browser.contentDocument;
    let xpe = new XPathEvaluator();
    // 6 = UNORDERED_NODE_SNAPSHOT_TYPE
    let result = xpe.evaluate("//*[@class='lucitemp']", doc.documentElement,
			      xpe.createNSResolver(doc), 6, null);
    for (let i = 0; i < result.snapshotLength; i++) {
	let node = result.snapshotItem(i);
	node.parentNode.removeChild(node);
    }
}

function import_notes() {
    let fp = Components.classes["@mozilla.org/filepicker;1"].
	createInstance(Components.interfaces.nsIFilePicker);
    let bundle = document.getElementById("lucifox-strings");
    let text = bundle.getString("import_annotations");
    fp.init(window, text, Components.interfaces.nsIFilePicker.modeOpen);

    fp.appendFilters(fp.filterAll | fp.filterText);

    let callback = {
	done: function(result) {
	    if (result == fp.returnOK) {
		NetUtil.asyncFetch(fp.file, function(inputStream, status) {
		    if (!Components.isSuccessCode(status)) {
			// Handle error!
			return;
		    }
		    let data = NetUtil.readInputStreamToString(inputStream, inputStream.available(), { charset: "UTF-8" });
		    try_import_notes(data);
		});
	    }
	}
    };

    if (fp.open != undefined)
	fp.open(callback);
    else
	callback.done(fp.show());
}

function try_import_notes(data) {
    debug("try_import_notes");
    let ret = check_import_notes(data);
    debug("try_import_notes: " + ret);
    if (ret == 0) {		// No annotations
    } else if (ret == -3 || ret == -2) {	// Invalid file
	let prompts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
	    .getService(Components.interfaces.nsIPromptService);
	let bundle = document.getElementById("lucifox-strings");
	let title = bundle.getString("import_annotations");
	let text = bundle.getString("import_invalid");
	prompts.alert(null, title, text);
    } else if (ret == -1) {	// Annotations for other book
	let prompts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
	    .getService(Components.interfaces.nsIPromptService);
	let bundle = document.getElementById("lucifox-strings");
	let title = bundle.getString("import_annotations");
	let text = bundle.getString("import_original_question");
	let ch_text = bundle.getString("import_current");
	let check = { value: false };
	let val = prompts.confirmCheck(null, title, text, ch_text, check);
	if (val)
	    do_import_notes(data, check.value, false);
    } else if (ret == 1) {	// Annotations for current book
	do_import_notes(data, false, true);
    } else if (ret == 2 || ret == 3) {	// Annotations for several books
	let prompts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
	    .getService(Components.interfaces.nsIPromptService);
	let bundle = document.getElementById("lucifox-strings");
	let title = bundle.getString("import_annotations");
	let text = bundle.getString("import_several_question");
	let ch_text = bundle.getString("import_current");
	let check = { value: false };
	let val = prompts.confirmCheck(null, title, text, ch_text, check);
	if (val)
	    do_import_notes(data, false, check.value);
    }
}

function check_import_notes(data) {
    let inp;

    try {
	inp = JSON.parse(data);
    } catch (e) {
	return -3;
    }

    if (!inp || !inp.type || !inp.type == "http://lucidor.org/-/annotations/")
	return -2;

    if (!inp.annotations || !inp.annotations.length ||
	inp.annotations.length == 0)
	return 0;

    if (inp.annotations.length > 1)
	return 2;

    let pub = inp.annotations[0].publication;
    if (pub.sha1 != book.sha1 || pub.root != book.rootfile)
	return -1;
    return 1;
}

function force_to_current(note) {
    if (!note.publication)
	note.publication = {};
    note.publication.sha1 = book.sha1;
    note.publication.root = book.rootfile;
    note.publication.rootfile = book.rootfiles[book.rootfile];
    if (book.opf.uid)
	note.publication.uid = book.opf.uid;
    if (book.opf.metadata.modified)
	note.publication.modified = book.opf.metadata.modified;

    if (!note.publication.info)
	note.publication.info = {};
    if (book.opf.metadata.title)
	note.publication.info.title = book.opf.metadata.title;
    if (book.opf.metadata.author)
	note.publication.info.creator = book.opf.metadata.author.name;
    else if (book.opf.metadata.creator.length > 0)
	note.publication.info.creator = book.opf.metadata.creator[0].name;
    if (book.opf.metadata.publisher)
	note.publication.info.publisher = book.opf.metadata.publisher;
}

function note_check(note) {
    if (note.publication == undefined)
	note.publication = [];
    if (note.publication.sha1 == undefined)
	note.publication.sha1 = null;
    if (note.publication.root == undefined)
	note.publication.root = 0;
    if (note.publication.rootfile == undefined)
	note.publication.rootfile = null;
    if (note.publication.uid == undefined)
	note.publication.uid = null;
    if (note.publication.modified == undefined)
	note.publication.modified = null;
    if (note.publication.info == undefined)
	note.publication.info = null;

    if (note.uid == undefined)
	note.uid = null;
    if (note.annotator == undefined)
	note.annotator = null;
    if (note.source == undefined)
	note.source = null;
    if (note.targets == undefined)
	note.targets = [];
    for (let target in loop(note.targets)) {
	if (target.part == undefined)
	    target.part = null;
	if (target.endpart == undefined)
	    target.endpart = null;
	if (target.fixptr == undefined)
	    target.fixptr = null;
	if (target.epubcfi == undefined)
	    target.epubcfi = null;
	if (target.text == undefined)
	    target.text = null;
	if (target.endtext == undefined)
	    target.endtext = null;
	if (target.pretext == undefined)
	    target.pretext = null;
	if (target.posttext == undefined)
	    target.posttext = null;
    }
    if (note.content == undefined)
	note.content = null;
    if (note.type == undefined)
	note.type = null;
    if (note.color == undefined)
	note.color = null;
    if (note.style == undefined)
	note.style = null;
    if (note.extra == undefined)
	note.extra = null;
    if (note.position == undefined)
	note.position = null;
}

function do_import_notes(data, force_current, only_current) {
    debug("do_import_notes");
    let num = 0;

    let inp = JSON.parse(data);
    for (let entry in loop(inp.annotations)) {
	for (let note in loop(entry.annotations)) {
	    if (only_current && (entry.publication.sha1 != book.sha1 ||
				 entry.publication.root != book.rootfile))
		continue;

	    if (force_current) {
		force_to_current(note);
	    } else {
		if (!note.publication)
		    note.publication = {};
		note.publication.sha1 = entry.publication.sha1;
		note.publication.root = entry.publication.root;
		note.publication.rootfile = entry.publication.rootfile;
		note.publication.uid = entry.publication.uid;
		note.publication.modified = entry.publication.modified;
		note.publication.info = {};
		if (entry.publication.title)
		    note.publication.info.title = entry.publication.title;
		if (entry.publication.creator)
		    note.publication.info.creator = entry.publication.creator;
		if (entry.publication.publisher)
		    note.publication.info.publisher = entry.publication.publisher;
	    }
	    note.source = "imported";
	    note_check(note);
	    note_manager.add(note);
	    num++;
	}
    }

    done_importing(num);
    return 0;
}

function done_importing(num) {
    debug("done_importing");
    update_target_parts();
    if (true) {
	list_notes();
	if (update_targets(false))
	    reload_page();
	else
	    refresh_marks();
    }

    let bundle = document.getElementById("lucifox-strings");
    let text = bundle.getFormattedString("import_done", [num.toString()]);
    let nb = document.getElementById("notification_box");
    let no = nb.appendNotification(text, "imp", null,
				   nb.PRIORITY_INFO_HIGH, [], null);
    autohide(no, true);

    return 0;
}

function export_notes() {
    main_control.add_web_tab("x-lucidor://notes/" + bookid);
}

function export_all_notes() {
    main_control.add_web_tab("x-lucidor://notes/");
}

function print_notes() {
    main_control.add_web_tab("x-lucidor://printnotes/" + bookid);
}

function annotation_manager() {
    main_control.add_web_tab("chrome://lucidor/content/annotations.xul?bookid="
			     + bookid);
}

function init_line_selection() {
    line_move_line_selection(false);
}

function page_move_line_selection(forward) {
    let selcon = get_selection_controller();
    let sel = selcon.getSelection(selcon.SELECTION_NORMAL);
    let node = sel.anchorNode;
    let offset = sel.anchorOffset;
    let extend = true;
    if (!node) {
	forward = !forward;
	extend = false;
    }
    selcon.pageMove(forward, extend);
    selcon.intraLineMove(false, false);
    selcon.intraLineMove(true, true);
    if (sel.anchorNode == node && sel.anchorOffset == offset) {
	if (forward)
	    show_next_part(false);
	else
	    show_previous_part(false);
    } else {
	scroll_selection_into_view();
    }
}

function line_move_line_selection(forward) {
    let el = document.commandDispatcher.focusedElement;
    if (el)
	debug("focus: " + el.tagName + " " + el.getAttribute("tabindex"));
    el = document.activeElement;
    if (el)
	debug("focus: " + el.tagName + " " + el.getAttribute("tabindex"));


    let selcon = get_selection_controller();
    let sel = selcon.getSelection(selcon.SELECTION_NORMAL);
    let node = sel.anchorNode;
    let offset = sel.anchorOffset;
    let extend = true;
    if (!node) {
	forward = !forward;
	extend = false;
    }
    try {
	selcon.lineMove(forward, extend);
	selcon.intraLineMove(false, false);
	selcon.intraLineMove(true, true);
	if (sel.anchorNode == node && sel.anchorOffset == offset) {
	    if (forward)
		show_next_part(false);
	    else
		show_previous_part(false);
	} else {
	    scroll_selection_into_view();
	}
    } catch (e) {
	debug("scroll error: " + e);
    }
}

function line_scroll(event) {
    if (event.detail < 0)
	line_move_line_selection(false);
    else
	line_move_line_selection(true);
}

function scroll_selection_into_view() {
    let selcon = get_selection_controller();

    try {
	let sel = selcon.getSelection(selcon.SELECTION_NORMAL);
	sel.QueryInterface(Components.interfaces.nsISelectionPrivate);
	sel.scrollIntoView(selcon.SELECTION_ANCHOR_REGION, false, 50, -1);
	sel.scrollIntoView(selcon.SELECTION_FOCUS_REGION, false, 50, -1);
    } catch (e) {
	debug("scroll error 1: " + e);
	try {
	    let selcon = get_selection_controller();
	    selcon.scrollSelectionIntoView(selcon.SELECTION_NORMAL,
					   selcon.SELECTION_ANCHOR_REGION,
					   selcon.SCROLL_CENTER_VERTICALLY);
	    selcon.scrollSelectionIntoView(selcon.SELECTION_NORMAL,
					   selcon.SELECTION_FOCUS_REGION,
					   selcon.SCROLL_CENTER_VERTICALLY);
	} catch (e) {
	    debug("scroll error 2: " + e);
	}
    }
}

function play_pause() {
    debug("play_pause");
    let button = document.getElementById("play-button");
    let is_playing = button.hasAttribute("checked");	// FIXME or !

    if (is_playing) {
	//is_playing = false;
	pause();
	button.label = "Play";	// TRANSLATE-ME
	button.removeAttribute("checked");
    } else {
	//is_playing = true;
	play();
	button.label = "Pause";	// TRANSLATE-ME
	button.setAttribute("checked", "true");
    }
}

function play() {
    debug("play");

    play_next();
}

function pause() {
    speak.pause();
}

function stop() {
    speak.pause();
    play_queue = [];
}

var play_queue = [];

function play_next() {
    debug("play_next");

    while (play_queue.length < 3) {
	// Add more
	if (!play_add())
	    break;
    }

    if (play_queue.length == 0) {
	let button = document.getElementById("play-button");
	button.removeAttribute("checked");
	return;
    }

    // Get next line from queue
    let range = play_queue[0];
    let selcon = get_selection_controller();
    let sel = selcon.getSelection(selcon.SELECTION_NORMAL);
    sel.removeAllRanges();
    sel.addRange(range);
}

function play_add() {
    debug("play_add");
    if (!play_select_next_line())
	return false;
    let selcon = get_selection_controller();
    let sel = selcon.getSelection(selcon.SELECTION_NORMAL);
    play_queue.push(sel.getRangeAt(0));
    speak.myplay(sel.toString(), null, play_ended);
    debug("text: " + sel.toString());
    return true;
}

function play_ended() {
    debug("play_ended");
    if (play_queue.length > 0)		// Should always be true
	play_queue.shift();		// Remove finished range
    play_next();
}

function play_select_next_line() {
    let forward = true;
    let selcon = get_selection_controller();
    let sel = selcon.getSelection(selcon.SELECTION_NORMAL);
    let node = sel.anchorNode;
    let offset = sel.anchorOffset;
    if (!node)
	forward = !forward;
    selcon.lineMove(forward, false);
    selcon.intraLineMove(false, false);
    selcon.intraLineMove(true, true);
    if (sel.anchorNode == node && sel.anchorOffset == offset) {
	return false;
	/*
	if (forward)
	    show_next_part();
	else
	    show_previous_part();
	*/
    }
    return true;
}

function open_searchpanel() {
    if (document.getElementById("mysidebar").hidden)
	toggle_sidebar();
    document.getElementById("search_button").doCommand();
}

function on_bar_search_keypress(event) {
    if (event.keyCode == 13)
	bar_find_in_document();
}

function bar_find_in_document() {
    let box = document.getElementById("bar_textbox");
    if (do_find_in_document(box.value))
	open_searchpanel();
}

function on_search_keypress(event) {
    if (event.keyCode == 13)
	find_in_document();
}

function find_in_document() {
    let box = document.getElementById("search-box");
    do_find_in_document(box.value);
}

var finder;
var spart;
var search_string;
var search_stylesheet;
var searches;

function do_find_in_document(sstr) {
    debug("do_find_in_document");
    if (sstr == "")
	return false;

    search_string = sstr;

    finder = Components.classes["@mozilla.org/embedcomp/rangefind;1"].
	createInstance(Components.interfaces.nsIFind);
    finder.caseSensitive = false;
    finder.findBackwards = false;

    spart = 0;
    searches = new Array();

    let frame = document.getElementById("search_results");
    while (frame.childNodes.length > 0)
	frame.removeChild(frame.firstChild);

    let div = document.createElementNS(XHTML_NS, "div");
    div.setAttribute("class", "navitem");
    let bundle = document.getElementById("lucifox-strings");
    let text = bundle.getString("searching");
    div.appendChild(document.createTextNode(text));
    frame.appendChild(div);

    start_search();
    return true;
}

var searchProgressListener = {
    onLocationChange: function(webProgress, request, location) {
    },

    onProgressChange: function(webProgress, request, curSelfProgress,
			       maxSelfProgress, curTotalProgress,
			       maxTotalProgress) {
    },

    onSecurityChange: function(webProgress, request, state) {
    },

    onStateChange: function(webProgress, request, stateFlags, status) {
	if ((stateFlags & CIWPL.STATE_STOP) &&
	    (stateFlags & CIWPL.STATE_IS_DOCUMENT)) {
	    let hbrowser = document.getElementById("hidden_browser_s");
	    if (webProgress.DOMWindow != hbrowser.contentWindow)
		return null;

	    let doc = webProgress.DOMWindow.document;
	    /*
	    for (let el in loop(doc.getElementsByTagName("html")))
		el.setAttribute("style", "-moz-user-focus: ignore");
	    for (let el in loop(doc.getElementsByTagName("html")))
		el.setAttribute("tabindex", "-1");
	    for (let el in loop(doc.getElementsByTagName("a")))
		el.setAttribute("tabindex", "-1");
	    */

	    /*
	    let oldroot = doc.documentElement;
	    let root = doc.createElementNS(XHTML_NS, "x-ordbrand-root");
	    let glue = doc.createElementNS(XHTML_NS, "x-ordbrand-glue");
	    doc.replaceChild(root, oldroot);
	    root.appendChild(glue);
	    glue.appendChild(oldroot);
	    if (search_stylesheet) {
		let pi = doc.createProcessingInstruction("xml-stylesheet",
							 search_stylesheet);
		doc.insertBefore(pi, doc.firstChild);
	    }
	    */
	    find_in_part(doc);
	    return start_search();
	}
	return null;
    },

    onStatusChange: function(webProgress, request, status, message) {
    },

    QueryInterface: function(aIID) {
	if (aIID.equals(Components.interfaces.nsIWebProgressListener) ||
	    aIID.equals(Components.interfaces.nsISupportsWeakReference))
	    return this;
	throw Components.results.NS_NOINTERFACE;
    }
};

function start_search() {
    debug("start_search: " + search_string);
    if (spart >= spine.length) {
	debug("Search ended");

	let hbrowser = document.getElementById("hidden_browser_s");
	hbrowser.setAttribute("hidden", "true");

	let iframe = document.getElementById("search_results");
	let div = document.createElementNS(XHTML_NS, "div");
	div.setAttribute("class", "navitem");
	let bundle = document.getElementById("lucifox-strings");
	let text = bundle.getFormattedString("matches",
					     ['"' + search_string + '"',
					      searches.length.toString()]);
	div.appendChild(document.createTextNode(text));

	let oldfirst = iframe.firstChild;
	iframe.insertBefore(div, oldfirst);
	iframe.removeChild(oldfirst);

	return null;
    }

    debug("Part " + spart);
    let location = "x-lucidor://" + bookid + "/" + spine[spart].href;
    search_stylesheet = "x-lucidor://" + bookid + "/" + spine[spart].stylesheet;
    spart++;

    try {
	let hbrowser = document.getElementById("hidden_browser_s");
	hbrowser.removeAttribute("hidden");
	hbrowser.loadURI(location, null, null);
    } catch (e) {
	// FIXME show_warning
	debug("Error: opening page for search failed: " + location);
	return start_search();
    }
    return null;
}

function find_in_part(doc) {
    let rootNode = doc.documentElement;
    let searchRange = doc.createRange();
    searchRange.selectNodeContents(rootNode);
    //searchRange.selectNode(rootNode);
    let startPoint = searchRange.cloneRange();
    startPoint.collapse(true);
    let endPoint = searchRange.cloneRange();
    endPoint.collapse(false);

    debug("ROOT: " + rootNode.localName + " " + searchRange.startOffset + " " +
	 searchRange.endOffset);

    let iframe = document.getElementById("search_results");
    while (true) {
	let res = finder.Find(search_string, searchRange, startPoint, endPoint);
	debug("RES: " + res);
	if (!res)
	    return;
	debug(res.startOffset + " " + res.endOffset);

	let div = document.createElementNS(XHTML_NS, "div");
	div.setAttribute("class", "navitem");
	iframe.appendChild(div);
	let link = document.createElementNS(XHTML_NS, "a");
	div.appendChild(link);
	link.setAttribute("href", "#");
	link.addEventListener("click", {
	    number: searches.length,
	    handleEvent: function(e) { view_search(this.number); }
	}, false);

	let span = document.createElementNS(XHTML_NS, "span");
	span.setAttribute("class", "infoheader");
	span.appendChild(document.createTextNode(searches.length + 1));
	link.appendChild(span);
	let context = get_pretext(res, 22, true);
	link.appendChild(document.createTextNode(" " + context));
	span = document.createElementNS(XHTML_NS, "span");
	span.setAttribute("class", "match");
	span.appendChild(document.createTextNode(res.toString()));
	link.appendChild(span);
	context = get_posttext(res, 22, true);
	link.appendChild(document.createTextNode(context));

	let item = {};
	item.location = "" + doc.location;	// Must make a copy
	//item.location = spine[spart - 1].href;
	item.stylesheet = spine[spart - 1].stylesheet;
	try {
	    item.fixptr = domrange_to_fixptr(res);
	} catch (e) {
	    debug("EEE: " + e);
	}
	searches.push(item);

	startPoint.setStart(res.endContainer, res.endOffset);
	startPoint.setEnd(res.endContainer, res.endOffset);
    }
}

function is_block(node) {
    let d = node.ownerDocument.defaultView.getComputedStyle(node, "");
    if (d)
	return d.display == "block";
    return false;
}

function is_blocklike(node) {
    let d = node.ownerDocument.defaultView.getComputedStyle(node, "").display;
    //return ["inline", "none"].indexOf(d) < 0;
    return ["block", "list-item", "run-in", "inline-block",
	    "table-cell", "table-caption",
	    "flex", "inline-flex", "grid", "inline-grid"].indexOf(d) >= 0;
}

function is_br(node) {
    return node.localName == "br" && node.namespaceURI == XHTML_NS;
}

// Get the first enclosing node (or self) which is blocklike
function get_first_block(node) {
    for (; node.parentNode; node = node.parentNode)
	if ((node.nodeType == node.ELEMENT_NODE) && is_blocklike(node))
	    return node;
    return node;
}

// Get the first enclosing node (or self) which is blocklike, and which
// has no nonwhitespace text before self
function get_first_starting_block(node) {
    let top = node;
    for (; node.parentNode; node = node.parentNode) {
	if ((node.nodeType == node.ELEMENT_NODE) && is_blocklike(node))
	    return node;
	for (let n = node.previousSibling; n; n = n.previousSibling)
	    if (n.nodeType == node.TEXT_NODE)
		for (let ch in loop(n.nodeValue))
		    if (!is_wss[ch])
			return top;
    }
    return top;
}

// FIXME should collapse whitespace
function get_pretext(drange, length, trim) {
    let doc = drange.startContainer.ownerDocument;

    // Pre-context
    let pre = doc.createRange();
    pre.setStart(drange.startContainer,
		 Math.max(0, drange.startOffset - length));
    pre.setEnd(drange.startContainer, drange.startOffset);

    // Expand range to previous nodes if necessary
    let node = pre.startContainer;
    let block = get_first_block(node);
    let walker = doc.createTreeWalker(block, 0x04, null, false);
    walker.currentNode = node;
    //debug("LEN: " + pre.toString().length + " " + pre.toString());
    while (pre.toString().length < length) {
	node = walker.previousNode();
	if (!node)
	    break;
	pre.setStart(node, Math.max(0, node.textContent.length -
				    (length - pre.toString().length)));
	//debug("len: " + pre.toString().length + " " + pre.toString());
    }

    // Start of a block?
    let prebreak = false;
    if ((pre.startOffset == 0) && !walker.previousNode())
	prebreak = true;

    let pretext = pre.toString();

    // Align with word-breaks if not too much context is lost
    if (trim && !prebreak) {
	let i = pretext.indexOf(" ");
	if (i < 5)
	    pretext = pretext.substring(i);

	pretext = "\u2026" + pretext;	// "..."
    }

    return collapse_whitespace(pretext);
}

// FIXME should collapse whitespace
function get_posttext(drange, length, trim) {
    let doc = drange.startContainer.ownerDocument;

    // Post-context
    let post = doc.createRange();
    post.setStart(drange.endContainer, drange.endOffset);
    post.setEnd(drange.endContainer,
		Math.min(drange.endContainer.textContent.length,
			 drange.endOffset + length));

    // Expand range to following nodes if necessary
    let node = post.endContainer;
    let block = get_first_block(node);
    let walker = doc.createTreeWalker(block, 0x04, null, false);
    walker.currentNode = node;
    //debug("lEN: " + post.toString().length + " " + post.toString());
    while (post.toString().length < length) {
	node = walker.nextNode();
	if (!node)
	    break;
	post.setEnd(node, Math.min(node.textContent.length,
				   (length - post.toString().length)));
	//debug("len: " + post.toString().length + " " + post.toString());
    }

    // End of the block?
    let postbreak = false;
    if ((post.endOffset == post.endContainer.textContent.length) &&
	!walker.nextNode())
	postbreak = true;

    let posttext = post.toString();

    // Align with word-breaks if not too much context is lost
    if (trim && !postbreak) {
	let i = posttext.lastIndexOf(" ");
	if (i >= (posttext.length - 5))
	    posttext = posttext.substring(0, i + 1);

	posttext += "\u2026";	// "..."
    }

    return collapse_whitespace(posttext);
}

function view_search(num) {
    debug("view_search: " + searches[num].fixptr);
    if (browser.currentURI.spec == searches[num].location) {
	scroll_to_target({ fixptr: searches[num].fixptr },
			 browser.contentDocument, browser.contentWindow);
    } else {
	view_uri_x(searches[num].location, searches[num].stylesheet,
		   0, { fixptr: searches[num].fixptr });
    }
}

var is_ws = [];		// ASCII white space as used in XML Schema
is_ws[" "] = true;
is_ws["\t"] = true;
is_ws["\n"] = true;
is_ws["\r"] = true;

var is_wss = [];	// Unicode white space
for (let ch in loop(" \f\n\r\t\v​\u00a0\u1680​\u180e\u2000​\u2001\u2002​\u2003\u2004​\u2005\u2006​\u2007\u2008​\u2009\u200a​\u2028\u2029​\u202f\u205f​\u3000"))
    is_wss[ch] = true;

// FIXME This can be improved
function get_nodetext(node) {
    let text = "";
    let cut = false;
    let nls = 0;
    let walker = node.ownerDocument.createTreeWalker(node, 0x01 | 0x04,
						     null, false);
    for (let n = walker.currentNode; n; n = walker.nextNode()) {
	if (n.nodeType == n.TEXT_NODE) {
	    let aa = "";
	    let ws = (text.length > 0) && is_wss[text[text.length - 1]];
	    for (let ch in loop(n.nodeValue)) {
		if (is_wss[ch]) {
		    if (!ws) {
			aa += " ";
			ws = true;
		    }
		} else {
		    aa += ch;
		    ws = false;
		}
	    }
	    text += aa;
	} else if (text.length > 0) {
	    if (is_block(n)) {
		if (is_wss[text[text.length - 1]]) {
		    if ((text.length > 1) && !is_wss[text[text.length - 2]]) {
			text = text.substring(0, text.length - 1) + "\n\n";
			nls++;
		    }
		} else {
		    text += "\n\n";
		    nls += 2;
		}
	    } else if (n.localName == "br") {
		text += "\n";
		nls++;
	    }
	}
	if (text.length > 600 || nls > 5) {
	    cut = true;
	    break;
	}
    }

    if (cut)
	return strip_whitespace(text.substring(0, 600)) + "[\u2026]";	// "..."

    return strip_whitespace(text);
}

// Strip whitespace from the beginning and end
function strip_whitespace(text) {
    let i;
    for (i = 0; i < text.length; i++)
	if (!is_wss[text[i]])
	    break;
    let j;
    for (j = text.length - 1; j > i; j--)
	if (!is_wss[text[j]])
	    break;
    return text.substring(i, j + 1);
}

function get_linktext(link) {
    let href = link.getAttribute("href");
    if (!href)
	return null;

    let ioService = Components.classes["@mozilla.org/network/io-service;1"].
	getService(Components.interfaces.nsIIOService);
    let baseuri = ioService.newURI(link.ownerDocument.baseURI, "UTF-8", null).
	QueryInterface(Components.interfaces.nsIURL);
    let linkuri = ioService.newURI(href, "UTF-8", baseuri);

    if (linkuri.schemeIs("http") || linkuri.schemeIs("https") ||
	linkuri.schemeIs("mailto")) {
	if (is_extension)
	    return null;
	link.tip_text = href;
	link.tip_color = "#ffe7e7";
	return show_tooltip(link);
    } else if (!linkuri.schemeIs("x-lucidor")) {
	return null;
    }

    linkuri.QueryInterface(Components.interfaces.nsIURL);
    if (linkuri.filePath == baseuri.filePath)
	return get_and_show_tooltip(link, link.ownerDocument, linkuri.ref, 2);
    else
	return load_tooltip(link, linkuri.spec, linkuri.ref);
}

function get_and_show_tooltip(link, doc, idref, mmm) {
    let node;
    if (idref) {
	node = doc.getElementById(idref);
	if (!node) {
	    book.add_error("fragment identifier missing: " + idref + " " + mmm);
	    for (let n in loop(doc.getElementsByTagNameNS(XHTML_NS, "a")))
		if (n.hasAttribute("name") && n.getAttribute("name") == idref) {
		    node = n;
		    break;
		}
	}
	if (!node) {
	    link.tip_text = null;
	    return;
	}

	node = get_first_starting_block(node);
    } else {
	node = doc.documentElement;
	for (let n in loop(node.childNodes))
	    if (n.nodeType == node.ELEMENT_NODE &&
		((n.localName == "body" && n.namespaceURI == XHTML_NS) ||
		 (n.localName == "book" && n.namespaceURI == DTB_NS))) {
		node = n;
		break;
	    }
    }
    link.tip_text = get_nodetext(node);
    link.tip_color = "";
    if (link.tip_text)
	show_tooltip(link);
}

var tt_target = null;

function show_tooltip(target) {
    tt_target = target;
    setTimeout(function() {
	if (!tt_target)
	    return;
	let target = tt_target;
	tt_target = null;
	let tooltip = document.getElementById("tooltip");
	let parent = tooltip.parentNode;
	parent.removeChild(tooltip);
	parent.appendChild(tooltip);
	tooltip.label = target.tip_text;
	if (target.tip_color)
	    tooltip.setAttribute("style",
				 "background-color: " + target.tip_color + ";");
	else
	    tooltip.removeAttribute("style");
	tooltip.openPopup(target, "after_pointer", -1 , -1, false, false);
    }, 250);
}

var link_handler = {
    handleEvent: function(event) {
	let target = event.currentTarget;
	do_remove = false;
	if (target.tip_text) {
	    show_tooltip(target);
	    return;
	}
	if (target.tip_text == undefined)
	    get_linktext(target);
	//if (target.parentNode.localName == "noteref")	// FIXME
    }
};

var do_remove = false;

var link_out_handler = {
    handleEvent: function(event) {
	do_remove = true;
	tt_target = null;
	setTimeout(function() {
	    if (do_remove) {
		let tooltip = document.getElementById("tooltip");
		let parent = tooltip.parentNode;
		parent.removeChild(tooltip);
		parent.appendChild(tooltip);
	    }
	}, 25);
    }
};

function tt_over() {
    do_remove = false;
}

function tt_out() {
    do_remove = true;
    tt_target = null;
    setTimeout(function() {
	if (do_remove) {
	    let tooltip = document.getElementById("tooltip");
	    let parent = tooltip.parentNode;
	    parent.removeChild(tooltip);
	    parent.appendChild(tooltip);
	}
    }, 25);
}

function load_tooltip(link, location, fragment) {
    tooltipProgressListener.ok = false;
    let hbrowser = document.getElementById("hidden_browser_tt");
    hbrowser.stop();

    if (hbrowser.currentURI.spec == location)
	return get_and_show_tooltip(link, hbrowser.contentDocument, fragment, 1);

    hbrowser.removeAttribute("hidden");
    tooltipProgressListener.link = link;
    tooltipProgressListener.idref = fragment;
    //tooltipProgressListener.location = location;
    try {
	hbrowser.loadURI(location, null, null);
    } catch (e) {
    }
    return null;
}

var tooltipProgressListener = {
    onLocationChange: function(webProgress, request, location) {
    },

    onProgressChange: function(webProgress, request, curSelfProgress,
			       maxSelfProgress, curTotalProgress,
			       maxTotalProgress) {
    },

    onSecurityChange: function(webProgress, request, state) {
    },

    onStateChange: function(webProgress, request, stateFlags, status) {
	if ((stateFlags & CIWPL.STATE_START) &&
	    (stateFlags & CIWPL.STATE_IS_DOCUMENT)) {
	    let hbrowser = document.getElementById("hidden_browser_tt");
	    if (webProgress.DOMWindow != hbrowser.contentWindow)
		return;
	    this.ok = true;
	} else if ((stateFlags & CIWPL.STATE_STOP) &&
		   (stateFlags & CIWPL.STATE_IS_DOCUMENT)) {
	    if (!this.ok)
		return;

	    let hbrowser = document.getElementById("hidden_browser_tt");
	    if (webProgress.DOMWindow != hbrowser.contentWindow)
		return;

	    hbrowser.setAttribute("hidden", "true");

	    /*
	    request.QueryInterface(Components.interfaces.nsIChannel);
	    if (request.URI.spec != this.location) {
		debug("QERROR");
		return;
	    }
	     */

	    let doc = webProgress.DOMWindow.document;
	    get_and_show_tooltip(this.link, doc, this.idref, 3);
	    this.link = null;
	    this.idref = null;
	    //this.location = null;
	}
    },

    onStatusChange: function(webProgress, request, status, message) {
    },

    QueryInterface: function(aIID) {
	if (aIID.equals(Components.interfaces.nsIWebProgressListener) ||
	    aIID.equals(Components.interfaces.nsISupportsWeakReference))
	    return this;
	throw Components.results.NS_NOINTERFACE;
    }
};

var myPrefObserver = {
  register: function() {
      let prefService = Components.classes["@mozilla.org/preferences-service;1"].
	  getService(Components.interfaces.nsIPrefService);
      this._branch = prefService.getBranch("extensions.lucifox.");
      this._branch.addObserver("", this, false);
  },

  unregister: function() {
      if (!this._branch)
	  return;
      this._branch.removeObserver("", this);
  },

  observe: function(aSubject, aTopic, aData) {
      if (aTopic != "nsPref:changed")
	  return;

      let prefs = Components.classes["@mozilla.org/preferences-service;1"].
	  getService(Components.interfaces.nsIPrefService).
	  getBranch("extensions.lucifox.");
      if (aData == "viewer.border") {
	  view_border = prefs.getBoolPref("viewer.border");
	  set_view_mode(null);
      } else if (aData == "viewer.columnAdjustment") {
	  column_adjustment = prefs.getCharPref("viewer.columnAdjustment");
      } else if (aData == "viewer.columnAdjustment") {
	  column_adjustment = prefs.getCharPref("viewer.columnAdjustment");
      } else if (aData == "viewer.enableExtraTools") {
	  set_extra_bar(prefs.getBoolPref("viewer.enableExtraTools"));
      } else if (aData == "viewer.enableNavigationBar") {
	  set_navigation_bar(prefs.getBoolPref("viewer.enableNavigationBar"));
      }
  }
};

var fakeControlService = {
    init: function(win) {
    },

    online: function() {
    },

    add_tab: function(location) {
	debug("ADD_TAB: " + location);
	let mainWindow = window.QueryInterface(Components.interfaces.nsIInterfaceRequestor)
	    .getInterface(Components.interfaces.nsIWebNavigation)
	    .QueryInterface(Components.interfaces.nsIDocShellTreeItem)
	    .rootTreeItem
	    .QueryInterface(Components.interfaces.nsIInterfaceRequestor)
	    .getInterface(Components.interfaces.nsIDOMWindow);
	let tab = mainWindow.gBrowser.addTab(location);

	let prefs = Components.classes["@mozilla.org/preferences-service;1"].
	    getService(Components.interfaces.nsIPrefService).
	    getBranch("browser.");
	if (!prefs.getBoolPref("tabs.loadInBackground"))
	    mainWindow.gBrowser.selectedTab = tab;
    },

    add_epub_tab: function(location, root) {
	return this.add_tab("chrome://lucidor/content/viewepub.xul?uri=" +
			    encodeURIComponent(location) + "&root=" + root);
    },

    add_web_tab: function(location) {
	return this.add_tab(location);
    },

    add_web_tab_post: function(location, post) {
	// FIXME
	let stringStream = Components.classes["@mozilla.org/io/string-input-stream;1"].
	    createInstance(Components.interfaces.nsIStringInputStream);
	stringStream.data = post;
	postData = Components.classes["@mozilla.org/network/mime-input-stream;1"].
	    createInstance(Components.interfaces.nsIMIMEInputStream);
	postData.addHeader("Content-Type", "application/x-www-form-urlencoded");
	postData.addContentLength = true;
	postData.setData(stringStream);

	let mainWindow = window.QueryInterface(Components.interfaces.nsIInterfaceRequestor)
	    .getInterface(Components.interfaces.nsIWebNavigation)
	    .QueryInterface(Components.interfaces.nsIDocShellTreeItem)
	    .rootTreeItem
	    .QueryInterface(Components.interfaces.nsIInterfaceRequestor)
	    .getInterface(Components.interfaces.nsIDOMWindow);
	mainWindow.gBrowser.webNavigation.loadURI(location, null, null,
						  postData, null);
    },

    open_epub_location: function(location) {
    },

    open_location: function(location) {
    },

    update_bar: function() {
	return;
	/*
	let button = document.getElementById("forward-button");
	if (browser.canGoForward)
	    button.disabled = false;
	else
	    button.disabled = true;

	button = document.getElementById("back-button");
	if (browser.canGoBack)
	    button.disabled = false;
	else
	    button.disabled = true;
	*/
    },

    update_tab_title: function(win) {
    },

    toggle_tabbar: function() {
    },
};

function lucifox_save_file() {
    let file = get_save_file();

    let fp = Components.classes["@mozilla.org/filepicker;1"].
	createInstance(Components.interfaces.nsIFilePicker);
    let sharedbundle = document.getElementById("lucifox-shared_strings");
    let text = sharedbundle.getString("save_book");
    fp.init(window, text, Components.interfaces.nsIFilePicker.modeSave);

    text = sharedbundle.getString("books");
    fp.appendFilter(text, "*.epub ; *.EPUB");
    //text = sharedbundle.getString("all_files");
    //fp.appendFilter(text, "*");
    fp.appendFilters(fp.filterAll);

    let defaultString = get_save_name();
    if (!defaultString)
	defaultString = file.leafName;
    if (!defaultString)
	defaultString = "book.epub";

    if ((defaultString.length < 5) ||
	(defaultString.substring(defaultString.length - 5).toLowerCase()
	 != ".epub")) {
	defaultString += ".epub";
    }

    fp.defaultString = defaultString;
    fp.defaultExtension = "epub";

    try {
	let prefs = Components.classes["@mozilla.org/preferences-service;1"].
	    getService(Components.interfaces.nsIPrefService).
	    getBranch("extensions.lucifox.");
	let dir = prefs.getComplexValue("lucifox.folder",
					Components.interfaces.nsIFile);
	fp.displayDirectory = dir;
    } catch (e) {
    }

    function my_save_file(targetFile, replace) {
	debug("my_save_file");
	if (!replace && targetFile.exists())
	    return;
	// FIXME honour "replace", it is now OS specific, ok on GNU/Linux
	try {
	    file.copyTo(targetFile.parent, targetFile.leafName);
	} catch (e) {
	    let bundle = document.getElementById("lucifox-strings");
	    let text = bundle.getString("save_error");
	    let nb = document.getElementById("notification_box");
	    nb.appendNotification(text, "save_err", null,
				  nb.PRIORITY_CRITICAL_MEDIUM, [], null);
	}
    }
    let save_it = my_save_file;
    if (!file)
	save_it = browser.contentWindow.save_file;

    let callback = {
	done: function(result) {
	    if (result == fp.returnOK || result == fp.returnReplace) {
		try {
		    let prefs = Components.classes["@mozilla.org/preferences-service;1"].
			getService(Components.interfaces.nsIPrefService).
			getBranch("extensions.lucifox.");

		    if (!prefs.prefHasUserValue("lucifox.folder"))
			prefs.setComplexValue("lucifox.folder",
					      Components.interfaces.nsIFile,
					      fp.file.parent);
		} catch (e) {
		}
		if (result == fp.returnOK)
		    save_it(fp.file, false);
		else if (result == fp.returnReplace)
		    save_it(fp.file, true);
	    }
	}
    };

    if (fp.open != undefined)
	fp.open(callback);
    else
	callback.done(fp.show());
}

function lucidor_event(evt) {
    let cmd = evt.target.getAttribute("command");
    if (cmd == "previous_part") {
	show_previous_part();
    } else if  (cmd == "next_part") {
	show_next_part();
    } else if  (cmd == "previous_column") {
	previous_column();
    } else if  (cmd == "next_column") {
	next_column();
    }
    document.documentElement.removeChild(evt.target);
}

function add_music_object(obj) {
    debug("add_music_object");
    if (!obj.hasAttribute("data"))
	return;
    let data = obj.getAttribute("data");

    let ioService = Components.classes["@mozilla.org/network/io-service;1"].
	getService(Components.interfaces.nsIIOService);
    let baseuri = ioService.newURI(obj.ownerDocument.baseURI, "UTF-8", null).
	QueryInterface(Components.interfaces.nsIURL);
    let uri = ioService.newURI(data, "UTF-8", baseuri);

    add_music(obj, uri.spec);
}

function add_music(parent, music) {
    debug("add_music: " + music);

    //let canv = document.createElementNS(XHTML_NS, "canvas");
    let canv = document.createElementNS(XHTML_NS, "div");
    parent.appendChild(canv);
    //let canvas_sel = "canvas_1";
    //canv.setAttribute("id", canvas_sel);

    let url = music;

    let xhr = new XMLHttpRequest();
    xhr.onload = function() {
	let oSerializer = new XMLSerializer();
	let st = oSerializer.serializeToString(xhr.responseXML);
	do_music(xhr.responseXML);
    }
    xhr.open("GET", url);
    xhr.send();

    function do_music(music) {
	Components.utils.import("chrome://lucidor/content/raphael.jsm");
	Raphael.setWindow({document: document});
	Components.utils.import("chrome://lucidor/content/vexflow.jsm");

	//let ctx = Vex.Flow.Renderer.getCanvasContext(canv, 680, 200);
	let ctx = Vex.Flow.Renderer.getRaphaelContext(canv, 680, 200);
	ctx.scale(0.6, 0.6);
	ctx.setFillStyle("#221");
	ctx.setStrokeStyle("#221");
	let doc = new Vex.Flow.MusicXML.Document(music);
	let staffSystem = new Vex.Flow.MusicXML.StaffSystem(doc, {x: 20, y: 0, width: 2400});
	staffSystem.draw(ctx);
	staffSystem.drawContents(ctx);
    }
}

function set_luci_node(node) {
    node.setAttribute("class", "lucitemp");
}

function is_luci_node(node) {
    return (node.nodeType == node.ELEMENT_NODE &&
	    node.hasAttribute("class") &&
	    node.getAttribute("class") == "lucitemp")
}

function get_root(doc) {
    if (!doc)
	doc = browser.contentDocument;
    if (doc.documentElement) {
	if (doc.documentElement.localName == "x-ordbrand-root")
	    return doc.documentElement.firstChild.firstChild;
	else
	    return doc.documentElement;
    }
    return doc;
}

function target_to_domrange(target, root) {
    /*
    debug("target_to_domrange: " + target.fixptr + " ; " + target.start +
	  " ; " + target.pstart + " ; " + target.epubcfi +
	  " ; " + target.xpointer);
    */
    if (!root) {
	root = get_root(null);
    }

    let range = null;
    if (target.fixptr) {
	range = fixptr_to_domrange(target.fixptr, root);
    }
    if (!range && target.start) {
	range = xpathlocation_to_domrange(target, root);
    }
    if (!range && target.pstart) {
	range = adelocation_to_domrange(target, root);
    }
    if (!range && target.epubcfi) {
	range = epubcfi_to_domrange(target.epubcfi);
    }
    if (!range && target.xpointer) {
	range = xpointer_to_domrange(target.xpointer, root);
    }
    return range;
}

function xpointer_to_domrange(xptr, root) {
    /*
    if (xptr.substring(0, 23) == "xpointer(string-range(/") {
	xptr = "string-range(/x-ordbrand-root[1]/x-ordbrand-glue[1]/" + xptr.substring(23);
    } else if (xptr.substring(0, 10) == "xpointer(/") {
	xptr = "xpointer(/x-ordbrand-root[1]/x-ordbrand-glue[1]" + xptr.substring(10);
    }
    */
    xptr = xptr.replace("(/", "(/x-ordbrand-root[1]/x-ordbrand-glue[1]/");

    /*
      let xpeval = Components.classes["@mozilla.org/xmlextras/xpointerevaluator;1"].createInstance(Components.interfaces.nsIXPointerEvaluator);
      let xpres = xpeval.evaluate(doc, xptr);
      if (xpres.length == 0)
	return;
      let dr = xpres.item(0);
    */

    let dr;
    try {
	let xpos = Components.classes["@lucidor.org/xpointerService;1"].getService().wrappedJSObject;
	dr = xpos.parseXPointerToRange(xptr, doc);
	debug(dr + ":" + dr.startOffset);
    } catch (e) {
	debug("Couldn't resolve: " + e);
	return null;
    }

    return dr;
}

function domrange_to_xpointer(range) {
    try {
	let xpos = Components.classes["@lucidor.org/xpointerService;1"].getService().wrappedJSObject;
	let doc = range.startContainer.ownerDocument;
	xpo = xpos.createXPointerFromRange(range, doc);
	xpo = xpo.replace("/x-ordbrand-root[1]/x-ordbrand-glue[1]/", "/");
	return xpo;
    } catch (e) {
    }
    return null;
}

function adelocation_to_domrange(adeloc, root) {
    debug("adelocation_to_domrange: " + adeloc.pstart + " " + adeloc.pend);
    let doc = root.ownerDocument;
    let point = parse_adepoint(adeloc.pstart);
    let node = evaluate_adepoint(point, root);
    if (!node) {
	debug("ade start error");
	return null;
    }
    let offset = utf8_to_utf16_offset(node, point.offset);
    let dr = doc.createRange();
    try {
	dr.setStart(node, offset);
    } catch (e) {
	return null
    }
    if (adeloc.pend) {
	point = parse_adepoint(adeloc.pend);
	node = evaluate_adepoint(point, root);
	if (!node) {
	    debug("ade end error");
	    return null;
	}
	offset = utf8_to_utf16_offset(node, point.offset);
	try {
	    dr.setEnd(node, offset);
	} catch (e) {
	    return null
	}
    } else {
	dr.collapse(true);
    }
    //debug(dr + ":" + dr.startOffset);
    return dr;
}

// Location syntax used by Adobe Reader Mobile SDK (Digital Editions, Nook)
function parse_adepoint(adestring) {
    debug("parse_adepoint: " + adestring);
    let re = /((.+)#)?point\(([\/0-9]+)(:([0-9]+))?\)/;
    let match = re.exec(adestring);
    if (!match)
	return null;
    return { part: match[2],
	     node: match[3],
	     offset: match[5] };
}

// Evaluate adepoint to node (but not offset)
function evaluate_adepoint(adepoint, root) {
    debug("evaluate_adepoint: " + adepoint.node);
    let parts = adepoint.node.split("/");
    parts.shift();
    let node = root.parentNode;
    for (let part in loop(parts)) {
	let pa = parseInt(part, 10);
	debug("node: " + pa + " " + node.localName + " " + node.childNodes.length);
	if (pa === NaN || pa < 1)
	    return null;
	//node = node.childNodes[pa - 1];
	let n = 0;
	for (let ch in loop(node.childNodes)) {
	    if (ch.nodeType == ch.ELEMENT_NODE || ch.nodeType == ch.TEXT_NODE) {
		n++;
		if (n == pa) {
		    node = ch;
		    break;
		}
	    }
	}
	if (n != pa)
	    return null;
    }
    return node;
}

function utf16_to_utf8_offset(node, offset) {
    if (offset == 0 || offset == undefined)
	return 0;

    if (node.nodeType == node.TEXT_NODE) {
	let converter = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"].
	    createInstance(Components.interfaces.nsIScriptableUnicodeConverter);
	converter.charset = "UTF-8";

	if (node.nodeValue.length >= offset) {
	    let data = converter.convertToByteArray(node.nodeValue.substring(0, offset), {});
	    return data.length;
	}
    }

    debug("utf16_to_utf8_offset error");
    return offset;
}

function utf8_to_utf16_offset(node, offset) {
    debug("utf8_to_utf16_offset: " + offset);
    if (offset == 0 || offset == undefined)
	return 0;

    //debug("value: " + node.nodeValue);
    if (node.nodeType == node.TEXT_NODE) {
	let converter = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"].
	    createInstance(Components.interfaces.nsIScriptableUnicodeConverter);
	converter.charset = "UTF-8";

	let data = converter.convertToByteArray(node.nodeValue.substring(0, offset), {});
	//debug("data: " + data.length + " " + data);
	if (data.length >= offset) {
	    let chars = converter.convertFromByteArray(data, offset);
	    //debug("len: " + node.nodeValue.length + " " + chars.length);
	    return chars.length;
	}
    }

    debug("utf8_to_utf16_offset error");
    return offset;
}

function utf16_to_ucs_offset(node, offset) {
    return offset;
    /*
    if (offset == 0 || offset == undefined)
	return 0;

    let off = 0;
    if (node.nodeType == node.ELEMENT_NODE)
	node = node.firstChild;
    for (; node; node = node.nextSibling) {
	if (node.nodeType == node.TEXT_NODE) {
	    let str = node.nodeValue;
	    for (let i = 0; i < str.length; i++) {
		let cha = str.charCodeAt(i);
		if (cha >= 0xD800 && cha <= 0xDFFF) {	// Surrogate
		    if (cha >= 0xDC00)	// Low surrogate
			off++;
		} else {
		    off++;
		}
		offset--;
		if (offset == 0)
		    break;
	    }
	}
    }
    if (offset != 0) {
	debug("utf16_to_ucs_offset error");
    }
    return off;
    */
}

function ucs_to_utf16_offset(node, offset) {
    return offset;
    /*
    if (offset == 0 || offset == undefined)
	return 0;

    let off = 0;
    if (node.nodeType == node.ELEMENT_NODE)
	node = node.firstChild;
    for (; node; node = node.nextSibling) {
	if (node.nodeType == node.TEXT_NODE) {
	    let str = node.nodeValue;
	    for (let i = 0; i < str.length; i++) {
		let cha = str.charCodeAt(i);
		if (cha >= 0xD800 && cha <= 0xDFFF) {	// Surrogate
		    if (cha >= 0xDC00)	// Low surrogate
			offset--;
		} else {
		    offset--;
		}
		off++;
		if (offset == 0)
		    break;
	    }
	}
    }
    if (offset != 0) {
	debug("ucs_to_utf16_offset error");
    }
    return off;
    */
}

function domrange_to_adelocation(range) {
    if (!range || !range.startContainer)
	return null;

    //let doc = browser.contentDocument;
    let doc = range.startContainer.ownerDocument;
    let root = get_root(doc);
    let adepoint = get_adepoint(range.startContainer, range.startOffset, root);
    //debug("adepoint start: " + adepoint);
    let adeloc = { pstart: adepoint };
    if (!range.collapsed) {
	adepoint = get_adepoint(range.endContainer, range.endOffset, root);
	//debug("adepoint end: " + adepoint);
	adeloc.pend = adepoint;
    }
    return adeloc;
}

// Get adepoint in string form
function get_adepoint(target, offset, root) {
    let path = "";
    for (let tar = target;
	 tar != root.parentNode && tar != target.ownerDocument;
	 tar = tar.parentNode) {
	if (is_luci_node(tar))
	    continue;
	let count = 1;
	for (let el = tar.previousSibling; el; el = el.previousSibling) {
	    if (el.nodeType == el.ELEMENT_NODE) {
		if (!is_luci_node(el))
		    count++;
	    } else if (el.nodeType == el.TEXT_NODE) {
		count++;
	    }
	}
	path = "/" + count + path;
    }
    return "point(" + path + ":" + utf16_to_utf8_offset(target, offset) + ")";
}

function xpathlocation_to_domrange(xploc, root) {
    let doc = root.ownerDocument;

    /*
    let xpe = new XPathEvaluator();
    let result = xpe.evaluate(//"/x-ordbrand-root[1]/x-ordbrand-glue[1]" +
			      range.start.substring(1),
			      get_root(doc),
			      xpe.createNSResolver(doc), 9, null);
    let node = result.singleNodeValue;
    */

    let node = evaluate_xpath(xploc.start, root);
    if (!node) {
	debug("xpath start error");
	return null;
    }
    let pair = element_to_node_offset(node, xploc.startoffset);
    let dr = doc.createRange();
    dr.setStart(pair.node, pair.offset);

    if (xploc.end) {
	node = evaluate_xpath(xploc.end, root);
	if (!node) {
	    debug("xpath end error");
	    return null;
	}
	pair = element_to_node_offset(node, xploc.endoffset);
	dr.setEnd(pair.node, pair.offset);
    } else {
	dr.collapse(true);
    }
    //debug(dr + ":" + dr.startOffset);
    return dr;
}

function domrange_to_xpathlocation(range) {
    if (!range || !range.startContainer)
	return null;

    let pair = node_to_element_offset(range.startContainer, range.startOffset);
    let xp = get_element_xpath(pair.element);
    let xploc = { start: xp, startoffset: pair.offset };
    debug("startpair: " + pair.element + " " + pair.offset);
    debug("xpath: " + xp);
    if (!range.collapsed) {
	pair = node_to_element_offset(range.endContainer, range.endOffset);
	xp = get_element_xpath(pair.element);
	xploc.end = xp;
	xploc.endoffset = pair.offset;
	debug("endpair:" + pair.element + " " + pair.offset);
	debug("xpath: " + xp);
    }
    return xploc;
}

function get_node_offset(node) {
    let offset = 0;
    for (let n = node.parentNode.firstChild; n; n = n.nextSibling) {
	if (n == node)
	    return offset;
	if (n.nodeType == n.TEXT_NODE)
	    offset += n.nodeValue.length;
    }
    return offset;
}

// Convert textnode+offset to element+offset
function node_to_element_offset(node, offset) {
    if (node.nodeType == node.TEXT_NODE) {
	let off = offset;
	for (let con = node.previousSibling; con; con = con.previousSibling) {
	    if (con.nodeType == con.TEXT_NODE)
		off += con.nodeValue.length;
	}
	return { element: node.parentNode,
		 offset: off };
    } else if (node.nodeType == node.ELEMENT_NODE) {
	let off = 0;
	for (let i = 0; i < offset; i++) {
	    let con = node.childNodes[i];
	    //debug("con: " + con);
	    if (con.nodeType == con.TEXT_NODE)
		off += con.nodeValue.length;
	}
	return { element: node,
		 offset: off };
    } else {
	return { element: node.parentNode };
    }
}

// Convert element+offset to (text)node+offset
function element_to_node_offset(element, offset) {
    //if (element.nodeType != element.ELEMENT_NODE)
    //	return { node: element, offset: offset };
    let off = 0;
    for (let node = element.firstChild; node; node = node.nextSibling) {
	if (node.nodeType == node.TEXT_NODE) {
	    if (offset == 0) {
		return { node: node, offset: 0 };
	    } else if (off + node.nodeValue.length == offset) {
		// Offset points to the end of a text node
		// Find beginning of next text node
		for (let n = node.nextSibling; n; n = n.nextSibling)
		    if (n.nodeType == n.TEXT_NODE)
			return { node: n, offset: 0 };
		// This should not happen
		return { node: node, offset: offset - off };
	    } else if (off + node.nodeValue.length > offset) {
		return { node: node, offset: offset - off };
	    } else {
		off += node.nodeValue.length;
	    }
	}
    }
    // This should not happen if offset > 0
    return { node: element, offset: offset };
    //return { node: element, offset: element.childNodes.length };
}

function evaluate_xpath(xpath, root) {
    let re = /(.+)\[([0-9]+)\]/;
    let parts = [];
    let spas = xpath.split("/");
    for (let spa in loop(spas)) {
	if (spa == "")
	    continue;
	let match = re.exec(spa);
	if (!match)
	    return null;
	parts.push([match[1], match[2]]);
    }
    let node = root.parentNode;
    for (let part in loop(parts)) {
	if (part[1] < 1)
	    return null;
	let n = 0;
	for (let ch = node.firstChild; ch; ch = ch.nextSibling) {
	    if ((ch.nodeType == ch.ELEMENT_NODE) &&
		(ch.nodeName == part[0])) {
		n++;
		if (n == part[1]) {
		    node = ch;
		    break;
		}
	    }
	}
	if (n != part[1])
	    return null;
    }
    return node;
}

function get_element_xpath(target) {
    let xpath = "";
    for (let tar = target; tar != target.ownerDocument; tar = tar.parentNode) {
	if (is_luci_node(tar))
	    continue;
	let count = 1;
	for (let el = tar.previousSibling; el; el = el.previousSibling) {
	    if ((el.nodeType == el.ELEMENT_NODE) &&
		(el.localName == tar.localName) &&
		(el.namespaceURI == tar.namespaceURI)) {
		count++;
	    }
	}
	let name = tar.localName;
	if (tar.prefix)
	    name = tar.prefix + ":" + name;
	xpath = "/" + name + "[" + count + "]" + xpath;
    }
    xpath = xpath.replace("/x-ordbrand-root[1]/x-ordbrand-glue[1]/", "/");
    return xpath;
}

function get_element_fixptr(target) {
    let doc = target.ownerDocument.documentElement.firstChild;
    if (doc.localName != 'x-ordbrand-glue')
	doc = target.ownerDocument;
    let fixptr = "";
    for (let tar = target; tar != doc; tar = tar.parentNode) {
	if (is_luci_node(tar))
	    continue;
	let count = 1;
	for (let el = tar.previousSibling; el; el = el.previousSibling)
	    if (el.nodeType == el.ELEMENT_NODE && !is_luci_node(el))
		count++;
	fixptr = "/" + count + fixptr;
    }
    return fixptr;
}

function domrange_to_fixptr(range) {
    if (!range || !range.startContainer)
	return null;

    let pair = node_to_element_offset(range.startContainer, range.startOffset);
    let fixptr = get_element_fixptr(pair.element);
    if (pair.offset) {
	let offset = utf16_to_ucs_offset(pair.element, pair.offset)
	fixptr += "(" + offset + ")";
    }

    if (!range.collapsed) {
	pair = node_to_element_offset(range.endContainer, range.endOffset);
	let fixp = get_element_fixptr(pair.element);
	if (pair.offset) {
	    let offset = utf16_to_ucs_offset(pair.element, pair.offset)
	    fixp += "(" + offset + ")";
	}
	fixptr += "," + fixp;
    }
    //debug("fixptr: " + fixptr);
    return fixptr;
}

function fixptr_to_domrange(fixptr, root) {
    debug("fixptr_to_domrange: " + fixptr);
    //debug("root: " + root.localName);
    let re = /([^\/(]+)?(\/[\/0-9]+)?(\(([0-9]+)\))?/;
    let fixp = [];
    for (let fi in loop(fixptr.split(","))) {
	let match = re.exec(fi);
	if (!match)
	    continue;
	let mid = match[1];
	let xp = match[2];
	let offset = match[4];
	if (!offset)
	    offset = 0;
	fixp.push({ mid: mid,
		    path: xp.split("/"),
		    offset: offset });
    }

    for (let fi in loop(fixp)) {
	let node = root.parentNode;
	if (fi.mid !== undefined) {
	    node = root.ownerDocumnet.getElementById(fi.mid);
	    if (!node) {
		debug("error 1");
		return null;
	    }
	}

	for (let path in loop(fi.path)) {
	    if (!path)
		continue;
	    let n = parseInt(path, 10);
	    //debug("path: " + n);
	    for (let ch = node.firstChild; ch; ch = ch.nextSibling) {
		//debug("node: " + ch.localName);
		if ((ch.nodeType == ch.ELEMENT_NODE) && !is_luci_node(ch)) {
		    n--;
		    if (n == 0) {
			node = ch;
			break;
		    }
		}
	    }
	    if (n != 0) {
		debug("error 2: " + n);
		return null;
	    }
	}
	fi.node = node;
    }

    //debug("fixp[0]: " + fixp[0].offset);
    let dr = root.ownerDocument.createRange();
    let offset = ucs_to_utf16_offset(fixp[0].node, fixp[0].offset);
    let pair = element_to_node_offset(fixp[0].node, offset);
    dr.setStart(pair.node, pair.offset);
    if (fixp.length == 2) {
	//debug("fixp[1]: " + fixp[1].offset);
	offset = ucs_to_utf16_offset(fixp[1].node, fixp[1].offset);
	pair = element_to_node_offset(fixp[1].node, offset);
	dr.setEnd(pair.node, pair.offset);
    } else {
	dr.collapse(true);
    }

    return dr;
}

function compare_fixptr(a, b) {
    let aloc = a.split(",");
    aloc = aloc[0].split("(");
    let aoff = -1;
    if (aloc.length >= 2)
	aoff = parseInt(aloc[1].replace(")", ""), 10);
    aloc = aloc[0].split("/").map(function(a) {
	return parseInt(a, 10);
    });

    let bloc = b.split(",");
    bloc = bloc[0].split("(");
    let boff = -1;
    if (bloc.length >= 2)
	boff = parseInt(bloc[1].replace(")", ""), 10);
    bloc = bloc[0].split("/").map(function(a) {
	return parseInt(a, 10);
    });

    for (let i = 0; i < Math.min(aloc.length, bloc.length); i++) {
	if (aloc[i] < bloc[i])
	    return -1;
	if (aloc[i] > bloc[i])
	    return 1;
    }

    if (aloc.length < bloc.length)
	return -1;
    if (aloc.length > bloc.length)
	return 1;

    if (aoff < boff)
	return -1;
    if (aoff > boff)
	return 1;
    return 0;
}

function compare_epubcfi(a, b) {
    //debug("compare_epubcfi: " + a + " " + b);
    let acfi = new EpubCFI(null);
    if (!acfi.parse(a))
	return 0;	// Error
    let bcfi = new EpubCFI(null);
    if (!bcfi.parse(b))
	return 0;	// Error
    return acfi.compare(bcfi);
}

function indirectionResolver(book, active) {
    this.book = book;
    this.active = active;
}

indirectionResolver.prototype = {
    active: false,

    resolve: function(node) {
	//debug("indirectionResolver.resolve: " + node);

	let res = { href: null,
		    node: null };
	if (node == null) {
	    res.node = this.book.opf.domroot;
	} else if (node.nodeType != node.ELEMENT_NODE) {
	    // error
	} else {
	    if (node.localName == "itemref") {
		let idref = node.getAttribute("idref");
		//debug("resolve_indirection spine: " + idref);
		res.href = this.book.idrefToHref(idref);
		//debug(". " + res.href + " " + href_to_local_href(browser.currentURI.spec));
		if (this.active)
		    if (res.href == href_to_local_href(browser.currentURI.spec))
			res.node = get_root(browser.contentDocument);
	    } else if (node.localName == "item") {	// Non-standard
		// get manifest item
		let href = node.getAttribute("href");
		//debug("resolve_indirection manifest: " + href);
		res.href = href;
		if (this.active)
		    if (res.href == href_to_local_href(browser.currentURI.spec))
			res.node = get_root(browser.contentDocument);
	    } else if (node.localName == "iframe" ||
		       node.localName == "embed") {
		// Ignore for now
		// src attribute
	    } else if (node.localName == "object") {
		// Ignore for now
		// data attribute
	    } else if (node.localName == "image" ||
		       node.localName == "use") {
		// Ignore for now
		// xlink:href attribute
	    } else {
		// error
	    }
	}
	return res;
    }
}

function epubcfi_to_domrange(epubcfi) {
    debug("epubcfi_to_domrange: " + epubcfi);
    let cfi = new EpubCFI(g_indirectionResolver);
    if (!cfi.parse(epubcfi))
	return null;
    return cfi.resolve();
}

function epubcfi_to_part(epubcfi) {
    debug("epubcfi_to_part: " + epubcfi);
    let cfi = new EpubCFI(g_indirectionResolver);
    if (!cfi.parse(epubcfi))
	return null;
    return cfi.resolve_part();
}

function domrange_to_epubcfi(range) {
    if (!range || !range.startContainer)
	return null;

    let doc = range.startContainer.ownerDocument;
    let root = get_root(doc);

    let cfi = new EpubCFI(null);
    let item = get_spine_item();
    if (!item)
	item = get_manifest_item();
    //debug("item: " + item + " " + item.element);
    if (!item)
	return null;
    cfi.create_from_element(item.element, book.opf.domroot);
    cfi.append_steps(cfi.CFI_PATH, [ cfi.get_indirection_step() ]);
    cfi.create_from_domrange(range, root);

    //debug("cfi: " + cfi.toString());
    return cfi.toString();
}
