/*
 *  Lucidor - e-book reader
 *  Copyright © 2008-2017  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("chrome://lucidor/content/utils.jsm");
Components.utils.import("chrome://lucidor/content/epubcfi.jsm");
Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

function bookmanager() {
    this.wrappedJSObject = this;
    this.books = {};
    this.files = [];
    this.lastid = 0;
    this.books['css'] = { type: 'css' };
    this.books['notes'] = { type: 'notes' };
    this.books['printnotes'] = { type: 'printnotes' };
    this.bundle = null;
    this.tempdir = null;
}

bookmanager.prototype = {
    classDescription: "bookmanager",
    classID: Components.ID("{c9ff0f94-2b9a-43bb-96ec-b86ddef81c54}"),
    contractID: "@lucidor.org/bookmanagerService;1",
    _xpcom_factory: bookmanagerFactory,
    QueryInterface: XPCOMUtils.generateQI([Components.interfaces.nsISupports]),

    getTempdir: function() {
	if (!this.tempdir) {
	    let tempdir = Components.classes["@mozilla.org/file/directory_service;1"].
		getService(Components.interfaces.nsIProperties).
		get("TmpD", Components.interfaces.nsIFile);
	    //this.tempdir.append("lucifox");
	    tempdir.append("l" + (new Date()).getTime().toString(36));
	    tempdir.createUnique(Components.interfaces.nsIFile.DIRECTORY_TYPE,
				 parseInt("0700", 8));
	    this.tempdir = tempdir;
	}
	return this.tempdir.clone();
    },

    tryRemoveTempdir: function() {
	debug("tryRemoveTempdir");
	try {
	    if (this.tempdir && this.tempdir.isDirectory())
		this.tempdir.remove(false);
	} catch (e) {
	}
    },

    removeTempdir: function() {
	debug("removeTempdir");
	try {
	    if (this.tempdir && this.tempdir.isDirectory())
		this.tempdir.remove(true);
	} catch (e) {
	}
    },

    getTempfile: function(suffix) {
	let fileid = (this.lastid++).toString();
	fileid += (0x100000 + ((new Date()).getTime() & 0x7fffff)).toString();
	let file = this.getTempdir();
	file.append("l" + fileid + suffix);
	return file;
    },

    addFile: function(window, filespec, filename) {
	debug("addFile " + filespec);
	this.files.push([window, filespec, filename]);
    },

    getFile: function(window) {
	debug("getFile");
	let filepair = null;
	for (let i = 0; i < this.files.length; i++) {
	    if (this.files[i][0] === window) {
		filepair = this.files[i].slice(1);
		this.files.splice(i, 1);
		break;
	    }
	}
	return filepair;
    },

    setBundle: function(bundle) {
	this.bundle = bundle;
    },

    addBook: function(type, book) {
	let bookid;
	do {
	    bookid = "w" + (this.lastid++).toString();
	    bookid += (0x100000 + ((new Date()).getTime() & 0x7fffff)).toString();
	    //bookid = Math.floor(Math.random() * 100000000).toString();
	} while (this.books[bookid] != undefined);

	this.books[bookid] = { type: type, book: book };
	debug("bookmanager.addBook: bookid=" + bookid);
        return bookid;
    },

    removeBook: function(bookid) {
	this.books[bookid] = undefined;
    },

    getBook: function(bookid) {
	return this.books[bookid].book;
    },

    getChannel: function(uri) {
	debug("bookmanager.getChannel: " + uri.spec);
	let entry = this.books[uri.host];
	if (!entry)
	    throw "Error: No such book";

	let frag = '';
	let path = decodeURIComponent(uri.path);
	if (path) {
	    let i = 0;
	    while (path.charAt(i) == "/")
		i++;
	    let j = path.indexOf("#", i);
	    if (j < 0)
		j = path.length;
	    else
		frag = path.substring(j + 1, path.length);
	    path = path.substring(i, j);
	}

	let instream = null;
	let mediatype = null;
	let charset = null;
	let length = 0;
	let newuri = uri;

	if (entry.type == "epub") {
	    let zipreader = entry.book.zipreader;

	    debug("PATH: " + path);
	    try {
		let item = entry.book.hrefToItem(path);
		debug("IDREF: " + item.id);
		let loader = entry.book.idrefToLoader(item.id);
		debug("NEW PATH: " + loader);
		path = loader[0];
	    } catch (e) {
		debug("Error: No such book entry");
	    }

	    // Normalize path
	    let ioService = Components.classes["@mozilla.org/network/io-service;1"].
		getService(Components.interfaces.nsIIOService);
	    let nuri = ioService.newURI("file://a/" + path, "UTF-8", null);
	    path = decodeURIComponent(nuri.path.substring(1));

	    if (!zipreader.hasEntry(path)) {
		debug("archive entry missing: " + path);
		entry.book.add_error("archive entry missing: " + path);
		if (path.substring(path.length - 4) == ".css") {
		    instream = Components.classes["@mozilla.org/io/string-input-stream;1"].
			createInstance(Components.interfaces.nsIStringInputStream);
		    instream.setData("", 0);
		    mediatype = "";
		} else
		    throw "Error: No such file entry";
	    } else {
		if (path == entry.book.rootfiles[entry.book.rootfile]) {
		    debug("bm epubcfi_to_part: " + frag);
		    let cfi = new EpubCFI(new indirectionResolver(entry.book,
								  false));
		    if (!cfi.parse(frag))
			throw "Error: Could not parse cfi";
		    let cfihref = cfi.resolve_part();
		    debug("cfihref: " + cfihref);
		    if (!cfihref)
			throw "Error: Could not resolve cfi";
		    path = cfihref;
		    let uspec = 'x-lucidor://' + uri.host + '/' +
			path + '#' + frag;
		    newuri = ioService.newURI(uspec, "UTF-8", null);
		}

		let item = entry.book.opf.manifest_href[path];
		if (!item) {
		    debug("manifest item missing: " + path);
		    entry.book.add_error("manifest item missing: " + path);
		    /* FIXME should not do this, only a workaround for
		       broken books (eg Death and the Senator) */
		    item = { mediatype: null };
		    //throw "Error: No such manifest item";
		} else if (item.encryption && !item.decrypt) {
		    throw "Error: encrypted";
		}
		mediatype = item.mediatype;
		if (mediatype == "application/vnd.adobe-page-template+xml")
		    throw "Error: adobe";

		/*
		let ios = Components.classes["@mozilla.org/network/io-service;1"].
		    getService(Components.interfaces.nsIIOService);
		let juri = 'jar:' + entry.book.location + '!/' + path;
		let ch = ios.newChannel(juri, null, null);
		*/

		instream = zipreader.getInputStream(path);
		//length = zipreader.getEntry(path).length;

		if (item.decrypt_font) {
		    // According to OCF 3.01 fonts must not compressed.
		    instream = item.decrypt_font(entry.book, item, instream);
		} else if (item.decrypt) {
		    instream = item.decrypt(entry.book, item, instream);

		    //let cs = Components.classes["@mozilla.org/streamConverters;1"].
		    //    getService(Components.interfaces.nsIStreamConverterService);
		    //instream = cs.convert(instream, "deflate", "uncompressed", null);
		    // According to OCF 3.0 section 2.5.2 encrypted data
		    // must be compressed.
		    instream = getDeflateStream(instream);
		}
	    }
	} else if (entry.type == "css") {
	    instream = Components.classes["@mozilla.org/io/string-input-stream;1"].
		createInstance(Components.interfaces.nsIStringInputStream);
	    instream.setData(get_css(), -1);
	    mediatype = "text/css";
	} else if (entry.type == "notes") {
	    if (path) {
		let ent = this.books[path];
		if (ent) {
		    let converter = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"].
			createInstance(Components.interfaces.nsIScriptableUnicodeConverter);
		    converter.charset = "UTF-8";
		    instream = converter.convertToInputStream(get_notes(ent.book));
		} else {
		    instream = Components.classes["@mozilla.org/io/string-input-stream;1"].
			createInstance(Components.interfaces.nsIStringInputStream);
		    instream.setData("", 0);
		}
	    } else {
		let converter = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"].
		    createInstance(Components.interfaces.nsIScriptableUnicodeConverter);
		converter.charset = "UTF-8";
		instream = converter.convertToInputStream(get_all_notes());
	    }
	    mediatype = "text/plain";
	    charset = "UTF-8";
	} else if (entry.type == "printnotes") {
	    let ent = "";
	    if (path)
		ent = this.books[path];
	    if (ent) {
		let converter = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"].
		    createInstance(Components.interfaces.nsIScriptableUnicodeConverter);
		converter.charset = "UTF-8";
		instream = converter.convertToInputStream(print_notes(ent.book, this.bundle));
	    } else {
		instream = Components.classes["@mozilla.org/io/string-input-stream;1"].
		    createInstance(Components.interfaces.nsIStringInputStream);
		instream.setData("", 0);
	    }
	    mediatype = "text/html";
	    charset = "UTF-8";
	} else
	    throw "Error: This will never happen";

        let channel = Components.classes["@mozilla.org/network/input-stream-channel;1"].createInstance(Components.interfaces.nsIInputStreamChannel);
	channel.contentStream = instream;
	channel.setURI(newuri);
	if (newuri != uri) {
	    channel.QueryInterface(Components.interfaces.nsIChannel);
	    channel.originalURI = uri;
	    channel.QueryInterface(Components.interfaces.nsIRequest);
	    channel.loadFlags |= Components.interfaces.nsIChannel.LOAD_REPLACE;
	}
	channel.QueryInterface(Components.interfaces.nsIChannel);
	if (length)
	    channel.contentLength = length;
	if (mediatype)
	    channel.contentType = mediatype;
	if (charset)
	    channel.contentCharset = charset;

	return channel;
    }
};

var serviceObject = new bookmanager();

var bookmanagerFactory = {
    createInstance: function(outer, aIID) {
	if (outer != null)
	    throw Components.results.NS_ERROR_NO_AGGREGATION;
	return serviceObject.QueryInterface(aIID);
    }
};

var components = [bookmanager];

if (XPCOMUtils.generateNSGetFactory)
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
else
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);

/*
function getBookId(bookid) {
    let parts = bookid.split(".");
    if (parts.length == 4) {
	bookid = ((parts[0] << 24) + (parts[1] << 16) +
		  (parts[2] << 8) + parts[3]).toString();
    }
    return bookid;
}
*/

function getDeflateStream(instream) {
    debug("getDeflateStream");
    let pipe = Components.classes["@mozilla.org/pipe;1"].
        createInstance(Components.interfaces.nsIPipe);
    pipe.init(false, false, 0, 0, null);
    debug("getDeflateStream 1");

    let blocker = Components.classes["@mozilla.org/binaryinputstream;1"].
        createInstance(Components.interfaces.nsIBinaryInputStream);
    blocker.setInputStream(pipe.inputStream);

    debug("getDeflateStream 3");

    let mylis = new listener(pipe);
    let cs = Components.classes["@mozilla.org/streamConverters;1"].
	getService(Components.interfaces.nsIStreamConverterService);
    let converter = cs.asyncConvertData("deflate", "uncompressed", mylis, null);
    try {
	converter.onDataAvailable(null, null, instream, 0, instream.available());
    } catch (e) {
	debug("error: " + e);
	return instream;
    }
    debug("getDeflateStream 5");

    try {
	blocker.read8();
    } catch (e) {
	debug("blocker failed");
    }
    debug("return: " + mylis.stream);
    return mylis.stream;
};

function listener(pipe) {
    this.pipe = pipe;
    try {
	let bookman = Components.classes["@lucidor.org/bookmanagerService;1"].
	    getService().wrappedJSObject;
	this.tempdir = bookman.getTempdir();
    } catch (e) {
	this.tempdir = null;
    }
};

listener.prototype = {
    onDataAvailable: function(request, context, inputStream, offset, count) {
	debug("onDataAvailable");
	if (!this.tempdir)
	    throw "Error: cannot create tempfile";
	let fil = this.tempdir.clone();
	fil.append("lucidor.ttf");
	fil.createUnique(fil.NORMAL_FILE_TYPE, parseInt("0600", 8));
 	let filout = Components.classes["@mozilla.org/network/file-output-stream;1"].
	    createInstance(Components.interfaces.nsIFileOutputStream);
	filout.init(fil, -1, -1, 0);

	let binin = Components.classes["@mozilla.org/binaryinputstream;1"].
            createInstance(Components.interfaces.nsIBinaryInputStream);
	binin.setInputStream(inputStream);
	let binout = Components.classes["@mozilla.org/binaryoutputstream;1"].
            createInstance(Components.interfaces.nsIBinaryOutputStream);
	binout.setOutputStream(filout);
	while (binin.available() > 0) {
	    let count = binin.available();
	    if (count > 1024)
		count = 1024;
	    let buf = binin.readByteArray(count);
	    binout.writeByteArray(buf, buf.length);
	}
	binout.close();
	filout.close();

 	let filin = Components.classes["@mozilla.org/network/file-input-stream;1"].
	    createInstance(Components.interfaces.nsIFileInputStream);
	filin.init(fil, -1, -1, 0);
	this.stream = filin;

	this.pipe.outputStream.close();
    },

    onStartRequest: function(aRequest, aContext) {
	debug("onStartRequest");
    },

    onStopRequest: function(aRequest, aContext, aStatusCode) {
	debug("onStopRequest");
    }
};

function get_css() {
    let prefs = Components.classes["@mozilla.org/preferences-service;1"].
	getService(Components.interfaces.nsIPrefService).
	getBranch("extensions.lucifox.");
    let name = prefs.getCharPref("style.name");
    let style = { background: '',
		  color: '',
		  lineheight: '',
		  fontsize: '',
		  fontfamily: '' };
    if (name == "custom") {
	style.background = prefs.getCharPref("style.background");
	style.color = prefs.getCharPref("style.color");
	style.fontsize = prefs.getCharPref("style.fontsize");
	if (style.fontsize)
	    style.fontsize += "pt";
	style.lineheight = prefs.getCharPref("style.lineheight");
	if (style.lineheight) {
	    //style.lineheight += "%";
	    style.lineheight = (parseInt(style.lineheight) * 0.01).toString();
	}
	style.fontfamily = prefs.getCharPref("style.fontfamily");
	if (style.fontfamily)
	    style.fontfamily = '"' + style.fontfamily + '"';
    } else if (name == "invert") {
	style.background = "#000000";
	style.color = "#ececec";
    } else if (name == "large") {
	style.fontsize = "x-large";
    }

    // Check values
    let re = /^#[0-9a-f]{6}$/i;
    if (style.background && !re.exec(style.background)) {
	style.background = '';
    }
    if (style.color && !re.exec(style.color)) {
	style.color = '';
    }
    //re = /^[0-9]+%$/;
    re = /^[0-9]+\.?[0-9]*$/;
    if (style.lineheight && !(re.exec(style.lineheight))) {
	style.lineheight = '';
    }
    re = /^[0-9]+pt$/;
    if (style.fontsize &&
	!(style.fontsize == "x-large" || re.exec(style.fontsize))) {
	style.fontsize = '';
    }
    //re = /^[^;,(){}\[\]"]+$/;
    re = /^"[^;{}"]+"$/;
    if (style.fontfamily && !re.exec(style.fontfamily)) {
	style.fontfamily = '';
    }

    let css = "";
    if (style.background || style.color || style.fontsize ||
	style.fontfamily || style.lineheight) {
	css = "@namespace url(http://www.w3.org/1999/xhtml);";
	css += " x-ordbrand-root, x-ordbrand-glue > html {"
	if (style.background)
	    css += " background-color: " + style.background + " !important;"
	css += " }";
	css += " x-ordbrand-glue > html > body {";
	if (style.background)
	    css += " background-color: " + style.background + " !important;";
	if (style.color)
	    css += " color: " + style.color + " !important;";
	if (style.fontfamily)
	    css += " font-family: " + style.fontfamily + " !important;";
	if (style.fontsize)
	    css += " font-size: " + style.fontsize + " !important;";
	if (style.lineheight)
	    css += " line-height: " + style.lineheight + " !important;";
	css += " }";
    }

    return css;
}

function get_all_notes() {
    let nm = Components.classes["@lucidor.org/notemanagerService;1"].getService().wrappedJSObject;
    if (!nm.open())
	return "";

    let all = [];
    for (let pid in loop(nm.find_all_publications())) {
	let notes = get_book_notes(nm, pid);
	if (notes)
	    all.push(notes);
    }

    return wrap_notes(all);
}

function get_notes(book) {
    let nm = Components.classes["@lucidor.org/notemanagerService;1"].getService().wrappedJSObject;
    if (!nm.open())
	return "";

    let pid = nm.find_publication(book.sha1, book.rootfile);
    if (pid < 0)
	return "";

    let notes = get_book_notes(nm, pid);
    if (notes)
	return wrap_notes([notes]);
    return wrap_notes([]);
}

function wrap_notes(notes) {
    let producer = "Lucifox";
    let appinfo = Components.classes["@mozilla.org/xre/app-info;1"].
	getService(Components.interfaces.nsIXULAppInfo);
    if (appinfo.ID == "Lucidor@lucidor.org")
	producer = "Lucidor";

    return JSON.stringify({ producer: producer,
			    created: new Date(),
			    type: "http://lucidor.org/-/annotations/",
			    version: 1,
			    annotations: notes }, null, 2);
}

function get_book_notes(nm, pid) {
    debug("get_book_notes: " + pid);
    let pub = nm.get_publication(pid);
    if (!pub)
	return null;

    delete pub.id;
    if (pub.info) {
	if (pub.info.title)
	    pub.title = pub.info.title;
	if (pub.info.creator)
	    pub.creator = pub.info.creator;
	if (pub.info.publisher)
	    pub.publisher = pub.info.publisher;
    }
    delete pub.info;

    let notes = nm.find_notes_id(pid);
    let notelist = [];
    for (let nid in loop(notes)) {
	let note = nm.get_note(nid);
	delete note.id;
	for (let target in loop(note.targets)) {
	    if (!target.endpart)
		delete target.endpart;
	    if (!target.endtext)
		delete target.endtext;
	    delete target.id;
	}
	notelist.push(note);
    }

    return { publication: pub,
	     annotations: notelist };
}

function print_notes(book, bundle) {
    let nm = Components.classes["@lucidor.org/notemanagerService;1"].getService().wrappedJSObject;
    if (!nm.open())
	return "";

    let pid = nm.find_publication(book.sha1, book.rootfile);
    if (pid < 0)
	return "";

    let notes = get_book_notes(nm, pid);
    if (!notes)
	return "";

    let title = "?";
    let pub = nm.get_publication(pid);
    if (pub && pub.info && pub.info.title)
	title = pub.info.title;
    let text = "Annotations for " + title;
    if (bundle)
	text = bundle.getFormattedString("annotations_for", [ title ]);
    let out = "<html xmlns='http://www.w3.org/1999/xhtml'>\n<head>\n" +
	"<title>" + escapeHtml(text) + "</title>\n";
    out += "<style type='text/css'>\n" +
	".item { margin-top: 7px; margin-bottom: 7px; }\n" +
	".line { margin-top: 2px; margin-bottom: 2px; }\n" +
	".text { font-style: italic; }\n" +
	".label { font-style: italic; color: gray; }\n" +
	".context { color: gray; }\n" +
	"</style>\n";

    out += "</head>\n<body>\n";
    out += "<h2>" + escapeHtml(text) + "</h2>\n";
    let first = true;
    for (let note in loop(notes.annotations)) {
	if (first) {
	    first = false;
	} else {
	    out += "<hr/>\n";
	}

	out += "<div class='item'>\n";
	text = "Created";
	if (bundle)
	    text = bundle.getString("date_creation");
	out += "<p class='line'><span class='label'>" + escapeHtml(text) +
	    ":</span> " + escapeHtml(note.created.toLocaleString());
	if (note.modified.toString() != note.created.toString()) {
	    text = "Modified";
	    if (bundle)
		text = bundle.getString("date_modification");
	    out += "<br/>\n<span class='label'>" + escapeHtml(text) +
		":</span> " + escapeHtml(note.modified.toLocaleString());
	}
	out += "</p>\n";

	if (note.type == "text") {
	    text = escapeHtml(note.content);
	    text = text.replace("\n", "<br/>\n", "g");
	    out += "<p class='line'>" + escapeHtml(text) + "</p>\n";
	}

	let tt = "";
	for (let target in loop(note.targets)) {
	    if (target.text || target.endtext ||
		target.pretext || target.posttext) {
		tt += "<li>";
		if (target.pretext) {
		    tt += "<span class='context'>" +
			escapeHtml(target.pretext) + "</span>";
		}
		if (target.text) {
		    tt += "<span class='text'>" +
			escapeHtml(target.text) + "</span>";
		}
		if (target.endtext) {
		    tt += "<span class='context'>[\u2026]</span>";	// ...
		    tt += "<span class='text'>" +
			escapeHtml(target.endtext) + "</span>";
		}
		if (target.posttext) {
		    tt += "<span class='context'>" +
			escapeHtml(target.posttext) + "</span>";
		}
		tt += "</li>\n";
	    }
	}
	if (tt)
	    out += "<ul class='line'>\n" + tt + "</ul>\n";
	out += "</div>\n";
    }
    out += "</body>\n</html>";
    return out;
}

function escapeHtml(text) {
    text = text.replace(/&/g, '&amp;');
    text = text.replace(/</g, '&lt;');
    text = text.replace(/>/g, '&gt;');
    return text;
}

function indirectionResolver(book, active) {
    this.book = book;
    this.active = active;
}

indirectionResolver.prototype = {
    active: false,

    resolve: function(node) {
	debug("indirectionResolver.resolve: " + node);

	let res = { href: null,
		    node: null };
	if (node == null) {
	    res.node = this.book.opf.domroot;
	} else if (node.nodeType != node.ELEMENT_NODE) {
	    // error
	} else {
	    if (node.localName == "itemref") {
		let idref = node.getAttribute("idref");
		debug("resolve_indirection spine: " + idref);
		res.href = this.book.idrefToHref(idref);
		if (this.active)
		    if (res.href == href_to_local_href(browser.currentURI.spec))
			res.node = browser.contentDocument.documentElement.firstChild.firstChild;
	    } else if (node.localName == "item") {	// Non-standard
		// get manifest item
		let href = node.getAttribute("href");
		debug("resolve_indirection manifest: " + href);
		res.href = href;
		if (this.active)
		    if (res.href == href_to_local_href(browser.currentURI.spec))
			res.node = browser.contentDocument.documentElement.firstChild.firstChild;
	    } else if (node.localName == "iframe" ||
		       node.localName == "embed") {
		// Ignore for now
		// src attribute
	    } else if (node.localName == "object") {
		// Ignore for now
		// data attribute
	    } else if (node.localName == "image" ||
		       node.localName == "use") {
		// Ignore for now
		// xlink:href attribute
	    } else {
		// error
	    }
	}
	return res;
    }
}
