/*
 *  Lucidor - e-book reader
 *  Copyright © 2009-2015  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("chrome://lucidor/content/utils.jsm");
Components.utils.import("resource://gre/modules/Downloads.jsm");
Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

const SCHEME = "x-lucidor-convert";
const XHTML_NS = "http://www.w3.org/1999/xhtml";

function convertProtocol() {
}

convertProtocol.prototype = {
    classDescription: "convertProtocol",
    classID: Components.ID("{e381f457-f813-4c1a-98d4-b0ade627983b}"),
    contractID: "@mozilla.org/network/protocol;1?name=" + SCHEME,
    QueryInterface: XPCOMUtils.generateQI([Components.interfaces.nsISupports,
					   Components.interfaces.nsIProtocolHandler]),

    defaultPort: -1,
    protocolFlags: Components.interfaces.nsIProtocolHandler.URI_NOAUTH |
	Components.interfaces.nsIProtocolHandler.URI_IS_LOCAL_FILE,
    scheme: SCHEME,

    allowPort: function(port, scheme) {
        return false;
    },

    newURI: function(spec, originCharset, baseURI) {
        let uri = Components.classes["@mozilla.org/network/standard-url;1"].
		createInstance(Components.interfaces.nsIStandardURL);
	uri.init(uri.URLTYPE_STANDARD, this.defaultPort, spec,
		 originCharset, baseURI);

        return uri.QueryInterface(Components.interfaces.nsIURI);
    },

    newChannel: function(URI) {
	debug("My newChannel: " + URI.spec);

	URI.QueryInterface(Components.interfaces.nsIURL);
	debug("PATH: " + URI.path);
	debug("HOST: " + URI.host);
	debug("QUERY: " + URI.query);

	let params = URI.query.split("&");
	//x-lucidor-convert://feed/?uri=...
	//x-lucidor-convert://feed/?uri=...&uri=...&title=...
	//x-lucidor-convert://text/?uri=...&charset=iso-8859-1
	let uris = [];
	let vals = {};
	for (let par in loop(params)) {
	    let pair = par.split("=");
	    if (pair.length != 2)
		continue;
	    pair[1] = decodeURIComponent(pair[1]);
	    debug("PARAM: " + pair[0] + " = " + pair[1]);
	    if (pair[0] == "uri")
		uris.push(pair[1]);
	    else	// author, title, charset
		vals[pair[0]] = pair[1];
	}

	let pipe = Components.classes["@mozilla.org/pipe;1"].
            createInstance(Components.interfaces.nsIPipe);
	pipe.init(true, false, 0, 0, null);

	if (URI.host == "text") {
	    let ioService = Components.classes["@mozilla.org/network/io-service;1"].
		getService(Components.interfaces.nsIIOService);
	    try {
		let uri = ioService.newURI(uris[0], null, null);
		this.download(URI.host, uri, pipe, vals);
	    } catch (e) {
		debug("URI ERR: " + e);
		return null;
	    }
	} else if (URI.host == "feed") {
	    this.do_feed(uris, vals, pipe);
	}

	let channel = Components.classes["@mozilla.org/network/input-stream-channel;1"].createInstance(Components.interfaces.nsIInputStreamChannel);
	channel.contentStream = pipe.inputStream;
	channel.setURI(URI);
	channel.QueryInterface(Components.interfaces.nsIChannel);
	channel.contentType = "application/epub+zip";
	channel.contentLength = -1;

	debug("RETURN");
	return channel;
    },

    do_feed: function(uris, vals, pipe) {
	function notify(out) {
	    if (out) {
		debug("h 4: " + out.fileSize);
		let nis = Components.classes["@mozilla.org/network/file-input-stream;1"].
		    createInstance(Components.interfaces.nsIFileInputStream);
		nis.init(out, -1, -1, 0);
		pipe.outputStream.writeFrom(nis, out.fileSize);
		pipe.outputStream.close();
		out.remove(false);
		debug("h 6");
	    } else {
		debug("h 4 no output");
		pipe.outputStream.close();
	    }
	}

	let fc = Components.classes["@lucidor.org/feedConverter;1"].
		createInstance().wrappedJSObject;
	let title = null;
	if (vals["title"])
	    title = vals["title"];
	fc.convert(title, uris, notify);
    },

    download: function(host, uri, pipe, vals) {
	let bookman = Components.classes["@lucidor.org/bookmanagerService;1"].
	    getService().wrappedJSObject;
	let fil = bookman.getTempdir();
	fil.append("lucidor.txt");
	fil.createUnique(Components.interfaces.nsIFile.NORMAL_FILE_TYPE,
			 parseInt("0600", 8));

	let downloadp = Downloads.createDownload({ source: uri, target: fil });
	downloadp.then(function (download) {
	    let prom = download.start();
	    prom.then(function (value) {
		debug("Persister has finished saving data.");

		let out;
		if (host == "text")
		    out = gutenberg_text_to_epub(fil, uri, vals);

		debug("H 4: " + out.fileSize);
		let nis = Components.classes["@mozilla.org/network/file-input-stream;1"].
		    createInstance(Components.interfaces.nsIFileInputStream);
		nis.init(out, -1, -1, 0);
		pipe.outputStream.writeFrom(nis, out.fileSize);
		pipe.outputStream.close();
		debug("H 6");
	    },
	    function (reason) {
		download.finalize(true);
		Components.utils.reportError(reason);
	    });
	},
	Components.utils.reportError);
    }
};

var components = [convertProtocol];

if (XPCOMUtils.generateNSGetFactory)
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
else
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);

function gutenberg_text_to_epub(fil, uri, vals) {
    debug("gutenberg_text_to_epub");
    let spine = [];
    let nonspine = [];
    let nav = [];

    let epub = Components.classes["@lucidor.org/epubCreator;1"].
	createInstance().wrappedJSObject;

    if (vals.title)
	epub.set_title(vals.title);

    if (vals.language)
	epub.add_language(vals.language);

    epub.set_uid(null);
    epub.set_modified(null);

    if (vals.author)
	epub.add_creator(vals.author, 'aut');

    if (vals.publisher)
	epub.set_publisher(vals.publisher);

    epub.set_source(uri);

    epub.add_autopub('text/gutenberg', uri);

    /*
    nonspine.push({name: "pg.css", mediatype: "text/css",
		href: "chrome://lucidor/content/lucidor-pg.css"});
    */

    let tp = epub.generate_titlepage();
    spine.push({name: "titlepage.xhtml", mediatype: "application/xhtml+xml",
		nonlinear: false, string: tp[0]});
    if (tp[1])
	nonspine.push(tp[1]);

    let document = Components.classes["@mozilla.org/xml/xml-document;1"].
	createInstance(Components.interfaces.nsIDOMDocument);
    let implementation = document.implementation;
    let dt = implementation.createDocumentType("html", "", "");
    let xdoc = implementation.createDocument("", "", dt);

    let el = xdoc.createElementNS(XHTML_NS, "html");
    xdoc.appendChild(el);
    debug("G 2");
    let head = xdoc.createElementNS(XHTML_NS, "head");
    el.appendChild(head);
    let body = xdoc.createElementNS(XHTML_NS, "body");
    el.appendChild(body);

    // title
    el = xdoc.createElementNS(XHTML_NS, "title");
    let text = xdoc.createTextNode(vals.title);
    el.appendChild(text);
    head.appendChild(el);

    /*
    el = xdoc.createElementNS(XHTML_NS, "link");
    el.setAttribute("rel", "stylesheet");
    el.setAttribute("type", "text/css");
    el.setAttribute("href", "pg.css");
    head.appendChild(el);
    */

    let div = xdoc.createElementNS(XHTML_NS, "div");
    //div.setAttribute("id", eid);
    div.setAttribute("class", "entry");
    body.appendChild(div);

    debug("G 3");

    fil.QueryInterface(Components.interfaces.nsIFile);
    let filis = Components.classes["@mozilla.org/network/file-input-stream;1"].
	createInstance(Components.interfaces.nsIFileInputStream);
    filis.init(fil, -1, -1, 0);

    debug("G 4: " + fil.fileSize);

    if (!vals.charset)
	vals.charset = "ISO-8859-1";
    let is = Components.classes["@mozilla.org/intl/converter-input-stream;1"].
	createInstance(Components.interfaces.nsIConverterInputStream);
    is.init(filis, vals.charset, 1024, is.DEFAULT_REPLACEMENT_CHARACTER);

    debug("G 5");

    let txt = "";
    let str = {};
    while (is.readString(512, str) != 0)
	txt += str.value;

    // FIXME
    txt = "hej";

    div.innerHTML = txt;

    debug("G 6");

    let xs = Components.classes["@mozilla.org/xmlextras/xmlserializer;1"].
	createInstance(Components.interfaces.nsIDOMSerializer);
    let ostring = xs.serializeToString(xdoc);

    spine.push({name: "text.xhtml", mediatype: "application/xhtml+xml",
		nonlinear: false, string: ostring});
    nav.push({title: "Text", href: "text.xhtml"});

    epub.add_items(spine, nonspine, nav);
    let ofil = epub.generate();
    debug("G 7");
    return ofil;
}
