/*
 *  Lucidor - e-book reader
 *  Copyright © 2008-2013  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("chrome://lucidor/content/utils.jsm");
Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");
Components.utils.import("resource://gre/modules/NetUtil.jsm");

const CONVERT = "?from=application/x-dtbook+xml&to=*/*";
const PIPE_SIZE = 9999999;

function dtbookStreamConverter() {
}

dtbookStreamConverter.prototype = {
    classDescription: "dtbookStreamConverter",
    classID: Components.ID("{75e6266a-37cb-4e52-8681-39fb8a71b58a}"),
    contractID: "@mozilla.org/streamconv;1" + CONVERT,
    _xpcom_categories: [{ category: "@mozilla.org/streamconv;1",
			  entry: CONVERT,
			  value: dtbookStreamConverter.prototype.classDescription }],
    QueryInterface: XPCOMUtils.generateQI([Components.interfaces.nsISupports,
					   Components.interfaces.nsIStreamConverter,
					   Components.interfaces.nsIStreamListener,
					   Components.interfaces.nsIRequestObserver]),

    asyncConvertData: function(fromType, toType, listener, ctxt) {
	debug("dtbookStreamConverter.asyncConvertData");
	this.listener = listener;
    },

    convert: function(fromStream, fromType, toType, ctxt) {
	debug("dtbookStreamConverter.convert");
	throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
    },

    onStartRequest: function(request, context) {
	debug("dtbookStreamConverter.onStartRequest");
	this.input = ""
	request.QueryInterface(Components.interfaces.nsIChannel).contentType =
	    "application/xhtml+xml";
	//request.contentCharset = "UTF-8";

	this.listener.QueryInterface(Components.interfaces.nsIRequestObserver).
	    onStartRequest(request, context);
    },

    onDataAvailable: function(request, context, instream, offset, count) {
	debug("dtbookStreamConverter.onDataAvailable");
	try {
	    let ops = { charset: "UTF-8", replacement: "¿" };
	    if (request.contentCharset)
		ops.charset = request.contentCharset;
	    this.input += NetUtil.readInputStreamToString(instream, count, ops);
	} catch (e) {
	    this.input += NetUtil.readInputStreamToString(instream, count);
	}
    },

    onStopRequest: function(request, context, statusCode) {
	debug("dtbookStreamConverter.onStopRequest: " + statusCode);

	if (statusCode) {
	    // FIXME
	    return;
	}

	this.transform(request, context, statusCode);
    },

    transform: function(request, context, statusCode) {
	debug("dtbookStreamConverter.transform");

	let this_proxy = this;
	function my_transform2(event) {
	    this_proxy.transform2(event, request, context, statusCode);
	}
	let oXHR = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Components.interfaces.nsIXMLHttpRequest);
	oXHR.open("GET", "chrome://lucidor/content/dtbook2dtbook+xhtml.xsl");
	oXHR.overrideMimeType("text/xml");
	oXHR.onload = my_transform2;
	oXHR.send(null);
    },

    transform2: function(event, request, context, statusCode) {
	let oXHR = event.target;
	let xslt_doc = oXHR.responseXML;

	let xslt = Components.classes["@mozilla.org/document-transformer;1?type=xslt"].createInstance(Components.interfaces.nsIXSLTProcessor);
	xslt.importStylesheet(xslt_doc.documentElement);

	let parser = Components.classes["@mozilla.org/xmlextras/domparser;1"].
		getService(Components.interfaces.nsIDOMParser);
	let doc = parser.parseFromString(this.input, "text/xml");

	let new_doc = xslt.transformToDocument(doc);

	let pipe = Components.classes["@mozilla.org/pipe;1"].
		createInstance(Components.interfaces.nsIPipe);
	pipe.init(false, true, 0, PIPE_SIZE, null);

	let dom_ser = Components.classes["@mozilla.org/xmlextras/xmlserializer;1"].createInstance(Components.interfaces.nsIDOMSerializer);
	dom_ser.serializeToStream(new_doc, pipe.outputStream, "");
	pipe.outputStream.close();

	let newstream = pipe.inputStream;
	this.listener.QueryInterface(Components.interfaces.nsIStreamListener).
	    onDataAvailable(request, context, newstream, 0,
			    newstream.available());
	this.listener.QueryInterface(Components.interfaces.nsIRequestObserver).
	    onStopRequest(request, context, statusCode);
    }
};

var components = [dtbookStreamConverter];

if (XPCOMUtils.generateNSGetFactory)
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
else
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);
