/*
 *  Lucidor - e-book reader
 *  Copyright © 2009-2013  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("chrome://lucidor/content/utils.jsm");
Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

const XHTML_NS = "http://www.w3.org/1999/xhtml";

function feedConverter() {
    this.wrappedJSObject = this;
}

feedConverter.prototype = {
    classDescription: "feedConverter",
    classID: Components.ID("{14d32bec-c8b9-4cf8-adc8-73b2b1d97bba}"),
    contractID: "@lucidor.org/feedConverter;1",
    QueryInterface: XPCOMUtils.generateQI([Components.interfaces.nsISupports]),

    notify_function: null,
    list: null,
    title: null,

    convert: function(title, uris, func) {
	debug("convert");
	this.title = title;
	this.notify_function = func;

	this.list = new Array();
	for (let uri in loop(uris)) {
	    let j = uri.indexOf(':');
	    if (j >= 0) {
		debug('URI: ' + uri);
		this.list.push({ url: uri.substring(j + 1),
				 type: uri.substring(0, j),
				 done: false,
				 meta: null,
				 error: null,
				 fail: false });
	    }
	}

	for (let num = 0; num < this.list.length; num++)
	    this.get_doc(num);
    },

    get_doc: function(num) {
	debug("get_doc: " + this.list[num].url);

	let xhr = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"].
	    createInstance(Components.interfaces.nsIXMLHttpRequest);
	this.list[num].httpRequest = xhr;
	let this_proxy = this;
	xhr.onload = function() {
	    this_proxy.handle_doc(num);
	}

	try {
	    xhr.open("GET", this.list[num].url, true);
	    xhr.responseType = "text";
	    xhr.send(null);
	    debug("SENT");
	} catch (e) {
	    debug("ERROR " + e);
	    this.list[num].error = e.message;
	    this.fail(num);
	}
    },

    fail: function(num) {
	debug("fail: " + num);
	this.list[num].fail = true;
	this.list[num].done = true;
	this.check();
    },

    check: function() {
	for (let i = 0; i < this.list.length; i++)
	    if (!this.list[i].done)
		return;
	try {
	    this.compose();
	} catch (e) {
	    debug("E: " + e);
	}
    },

    handle_doc: function(num) {
	debug("handle_doc: " + num);

	let xhr = this.list[num].httpRequest;
	if (xhr.responseText.length < 1) {
	    debug("No doc?");
	    return this.fail(num);
	}

	let parser = Components.classes["@mozilla.org/xmlextras/domparser;1"].
	    createInstance(Components.interfaces.nsIDOMParser);

	if (this.list[num].type == 'svg') {
	    let doc = parser.parseFromString(xhr.responseText, "image/svg+xml");
	    if (doc.documentElement.localName == "svg")
		return this.handleSVG(doc, xhr.responseText, num);
	} else if (this.list[num].type == 'html') {
	    doc = parser.parseFromString(xhr.responseText, "text/html");
	    if (doc.documentElement.localName.toLowerCase() == "html")
		return this.handleHTML(doc, xhr.responseText, num);
	} else if (this.list[num].type == 'feed') {
	    //debug("---\n" + xhr.responseText + "---");
	    parser = Components.classes["@mozilla.org/feed-processor;1"].
		createInstance(Components.interfaces.nsIFeedProcessor);
	    //parser.listener = this;
	    let this_proxy = this;
	    parser.listener = {
		handleResult: function(result) {
		    if (!result.doc)
			return this_proxy.fail(num);
		    return this_proxy.handleAtom(result, num);
		}
	    };
	    let ioService = Components.classes["@mozilla.org/network/io-service;1"].
		getService(Components.interfaces.nsIIOService);
	    let uri = ioService.newURI(this.list[num].url, null, null);
	    try {
		return parser.parseFromString(xhr.responseText, uri);
	    } catch (e) {
		debug("Error parsing doc.");
		return this.fail(num);
	    }
	} else {
	    debug("Unknown type: " + this.list[num].type);
	    return this.fail(num);
	}

	this.fail(num);
    },

    handleAtom: function(result, num) {
	debug("handleAtom");

	let meta = [];
	let spine = [];
	let nav = [];
	let filename = "luci-" + num + ".html";

	let parserUtils = Components.classes["@mozilla.org/parserutils;1"].
            getService(Components.interfaces.nsIParserUtils);

        let feed = result.doc.QueryInterface(Components.interfaces.nsIFeed);

	// Metadata

	let title = this.list[num].url;
	if (feed.title)
	    title = feed.title.plainText();
	meta.push({name: "title", value: title});

	let lang;
	if (feed.fields.hasKey("language"))
	    lang = feed.fields.getPropertyAsAUTF8String("language");
	if (!lang)
	    lang = "en";
	meta.push({name: "language", value: lang});

	for (let i = 0; i < feed.authors.length; i++) {
	    let creator = feed.authors.queryElementAt(i, Components.interfaces.nsIFeedPerson);
	    meta.push({name: "creator", value: creator.name,
			attrs: [{name: "role", value: "aut"}]});
	}
	for (let i = 0; i < feed.contributors.length; i++) {
	    let creator = feed.contributors.queryElementAt(i, Components.interfaces.nsIFeedPerson);
	    /*
	    meta.push({name: "contributor", value: creator.name,
			attrs: [{name: "role", value: "clb"}]});
	    */
	    meta.push({name: "creator", value: creator.name,
			attrs: [{name: "role", value: "clb"}]});
	}

	if (feed.id)
	    meta.push({name: "identifier", value: feed.id});

	if (feed.updated)
	    meta.push({name: "date", value: dateToValue(feed.updated),
			attrs: [{name: "event", value: "update"}]});	// FIXME syndication?

	// FIXME feed.rights

	if (feed.categories)
	    for (let i = 0; i < feed.categories.length; i++) {
		let cat = feed.categories.queryElementAt(i, Components.interfaces.nsISupports);
		debug("FEED CAT: " + cat);
	    }
	//let subj = feed.fields.getPropertyAsAUTF8String("dc:subject");

	let en = feed.fields.enumerator;
	while (en.hasMoreElements()) {
	    let field = en.getNext().QueryInterface(Components.interfaces.nsIProperty);
	    debug("FIELD: " + field.name + " " + field.value);
	}

	nav.push({title: title, href: filename, level: 0});

	// feed.image

	debug("W 0");

	// Feed Text

	let document = Components.classes["@mozilla.org/xml/xml-document;1"].
		createInstance(Components.interfaces.nsIDOMDocument);
	let implementation = document.implementation;
	let dt = implementation.createDocumentType("html", "", "");
	let xdoc = implementation.createDocument("", "", dt);

	let el = xdoc.createElementNS(XHTML_NS, "html");
	xdoc.appendChild(el);
	debug("2");
	let head = xdoc.createElementNS(XHTML_NS, "head");
	el.appendChild(head);
	let body = xdoc.createElementNS(XHTML_NS, "body");
	el.appendChild(body);

	// title

	el = xdoc.createElementNS(XHTML_NS, "title");
	let text = xdoc.createTextNode(title);
	el.appendChild(text);
	head.appendChild(el);

	el = xdoc.createElementNS(XHTML_NS, "link");
	el.setAttribute("rel", "stylesheet");
	el.setAttribute("type", "text/css");
	el.setAttribute("href", "feed.css");
	head.appendChild(el);

	el = xdoc.createElementNS(XHTML_NS, "h1");
	body.appendChild(el);
	debug("TYPE: " + feed.title.type);
	text = null;
	if (feed.title && (feed.title.type == "html" ||
			   feed.title.type == "xhtml")) {
	    debug("T0");
	    try {
		let frag = parserUtils.parseFragment(feed.title, 0, false,
						     xdoc.baseURIObject, el);
		el.appendChild(frag);
		text = "OK";
	    } catch (e) {
	    }
	}
	if (!text) {	// feed.title.type == "text"
	    debug("T1");
	    text = xdoc.createTextNode(title);
	    if (feed.link) {
		debug("T2");
		let el2 = xdoc.createElementNS(XHTML_NS, "a");
		el2.setAttribute("href", feed.link.spec);
		el2.appendChild(text);
		el.appendChild(el2);
	    } else
		el.appendChild(text);
	}

	let div = xdoc.createElementNS(XHTML_NS, "div");
	div.setAttribute("class", "info");
	text = xdoc.createTextNode(feed.subtitle.plainText());
	div.appendChild(text);
	body.appendChild(div);

	// covering [startdate, time] to [enddate, time].
	//div.innerHTML = '<p>Generated with <a href="http://lucidor.org/"><span class="sc">Lucidor</span></a>.</p>';	// FIXME date
	// date

	debug("W 2");

	let counter = 0;
	function get_id() {
	    // FIXME make sure the id does not exist in the content
	    return "lucidor-toc-" + counter++;
	}

	for (let i = 0; i < feed.items.length; i++) {
	    debug("W 2a");
	    let entry = feed.items.queryElementAt(i, Components.interfaces.nsIFeedEntry);
	    let eid = get_id();
	    div = xdoc.createElementNS(XHTML_NS, "div");
	    div.setAttribute("id", eid);
	    div.setAttribute("class", "entry");
	    body.appendChild(div);

	    if (entry.categories)
		for (let i = 0; i < entry.categories.length; i++) {
		    let cat = entry.categories.queryElementAt(i, Components.interfaces.nsISupports);
		    debug("CAT: " + cat);
		}

	    el = xdoc.createElementNS(XHTML_NS, "h2");
	    text = null;
	    if (entry.title && (entry.title.type == "html" ||
				entry.title.type == "xhtml")) {
		debug("T0");
		try {
		    let frag = parserUtils.parseFragment(entry.title, 0, false,
							 xdoc.baseURIObject, el);
		    el.appendChild(frag);
		    text = "OK";
		} catch (e) {
		}
	    }
	    if (!text) {	// entry.title.type == "text"
		debug("T1");
		let txt = "***";
		if (entry.title) {
		    txt = entry.title.plainText();
		    //txt = txt.replace("&", "&amp;", "g");
		}
		text = xdoc.createTextNode(txt);
		if (entry.link) {
		    debug("T2");
		    // FIXME entry.link.resolve("");

		    let el2 = xdoc.createElementNS(XHTML_NS, "a");
		    el2.setAttribute("href", entry.link.spec);
		    el2.appendChild(text);
		    el.appendChild(el2);
		} else
		    el.appendChild(text);
	    }
	    div.appendChild(el);

	    if (entry.published) {
		el = xdoc.createElementNS(XHTML_NS, "h4");
		text = xdoc.createTextNode(entry.published);
		div.appendChild(el);
	    }

	    if (entry.published) {
		try {
		    debug("DDD1: " + entry.published);
		    debug("DDD1: " + (new Date(entry.published)).toString());
		    debug("DDD2: " + dateToValue(entry.published));
		    debug("DDD3: " + dateToLocalValue(entry.published));
		} catch (e) {
		    debug("ERROR2: " + e);
		    debug("TEXT: " + entry.title.plainText());
		}
	    }

	    debug("W 1");
	    let txt;
	    if (entry.content)
		txt = entry.content.text;
	    else
		txt = entry.summary.text;

	    txt = unescapeEntities(txt);

	    while (txt.indexOf("&quot;") >= 0)
		txt = txt.replace("&quot;", '"', "g");	// FIXME
	    while (txt.indexOf("&nbsp;") >= 0)
		txt = txt.replace("&nbsp;", " ", "g");	// FIXME

	    debug("W 1 3");

	    //debug("TXT:" + txt);
	    el = xdoc.createElementNS(XHTML_NS, "div");
	    el.setAttribute("class", "my");
	    div.appendChild(el);

	    try {
		let frag = parserUtils.parseFragment(txt, 0, false,
						     xdoc.baseURIObject, el);
		el.appendChild(frag);
	    } catch (e) {
		debug("ERROR: " + e);
		debug("TEXT: " + txt);
	    }

	    debug("W 1 b");

	    nav.push({title: entry.title.plainText(),
		      href: filename + "#" + eid,
		      level: 1});
	}

	debug("W 3");

	for (let img in loop(xdoc.getElementsByTagName("img"))) {
	    debug("AAA");
	    img.parentNode.removeChild(img);
	}

	let xs = Components.classes["@mozilla.org/xmlextras/xmlserializer;1"].
		createInstance(Components.interfaces.nsIDOMSerializer);
	let out = xs.serializeToString(xdoc);

	//debug("---\n" + out + "\n---");

	spine.push({name: filename, mediatype: "application/xhtml+xml",
		    nonlinear: false, string: out});

	this.list[num].meta = meta;
	this.list[num].spine = spine;
	this.list[num].nav = nav;
	this.list[num].done = true;
	this.list[num].type = 'feed';

	this.check();
    },

    handleHTML: function(resultDoc, resultText, num) {
	let meta = [];
	let spine = [];
	let nav = [];
	let filename = "luci-" + num + ".html";
	let title = this.list[num].url;

	if (resultDoc.title)
	    title = resultDoc.title;
	meta.push({name: "title", value: title});
	meta.push({name: "source", value: this.list[num].url});

	let xs = Components.classes["@mozilla.org/xmlextras/xmlserializer;1"].
		createInstance(Components.interfaces.nsIDOMSerializer);
	let out = xs.serializeToString(resultDoc);

	spine.push({name: filename, mediatype: "application/xhtml+xml",
		    nonlinear: false, string: out});
	nav.push({title: title, href: filename, level: 0});

	this.list[num].meta = meta;
	this.list[num].spine = spine;
	this.list[num].nav = nav;
	this.list[num].done = true;
	this.list[num].type = 'html';

	this.check();
    },

    handleSVG: function(resultDoc, resultText, num) {
	let meta = [];
	let spine = [];
	let nav = [];
	let filename = "luci-" + num + ".html";
	let title = this.list[num].url;

	if (resultDoc.title)
	    title = resultDoc.title;
	meta.push({name: "title", value: title});
	meta.push({name: "source", value: this.list[num].url});

	let xs = Components.classes["@mozilla.org/xmlextras/xmlserializer;1"].
		createInstance(Components.interfaces.nsIDOMSerializer);
	let out = xs.serializeToString(resultDoc);

	spine.push({name: filename, mediatype: "image/svg+xml",
		    nonlinear: false, string: out});
	nav.push({title: title, href: filename, level: 0});

	this.list[num].meta = meta;
	this.list[num].spine = spine;
	this.list[num].nav = nav;
	this.list[num].done = true;
	this.list[num].type = 'svg';

	this.check();
    },

    compose: function() {
	debug("compose");

	let epub = Components.classes["@lucidor.org/epubCreator;1"].
		createInstance().wrappedJSObject;

	let spine = [];
	let nonspine = [];
	let nav = [];
	let title = this.title;

	/*
	let ok = false;
	for (let ls in loop(this.list))
	    if (!ls.fail) {
		ok = true;
		break;
	    }
	if (!ok)
	    return this.notify_function(null);
	*/

	for (let ls in loop(this.list)) {
	    for (let item in loop(ls.meta)) {
		if (item.name == "creator") {
		    epub.add_creator(item.value, null);
		} else if (item.name == "title") {
		    if (!title)
			title = item.value;
		} else if (item.name == "language") {
		    epub.add_language(item.value);
		} else if (item.name == "subject") {
		    epub.add_subject(item.value);
		}
	    }
	    epub.add_autopub(ls.type, ls.url);
	}
	if (title)
	    epub.set_title(title);

	epub.set_uid(null);
	epub.set_modified(null);

	/*
	meta.push({name: "date", value: dateToLocalValue(Date()),
		   attrs: [{name: "event", value: "creation"}]});
	*/

	nonspine.push({name: "feed.css", mediatype: "text/css",
		       href: "chrome://lucidor/content/lucidor-feed.css"});

	// Title Page
	let tp = epub.generate_titlepage();
	spine.push({name: "title.html", mediatype: "application/xhtml+xml",
		    nonlinear: false, string: tp[0]});
	if (tp[1])
	    nonspine.push(tp[1]);
	//nav.push({title: "Titlepage", href: "titlepage.html", level: 0});

	for (let i = 0; i < this.list.length; i++) {
	    if (this.list[i].fail) {
		let tit = this.list[i].url;
		let fn = "luci-" + i + ".html";
		let doc = '<p>';
		doc += 'Failed to generate page from document at ';
		doc += epub.escapeHtml(tit);
		doc += '</p>\n';
		if (this.list[i].error) {
		    doc += '<p>';
		    doc += epub.escapeHtml(this.list[i].error);
		    doc += '</p>\n';
		}

		spine.push({name: fn,
			    mediatype: "application/xhtml+xml",
			    nonlinear: false,
			    string: epub.wrap_content(tit, doc)});
		nav.push({title: tit, href: fn, level: 0});
	    } else {
		spine = spine.concat(this.list[i].spine);
		nav = nav.concat(this.list[i].nav);
	    }
	}

	epub.add_items(spine, nonspine, nav);

	let fil = null;
	try {
	    fil = epub.generate();
	} catch (e) {
	    debug("ERR: " + e);
	}
	return this.notify_function(fil);
    }
};

var components = [feedConverter];

if (XPCOMUtils.generateNSGetFactory)
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
else
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);

function unescapeEntities(inp) {
    let re0 = new RegExp("&#([0-9]+);");
    let re1 = new RegExp("&#[xX]([a-fA-F0-9]+);");
    let st = inp;
    let res;
    while ((res = re0.exec(st)) != null) {
	let chr = String.fromCharCode(parseInt(res[1], 10));
	st = st.substring(0, res.index) + chr +
	    st.substring(res.index + res[0].length);
    }
    while ((res = re1.exec(st)) != null) {
	let chr = String.fromCharCode(parseInt(res[1], 16));
	st = st.substring(0, res.index) + chr +
	    st.substring(res.index + res[0].length);
    }
    return st;
}

function escapeToEntities(inp) {
    let out = "";
    for (let i = 0; i < inp.length; i++) {
	let val = inp.charCodeAt(i);
	if (val > 127)
	    out += "&#" + val + ";";
	else
	    out += inp.charAt(i);
    }
    return out;
}

function zfill(num, n) {
    let r = "";
    let ns = num.toString();
    for (let i = (n - ns.length); i > 0; i--)
	r += "0";
    return r + ns;
}

// FIXME does not take into account Daylight Savings Time switches
// Using getUTCHours etc. instead of local time should fix this(?)
function dateToValue(st) {
    let da = new Date(st);

    let off = da.getTimezoneOffset();
    let tz = "Z";
    let res = (new RegExp("([-+])([0-9]{2})([0-9]{2})")).exec(st);
    if (res) {
	let ooff = parseInt(res[2], 10) * 60 + parseInt(res[3], 10);
	if (res[1] == "-")
	    off -= ooff;
	else
	    off += ooff;
	tz = res[1] + res[2] + ":" + res[3];
    }
    da.setTime(da.getTime() + off * 60 * 1000);

    let ret = "" + da.getFullYear() + "-" + zfill(da.getMonth() + 1, 2) + "-" +
	zfill(da.getDate(), 2) + "T" + zfill(da.getHours(), 2) + ":" +
	zfill(da.getMinutes(), 2) + ":" + zfill(da.getSeconds(), 2);
    let ms = da.getMilliseconds();
    if (ms != 0)
	ret += "." + ms;
    ret += tz;
    return ret;
}

function dateToLocalValue(st) {
    let da = new Date(st);
    let tz = "Z";
    let off = da.getTimezoneOffset();
    if (off != 0) {
	let h = Math.floor(Math.abs(off / 60));
	let m = Math.abs(off % 60);
	if (off > 0)
	    tz = "-";
	else
	    tz = "+";
	tz += zfill(h, 2) + ":" + zfill(m, 2);
    }
    return "" + da.getFullYear() + "-" + zfill(da.getMonth() + 1, 2) + "-" +
	zfill(da.getDate(), 2) + "T" + zfill(da.getHours(), 2) + ":" +
	zfill(da.getMinutes(), 2) + ":" + zfill(da.getSeconds(), 2) +
	tz;
    // getMilliseconds
}
