/*
 *  Lucidor - e-book reader
 *  Copyright © 2008-2013  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("chrome://lucidor/content/utils.jsm");
Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

function localCase() {
    this.wrappedJSObject = this;
    this.init0();
}

localCase.prototype = {
    classDescription: "localCase",
    classID: Components.ID("{8cd48e94-92e7-4776-87d4-0da5949dfc31}"),
    contractID: "@lucidor.org/localCase;1",
    QueryInterface: XPCOMUtils.generateQI([Components.interfaces.nsISupports]),

    local: true,

    info: { title: true,
	    author: true,
	    location: true },

    listers: [ "all" ],
    //listers: [ "all", "feeds" ],

    browsers: [ "tags" ],

    capabilities: { modify: true,
		    edit_tags: true,
		    standard_search: true,
		    search_meta: true,
		    search_title: true,
		    search_author: true,
		    search_tags: false,
		    search_text: false },

    init0: function() {
	this.bookcase = Components.classes["@lucidor.org/bookcaseSqlService;1"].
		getService().wrappedJSObject;
	this.bookcase.open();
    },

    init: function(coll, tree) {
	this.tree = tree;
	this.tree.datasources = this.bookcase.file_uri;
	this.tree.builder.rebuild();
    },

    close: function() {
    },

    add_book: function(file, callback) {
	this.bookcase.add_book(file, callback);
    },

    add_book_data: function(data) {
	return this.bookcase.add_book_data(data);
    },

    remove_book: function(id) {
	return this.bookcase.remove_book(id);
    },

    get_preview: function(id) {
	return this.bookcase.get_preview(id);
    },

    generate_preview: null,

    add_preview: function(id, url) {
	return this.bookcase.set_preview(id, url);
    },

    get_book_info: function(id) {
	return this.bookcase.get_book_info(id);
    },

    add_tag: function(id, label) {
	return this.bookcase.add_tag(id, label);
    },

    remove_tag: function(id, label) {
	return this.bookcase.remove_tag(id, label);
    },

    get_tags: function(id) {
	return this.bookcase.get_tags(id);
    },

    browse_gen: function(nid) {
	debug("NID: " + nid);
	if (nid == "tags") {
	    return this.browse_tags();
	} else if (nid.substring(0, 5) == "tags:") {
	    let tag = nid.substring(5);
	    return this.browse_tag(tag);
	}
	return null;
    },

    /*
      thisnode: true
      links:
      firstpath:
      nextpath:
      pageinfo:

      type: "folder"
      title:
      path:

      type: "epub" || "link"
      id:
      title:
      author:
      publisher:
      location:
     */
    browse_tags: function() {
	debug("browse_tags gen");
	for (let tag in loop(this.bookcase.get_all_tags())) {
	    debug("gen: " + tag);
 	    let link = { type: "folder",
			 path: "tags:" + tag,
			 title: tag };
	    yield link;
	}
    },

    browse_tag: function(label) {
	debug("browse_tag gen");
	for (let book in loop(this.bookcase.get_books_by_tag(label))) {
	    debug("gen: " + book);
	    let link = this.get_book_info(book);
	    yield link;
	}
    }
};

var components = [localCase];

if (XPCOMUtils.generateNSGetFactory)
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
else
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);
