/*
 *  Lucidor - e-book reader
 *  Copyright © 2013-2014  Mikael Ylikoski
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Components.utils.import("chrome://lucidor/content/utils.jsm");
Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

function notemanager() {
    this.wrappedJSObject = this;
}

notemanager.prototype = {
    classDescription: "notemanager",
    classID: Components.ID("{a3304dd0-6d49-11e2-a6fa-001b246c2d0e}"),
    contractID: "@lucidor.org/notemanagerService;1",
    _xpcom_factory: notemanagerFactory,
    QueryInterface: XPCOMUtils.generateQI([Components.interfaces.nsISupports]),

    db: null,
    file_uri: null,

    get_db_file: function() {
	let file = Components.classes["@mozilla.org/file/directory_service;1"].
		getService(Components.interfaces.nsIProperties).
		get("ProfD", Components.interfaces.nsIFile);
	file.append("lucifox");
	if (!file.exists()) {
	    try {
		file.create(file.DIRECTORY_TYPE, parseInt("0700", 8));
	    } catch (e) {
		return null;
	    }
	}
	file.append("annotations.sqlite");

	let fph = Components.classes["@mozilla.org/network/protocol;1?name=file"].getService(Components.interfaces.nsIFileProtocolHandler);
	this.file_uri = fph.getURLSpecFromFile(file);
	debug("DB_FILE: " + this.file_uri);
	return file;
    },

    open: function() {
	if (this.db && this.db.connectionReady)
	    return true;
	try {
	    let file = this.get_db_file();
	    let ss = Components.classes["@mozilla.org/storage/service;1"].
		getService(Components.interfaces.mozIStorageService);
	    this.db = ss.openDatabase(file);
	    if (this.db.schemaVersion < 1)
		this.create();
	    if (this.db.schemaVersion < 3)
		this.update_db();
	} catch (e) {
	    debug("Cannot open annotation database: " + e);
	    return false;
	}
	return true;
    },

    open_new: function() {
	try {
	    let ss = Components.classes["@mozilla.org/storage/service;1"].
		getService(Components.interfaces.mozIStorageService);
	    this.db = ss.openSpecialDatabase("memory");
	    this.create();
	} catch (e) {
	    debug("Cannot create temporary annotation database: " + e);
	    return false;
	}
	return true;
    },

    close: function() {
	if (this.db && this.db.connectionReady)
	    this.db.asyncClose();
    },

    /*
      ***************
      * publications *
      ***************
      id INTEGER PRIMARY KEY,

      sha1 TEXT,
      root INTEGER,
      rootfile TEXT,	// rootfile.substring(0, rootfile.lastIndexOf("/") + 1)
      uid TEXT,
      modified TEXT,	// dcterms:modified
      info TEXT,	// { title, creator, publisher }

      extra TEXT


      ***************
      * annotations *
      ***************
      id INTEGER PRIMARY KEY,
      uid TEXT,		// uuid; used for comments on comments

      publication INTEGER,	// (publications table id)

      annotator TEXT,	// annotator unique id (annotators table uid)
      created TEXT,	// date
      modified TEXT,	// date
      source TEXT,	// user / imported

      content TEXT,	// annotation content/body
      type TEXT,	// content type: text / html / image / uri (/ bookmark)
      color TEXT,
      style TEXT,	// highlight / underline / strikethrough / frame /
			// bubble / mark

      position REAL,
      extra TEXT,	// unused


      *********
      *targets*
      *********
      id INTEGER PRIMARY KEY,

      annotation INTEGER,	// (annotations table id)

      part TEXT,
      endpart TEXT,	// null if equal to part
      fixptr TEXT,
      epubcfi TEXT,
      location TEXT,	// { start: xpath, startoffset, end, endoffset,
			     pstart: adepoint, pend, xpointer }

      text TEXT,
      endtext TEXT,	// null if equal to text
      pretext TEXT,
      posttext TEXT,

      position REAL,
      extra TEXT,	// unused


      ************
      *annotators*
      ************
      id INTEGER PRIMARY KEY,
      uid TEXT,		// uuid

      name TEXT,
      signature TEXT,
      info TEXT,	// { email, description }

      extra TEXT,	// unused


      ************
      *references*
      ************
      id INTEGER PRIMARY KEY,

      annotation INTEGER,	// (annotation table id)

      linkend TEXT,	// referenced annotation uid (annotations table uid)
      created,		// date

      contenthash TEXT,	// sha1 of content of referenced annotation
      targetshash TEXT,	// sha1 of targets of referenced annotation

      extra TEXT,	// unused


      ******
      *tags*
      ******
      id INTEGER PRIMARY KEY,

      annotation INTEGER,	// (annotations table id)

      tagger TEXT,	// annotator unique id (annotators table id)
      created TEXT,	// date
      //comment TEXT,	// comment made by tagger
      //source TEXT,	// user (added) / imported (existed in original annotation)

      label TEXT,	// "music" "music/classical" "dc:.." "bookmark"
      type TEXT,	// category / origin (group) / permissions
			// label: "bookmark" type: "type"
      extra TEXT,	// unused
    */
    create: function() {
	debug("create");
	this.db.createTable("publications", "id INTEGER PRIMARY KEY, sha1 TEXT, root INTEGER, rootfile TEXT, uid TEXT, modified TEXT, info TEXT, extra TEXT");
	this.db.createTable("annotations", "id INTEGER PRIMARY KEY, uid TEXT, publication INTEGER, annotator TEXT, created TEXT, modified TEXT, source TEXT, content TEXT, type TEXT, color TEXT, style TEXT, position REAL, extra TEXT");
	this.db.createTable("targets", "id INTEGER PRIMARY KEY, annotation INTEGER, part TEXT, endpart TEXT, fixptr TEXT, epubcfi TEXT, location TEXT, text TEXT, endtext TEXT, pretext TEXT, posttext TEXT, position REAL, extra TEXT");

	//this.db.createTable("references", "id INTEGER PRIMARY KEY, annotation INTEGER, linkend TEXT, created TEXT, contenthash TEXT, targetshash TEXT, extra TEXT");

	this.db.schemaVersion = 3;

	let statement = this.db.createStatement("CREATE TRIGGER after_annotation_delete AFTER DELETE ON annotations FOR EACH ROW BEGIN DELETE FROM publications WHERE id=OLD.publication AND NOT EXISTS (SELECT id FROM annotations WHERE publication=OLD.publication); DELETE FROM targets WHERE annotation=OLD.id; END");
	// DELETE FROM references WHERE annotation=OLD.id;
	// DELETE FROM tags WHERE annotation=OLD.id;
	statement.execute();

	statement = this.db.createStatement("CREATE INDEX p_sha1 ON publications(sha1)");
	statement.execute();

	statement = this.db.createStatement("CREATE INDEX a_pub ON annotations(publication)");
	statement.execute();

	statement = this.db.createStatement("CREATE INDEX l_part ON targets(part)");
	statement.execute();
    },

    update_db: function() {
	if (this.db.schemaVersion < 2) {
	    let statement = this.db.createStatement("ALTER TABLE annotations ADD COLUMN publication TEXT");
	    statement.execute();

	    this.db.createTable("publications", "id INTEGER PRIMARY KEY, sha1 TEXT, root INTEGER, rootfile TEXT, uid TEXT, info TEXT, extra TEXT");
	    this.db.createTable("pointers", "id INTEGER PRIMARY KEY, annotation INTEGER, part TEXT, fixptr TEXT, location TEXT, text TEXT, position REAL, extra TEXT");

	    statement = this.db.createStatement("CREATE TRIGGER after_annotation_delete AFTER DELETE ON annotations FOR EACH ROW BEGIN DELETE FROM publications WHERE id=OLD.publication AND NOT EXISTS (SELECT id FROM annotations WHERE publication=OLD.publication); DELETE FROM pointers WHERE annotation=OLD.id; END");
	    statement.execute();

	    statement = this.db.createStatement("CREATE INDEX p_sha1 ON publications(sha1)");
	    statement.execute();

	    statement = this.db.createStatement("CREATE INDEX a_pub ON annotations(publication)");
	    statement.execute();

	    statement = this.db.createStatement("CREATE INDEX l_part ON targets(part)");
	    statement.execute();

	    this.db.schemaVersion = 2;
	}

	if (this.db.schemaVersion < 3) {
	    let statement = this.db.createStatement("ALTER TABLE publications ADD COLUMN modified TEXT");
	    statement.execute();

	    statement = this.db.createStatement("ALTER TABLE annotations ADD COLUMN uid TEXT");
	    statement.execute();

	    statement = this.db.createStatement("ALTER TABLE annotations ADD COLUMN source TEXT");
	    statement.execute();

	    statement = this.db.createStatement("ALTER TABLE annotations ADD COLUMN content TEXT");
	    statement.execute();

	    statement = this.db.createStatement("UPDATE annotations SET content=comment");
	    statement.execute();

	    statement = this.db.createStatement("ALTER TABLE pointers RENAME TO targets");
	    statement.execute();

	    statement = this.db.createStatement("ALTER TABLE targets ADD COLUMN endpart TEXT");
	    statement.execute();

	    statement = this.db.createStatement("ALTER TABLE targets ADD COLUMN epubcfi TEXT");
	    statement.execute();

	    statement = this.db.createStatement("ALTER TABLE targets ADD COLUMN endtext TEXT");
	    statement.execute();

	    statement = this.db.createStatement("ALTER TABLE targets ADD COLUMN pretext TEXT");
	    statement.execute();

	    statement = this.db.createStatement("ALTER TABLE targets ADD COLUMN posttext TEXT");
	    statement.execute();

	    statement = this.db.createStatement("DROP TRIGGER IF EXISTS after_annotation_delete");
	    statement.execute();

	    statement = this.db.createStatement("CREATE TRIGGER after_annotation_delete AFTER DELETE ON annotations FOR EACH ROW BEGIN DELETE FROM publications WHERE id=OLD.publication AND NOT EXISTS (SELECT id FROM annotations WHERE publication=OLD.publication); DELETE FROM targets WHERE annotation=OLD.id; END");
	    statement.execute();

	    this.db.schemaVersion = 3;
	}
    },

    clear: function() {
	this.db.schemaVersion = 0;
	let statement = this.db.createStatement("DROP TABLE IF EXISTS publications");
	statement.execute();
	statement = this.db.createStatement("DROP TABLE IF EXISTS annotations");
	statement.execute();
	statement = this.db.createStatement("DROP TABLE IF EXISTS targets");
	statement.execute();
	//statement = this.db.createStatement("DROP TABLE IF EXISTS annotators");
	//statement.execute();
	//statement = this.db.createStatement("DROP TABLE IF EXISTS references");
	//statement.execute();
	//statement = this.db.createStatement("DROP TABLE IF EXISTS tags");
	//statement.execute();
	statement = this.db.createStatement("DROP INDEX IF EXISTS p_sha1");
	statement.execute();
	statement = this.db.createStatement("DROP INDEX IF EXISTS a_pub");
	statement.execute();
	statement = this.db.createStatement("DROP INDEX IF EXISTS l_part");
	statement.execute();
	this.create();
    },

    add: function(note) {
	let id = this.find_publication(note.publication.sha1,
				       note.publication.root);

	if (id < 0) {
	    let statement = this.db.createStatement("INSERT INTO publications (id, sha1, root, rootfile, uid, modified, info) VALUES (NULL, :x_sha1, :x_root, :x_rootfile, :x_uid, :x_modified, :x_info)");
	    statement.params.x_sha1 = note.publication.sha1;
	    statement.params.x_root = note.publication.root;
	    statement.params.x_rootfile = note.publication.rootfile;
	    statement.params.x_uid = note.publication.uid;
	    statement.params.x_modified = note.publication.modified;
	    if (note.publication.info)
		statement.params.x_info = JSON.stringify(note.publication.info);
	    statement.execute();
	    id = this.db.lastInsertRowID;
	}

	if (id >= 0) {
	    let statement = this.db.createStatement("INSERT INTO annotations (id, uid, publication, created, modified, source, content, type, style) VALUES (NULL, :x_uid, :x_publication, datetime(:x_created), datetime('now'), :x_source, :x_content, :x_type, :x_style)");
	    if (note.uid)
		statement.params.x_uid = note.uid;
	    else
		statement.params.x_uid = Components.classes["@mozilla.org/uuid-generator;1"].
		getService(Components.interfaces.nsIUUIDGenerator).
		generateUUID().toString().substring(1, 37);
	    statement.params.x_publication = id;
	    if (note.created)
		statement.params.x_created = note.created;
	    else
		statement.params.x_created = "now";
	    statement.params.x_source = note.source;
	    statement.params.x_content = note.content;
	    statement.params.x_type = note.type;
	    statement.params.x_style = note.style;
	    statement.execute();
	    id = this.db.lastInsertRowID;

	    for (let target in loop(note.targets))
		this.add_target(id, target);

	    return id;
	}
	return -1;
    },

    update: function(id, content, type, style) {
	let statement = this.db.createStatement("UPDATE annotations SET modified=datetime('now'), content=:x_content, type=:x_type, style=:x_style WHERE id=:x_id");
	statement.params.x_content = content;
	statement.params.x_type = type;
	statement.params.x_style = style;
	statement.params.x_id = id;
	statement.execute();
    },

    update_modified: function(id) {
	let statement = this.db.createStatement("UPDATE annotations SET modified=datetime('now') WHERE id=:x_id");
	statement.params.x_id = id;
	statement.execute();
    },

    // Copy location information
    copy_location: function(source, target) {
	if (source.start) {
	    target.start = source.start;
	    target.startoffset = source.startoffset;
	    if (source.end) {
		target.end = source.end;
		target.endoffset = source.endoffset;
	    }
	}
	if (source.pstart) {
	    target.pstart = source.pstart;
	    if (source.pend)
		target.pend = source.pend;
	}
	if (source.xpointer) {
	    target.xpointer = source.xpointer;
	}
	return target;
    },

    add_target: function(id, target) {
	let statement = this.db.createStatement("INSERT INTO targets (id, annotation, part, endpart, fixptr, epubcfi, location, text, endtext, pretext, posttext) VALUES (NULL, :x_annotation, :x_part, :x_endpart, :x_fixptr, :x_epubcfi, :x_location, :x_text, :x_endtext, :x_pretext, :x_posttext)");
	statement.params.x_annotation = id;
	statement.params.x_part = target.part;
	statement.params.x_endpart = target.endpart;
	statement.params.x_fixptr = target.fixptr;
	statement.params.x_epubcfi = target.epubcfi;
	statement.params.x_location =
	    JSON.stringify(this.copy_location(target, {}));
	statement.params.x_text = target.text;
	statement.params.x_endtext = target.endtext;
	statement.params.x_pretext = target.pretext;
	statement.params.x_posttext = target.posttext;
	statement.execute();
    },

    update_target: function(target) {
	let statement = this.db.createStatement("UPDATE targets SET part=:x_part, endpart=:x_endpart, fixptr=:x_fixptr, epubcfi=:x_epubcfi, location=:x_location, text=:x_text, endtext=:x_endtext, pretext=:x_pretext, posttext=:x_posttext WHERE id=:x_id");
	statement.params.x_part = target.part;
	statement.params.x_endpart = target.endpart;
	statement.params.x_fixptr = target.fixptr;
	statement.params.x_epubcfi = target.epubcfi;
	statement.params.x_location =
	    JSON.stringify(this.copy_location(target, {}));
	statement.params.x_text = target.text;
	statement.params.x_endtext = target.endtext;
	statement.params.x_pretext = target.pretext;
	statement.params.x_posttext = target.posttext;
	statement.params.x_id = target.id;
	statement.execute();
    },

    /*
    update_targets: function(id, targets) {
	let statement = this.db.createStatement("DELETE FROM targets WHERE annotation=:x_annotation");
	statement.params.x_annotation = id;
	statement.execute();

	for (let target in loop(targets))
	    this.add_target(id, target);
    },
    */

    /*
    set_position: function(id, position) {
	let statement = this.db.createStatement("UPDATE annotations SET position=:x_position WHERE id=:x_id");
	statement.params.x_position = position;
	statement.params.x_id = id;
	statement.execute();
    },
    */

    update_publication: function(id, sha1, rootfile) {
	let statement = this.db.createStatement("UPDATE publications SET sha1=:x_sha1, rootfile=:x_rootfile WHERE id=:x_id");
	statement.params.x_sha1 = sha1;
	statement.params.x_rootfile = rootfile;
	statement.params.x_id = id;
	statement.execute();
    },

    remove_note: function(id) {
	let statement = this.db.createStatement("DELETE FROM annotations WHERE id=:x_id");
	statement.params.x_id = id;
	statement.execute();
    },

    remove_target: function(id) {
	let statement = this.db.createStatement("DELETE FROM targets WHERE id=:x_id");
	statement.params.x_id = id;
	statement.execute();
    },

    find_all_publications: function() {
	let pubs = [];
	let statement = this.db.createStatement("SELECT id FROM publications");
	try {
	    while (statement.executeStep())
		pubs.push(statement.row.id);
	} finally {
	    statement.reset();
	}
	return pubs;
    },

    find_publication: function(sha1, root) {
	let pid = -1;
	let statement = this.db.createStatement("SELECT id FROM publications WHERE sha1=:x_sha1 and root=:x_root");
	try {
	    statement.params.x_sha1 = sha1;
	    statement.params.x_root = root;
	    if (statement.executeStep())
		pid = statement.row.id;
	} finally {
	    statement.reset();
	}
	return pid;
    },

    find_publication_uid: function(uid) {
	let pid = -1;
	let statement = this.db.createStatement("SELECT id FROM publications WHERE uid=:x_uid");
	try {
	    statement.params.x_uid = uid;
	    if (statement.executeStep())
		pid = statement.row.id;
	} finally {
	    statement.reset();
	}
	return pid;
    },

    get_publication: function(pid) {
	let pub = null;
	let statement = this.db.createStatement("SELECT sha1, root, rootfile, uid, modified, info FROM publications WHERE id=:x_id");
	try {
	    statement.params.x_id = pid;
	    if (statement.executeStep()) {
		pub = {};
		pub.id = pid;
		pub.sha1 = statement.row.sha1;
		pub.root = statement.row.root;
		pub.rootfile = statement.row.rootfile;
		pub.uid = statement.row.uid;
		pub.modified = statement.row.modified;
		let info = statement.row.info;
		if (info)
		    pub.info = JSON.parse(info);
	    }
	} finally {
	    statement.reset();
	}
	return pub;
    },

    find_notes: function(sha1, root) {
	debug("find_notes");
	let pid = this.find_publication(sha1, root);
	if (pid < 0)
	    return [];

	return this.find_notes_id(pid);
    },

    find_notes_id: function(pid) {
	debug("find_notes_id");
	let statement = this.db.createStatement("SELECT id FROM annotations WHERE publication=?1 ORDER BY modified DESC");
	let nids = [];
	try {
	    statement.bindInt32Parameter(0, pid);
	    while (statement.executeStep())
		nids.push(statement.getInt32(0));
	} finally {
	    statement.reset();
	}
	return nids;
    },

    find_part_notes: function(sha1, root, part) {
	debug("get_part_notes");
	let pid = this.find_publication(sha1, root);
	if (pid < 0)
	    return [];

	let statement = this.db.createStatement("SELECT DISTINCT annotations.id FROM annotations, targets WHERE annotations.publication=:x_publication AND targets.part=:x_part");
	let nids = [];
	try {
	    statement.params.x_publication = pid;
	    statement.params.x_part = part;
	    while (statement.executeStep())
		nids.push(statement.getInt32(0));
	} finally {
	    statement.reset();
	}
	return nids;
    },

    find_highlighted_notes: function(sha1, root, part) {
	debug("get_highlighted_notes");
	let id = this.find_publication(sha1, root);
	if (id < 0)
	    return [];

	let statement = this.db.createStatement("SELECT DISTINCT annotations.id FROM annotations, targets WHERE annotations.publication=:x_publication AND annotations.style='highlight' AND targets.part=:x_part");
	let nids = [];
	try {
	    statement.params.x_publication = id;
	    statement.params.x_part = part;
	    while (statement.executeStep())
		nids.push(statement.getInt32(0));
	} finally {
	    statement.reset();
	}
	return nids;
    },

    get_targets: function(nid) {
	let statement = this.db.createStatement("SELECT id, part, endpart, fixptr, epubcfi, location, text, endtext, pretext, posttext FROM targets WHERE annotation=:x_annotation ORDER BY id ASC");
	let targets = [];
	try {
	    statement.params.x_annotation = nid;
	    while (statement.executeStep()) {
		let target = {};
		target.id = statement.row.id;
		target.part = statement.row.part;
		target.endpart = statement.row.endpart;
		target.fixptr = statement.row.fixptr;
		target.epubcfi = statement.row.epubcfi;
		let loc = statement.row.location;
		if (loc) {
		    let location = JSON.parse(loc);
		    this.copy_location(location, target);
		}
		target.text = statement.row.text;
		target.endtext = statement.row.endtext;
		target.pretext = statement.row.pretext;
		target.posttext = statement.row.posttext;
		targets.push(target);
	    }
	} finally {
	    statement.reset();
	}
	return targets;
    },

    get_note: function(id) {
	debug("get_note: " + id);

	let note = null;
	let statement = this.db.createStatement("SELECT uid, created, modified, source, content, type, style FROM annotations WHERE id=:x_id");
	try {
	    statement.params.x_id = id;
	    if (statement.executeStep()) {
		note = {};
		note.id = id;
		note.uid = statement.row.uid;
		//note.annotator = statement.row.annotator;
		let d = statement.row.created.replace(" ", "T") + "Z";
		note.created = new Date(d);
		d = statement.row.modified.replace(" ", "T") + "Z";
		note.modified = new Date(d);
		note.source = statement.row.source;
		note.content = statement.row.content;
		note.type = statement.row.type;
		note.style = statement.row.style;
	    }
	} finally {
	    statement.reset();
	}
	if (note)
	    note.targets = this.get_targets(note.id);
	return note;
    },

    get_notes_path: function(sha1, root) {
	debug("get_notes_path");
	let nids = this.find_notes(sha1, root);
	let notes = [];
	for (let nid in loop(nids))
	    notes.push({ id: nid, targets: this.get_targets(nid) })
	return notes;
    },

    /*
    add_reference: function(nid, endid) {
	let end = this.get_note(endid);
	let statement = this.db.createStatement("INSERT INTO references (id, annotation, linkend, created, contenthash, targetshash) VALUES (NULL, :x_annotation, :x_linkend, datetime('now'), :x_contenthash, :x_targetshash)");
	statement.params.x_annotation = nid;
	statement.params.x_linkend = end.uid;
	statement.params.x_contenthash = this.content_hash(end);
	statement.params.x_targetshash = this.targets_hash(end);
	statement.execute();
    },

    content_hash: function(note) {
	// FIXME
    },

    targets_hash: function(note) {
	// FIXME
    },

    get_references_from: function(nid) {
	let statement = this.db.createStatement("SELECT id, linkend, created, contenthash, targetshash FROM references WHERE annotation=:x_annotation ORDER BY id ASC");
	let refs = [];
	try {
	    statement.params.x_annotation = nid;
	    while (statement.executeStep()) {
		let ref = {};
		ref.id = statement.row.id;
		ref.linkend = statement.row.linkend;
		ref.created = statement.row.created;
		ref.contenthash = statement.row.contenthash;
		ref.targetshash = statement.row.targetshash;
		refs.push(ref);
	    }
	} finally {
	    statement.reset();
	}
	return refs;
    },

    get_references_to: function(linkend) {

    }
    */
};

var notemanagerService = new notemanager();

var notemanagerFactory = {
    createInstance: function(outer, aIID) {
	if (outer != null)
	    throw Components.results.NS_ERROR_NO_AGGREGATION;
	return notemanagerService.QueryInterface(aIID);
    }
};

var components = [notemanager];

if (XPCOMUtils.generateNSGetFactory)
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
else
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);
